#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "netcdf.h"
#include "cdflib.h"

#define MAX_FILE_NAME_LEN  240     /* Max. file name length                 */
#define MAX_ATTRS          3000    /* Max # of netCDF attributes            */
#define UNKNOWN_DT         -99     /* Unsupported netCDF datatype indicator */

#define DEFAULT_MAPPING_FILE   "netcdf_to_cdf_mapping.dat"
#define MAX_REC_LENGTH         NC_MAX_NAME+CDF_ATTR_NAME_LEN+2
                                 /* Max record length of DEFAULT_MAPPING_FILE */

#define TRUE               1
#define FALSE              0

#define MINIMUM(a,b) ((a) < (b) ? (a) : (b))

#define DEFAULT_BYTE_MAXVALUE           (127)
#define DEFAULT_INT1_MAXVALUE           (127)
#define DEFAULT_UINT1_MAXVALUE          ((unsigned char)255)
#define DEFAULT_INT2_MAXVALUE           ((short)32767)
#define DEFAULT_UINT2_MAXVALUE          ((unsigned short)65535)
#define DEFAULT_INT4_MAXVALUE           (2147483647)
#define DEFAULT_INT8_MAXVALUE           ((long long)9223372036854775807LL)
#define DEFAULT_UINT4_MAXVALUE          (4294967294U)
#define DEFAULT_REAL4_MAXVALUE          ((float)1.0E30f)
#define DEFAULT_FLOAT_MAXVALUE          ((float)1.0E30f)
#define DEFAULT_REAL8_MAXVALUE          1.0E30
#define DEFAULT_DOUBLE_MAXVALUE         1.0E30
#define DEFAULT_CHAR_MAXVALUE           ' '
#define DEFAULT_UCHAR_MAXVALUE          ' '
#define DEFAULT_EPOCH_MAXVALUE          1.0E30
#define DEFAULT_EPOCH16_MAXVALUE        1.0E30
#define DEFAULT_TT2000_MAXVALUE         ((long long)9223372036854775807LL)

static int FindEpoch (char *);
static int FindTime (char *);
int MakeLower (int);
int StrStrIgCaseX (char *, char *);
int ConvertData (void *, long, void *, int, int);

static char *staticMapping [] = { /* netCDF -> CDF */
            "Valid_Min",  "VALIDMIN",
            "Valid_Max",  "VALIDMAX",
            "_FillVal",   "FILLVAL",
            "_FillValue", "FILLVAL",
            "FillValue",  "FILLVAL" };

/*************************************
   Global variables and structures
**************************************/
CDFid    id;            /* pointer to the CDF output file                */

/*int  ncid,  */          /* netCDF file ID                                */
/*int  nDims, */          /* Number of dimensions in a netCDF file         */
/*     nVars, */          /* Number of variables in a netCDF file          */
int  nGattrs,           /* Number of global attributes in a netCDF file  */
     unlimitedDimId;    /* Unlimited dimension ID - if the dimension     */
                        /*      ID of a variable matches this ID, the    */
                        /*      variable is a record variable            */
int  unlimitedDimIds[NC_MAX_DIMS], numUnlimdim = 0;
int  *unlimitedDims = NULL;
int  dLevel = -1;
int  setAttributeType = -1; /* An entry is set to global attribute "NC_KIND" */
                            /* for string type attributes as "NC_CHAR" or    */
                            /* "NC_STRING".                                  */

struct ncdim {          /* netCDF dimension structure*/
    char name[NC_MAX_NAME];     /* Dimension name */
    size_t size;                /* Dimension size */
};

struct ncvar {          /* netCDF variable structure*/
    char name[NC_MAX_NAME];     /* Variable name                             */
    char ncname[NC_MAX_NAME];   /* Variable original name in its netCDF group*/
    char group[NC_MAX_NAME];    /* Variable group name                       */
    nc_type type;               /* Variable datatype                         */
    int ndims;                  /* Variable dimensionality                   */ 
    int dimIds[MAX_VAR_DIMS];   /* Dimension IDs                             */
    size_t dimSizes[NC_MAX_DIMS];     /* Dim sizes that correspond to dimIds */
    int natts;                  /* Number of attributes this variable has    */
    int has_fillval;
    double fillval;
    int recordVariable;         /* TRUE = yes, FALSE = no      */
    int coordinateVariable;     /* TRUE = yes, FALSE = no      */
    int deflate;
    int dLevel;
    int shuffle;		/* TRUE = yes, FALSE = no      */
    int storage;
    size_t chunkSizes[NC_MAX_DIMS];
};

struct ncatt {          /* netCDF attribute structure */
    int var;
    char name[NC_MAX_NAME];     /* Attribute name                           */
    nc_type type;               /* Attribute datatype                       */
    size_t len;                 /* # of attribute values                    */
    double *vals;               /* Applicable for the rest of datatypes -   */
};                              /*   i.e. numbers (NC_INT, NC_FLOAT, etc.)  */

struct ncdim dims[NC_MAX_DIMS]; /* Structure to hold dimension info     */
int    totDims = 0;             /* # of dimensions found in a netCDF file   */
int    totVars = 0;             /* # of variables found in a netCDF file   */
int    totAttrs = 0;            /* # of attributes found in a netCDF file   */
size_t totRecs;                 /* # of records written for unlimited dim */
size_t *totRecsX = NULL;
size_t totUnlims;
int    *varLen = NULL;       /* Max len of a string in a NC_STRING variable */
char   **textVars = NULL;
int    numTextVars = 0;
int    toDelimiter = 0;
int    curDiD = 0;
int    maxDiD = 4;
char   *delimiters[] = {"_", "@", "%", "."};

/**************************************************************************/
/*   Arrays to hold the netCDF-to-CDF mapping information that are loaded */
/*   from the default or user-define mapping file.                        */
/**************************************************************************/
char netcdfAttrNames[MAX_ATTRS][NC_MAX_NAME];
char cdfAttrNames[MAX_ATTRS][CDF_ATTR_NAME_LEN+1];
char comments[MAX_ATTRS][81];

int  totAttrsInMappingFile;           /* # of attributes in the mapping file */
char mappingFile[MAX_FILE_NAME_LEN];  /* Mapping file name                   */

struct attrstruct {
    char name[CDF_ATTR_NAME_LEN+1];   /* attribute name                      */
    long scope;                      /* attribute scope                     */
} cdfAttrs[MAX_ATTRS];


struct cdfvar {           /* CDF variable structure */
     char name[CDF_VAR_NAME_LEN+1];
     long num;                        /* variable number          */
     long datatype;                   
     long numElements;                   
     long dimensionality;             /* variable dimensionality      */
     long dimSizes[CDF_MAX_DIMS];     /* variable dimension sizes     */
     long numValues;                  /* # of data values in dimSizes */
};

char *cdfFile = NULL;
int DEBUG = FALSE;
int DELETE = FALSE;
int COLUMN = FALSE;
int ISTPepoch = FALSE;
int ADDdepend = FALSE;
int ADDvartype = FALSE;
int FORCE = FALSE;
int ENFORCE = FALSE;
int compression = TRUE;
int compressionE = FALSE;
int fillvalue = FALSE;
int deleteNULL = TRUE;
int EPOCH = FALSE;
int TIME = FALSE;
int NOEXTRA = FALSE;
int numEpochs = 0;
char **EPOCHnames;
int numTimes = 0;
char **TIMEnames;
int nc_kind;
char **vAttrs = NULL;
int numAttrs = 0;
char *addedAttrs;

/**************************
   Function prototypes
**************************/
void getDimensionInfo (int, int);
void getGroupDimensionInfo (int, int, int *, int);
void doGroups (int);
void processVariable (int, int, int *, int, char *);
void getVariableInfo (int, int, struct ncvar *);
void findTextVariables (int);
void allocate_cdf_var_records (long);
void get_attribute(int, int, struct cdfvar *, int);
int  get_ISTPepoch(int, int);
void getVariableData (int, struct ncvar *, int, struct cdfvar *);
char *NC_datatype (nc_type);
long get_cdf_datatype (nc_type);
void usage();
void create_cdf_file(char *);
long create_cdf_attribute (char *, long, int *);
void create_cdf_gAttribute_entry (long, long, long, long, void *);
void create_cdf_KIND_gAttribute (int);
void create_cdf_DELIMITER_gAttribute ();
void create_cdf_DIMENSIONS_gAttribute (int, int, int);
void create_cdf_DIMENSIONS_vAttribute ();
void create_cdf_GROUP_gAttribute ();
void write_cdf_DIMENSIONS_vAttribute_entry (int, int, int *, char *, int, long);
long create_cdf_var (int, char *group, int id, struct ncvar, struct cdfvar *);
void create_cdf_vAttribute_entry (char *, struct cdfvar *, long, long, void *);
void write_cdf_vardata (long, long, long, long, long *, void *);
void removeFilepath();
void handle_error (int, int);
void memory_error(int, char *, char *);
void cdf_status_handler (CDFstatus, char *);
void insufficient_memory();            /* used for debug */
char *getNextArgument (int, char *[], int);
void parseCmdLine (int, char *[]);
void errorMsg (char *);
void map_netCDF_attr_to_CDF (char *, char *);
char *strlaststr(char *, char *);
char *removeFileExtension(char *, char *);
int  findSpecialAttributes(char *);
long findEpochType(char *);
int  strncmpIgCase (char *, char *, size_t);
int  strcmpIgCase (char *, char *);
int  findIndex (char *, int *);
void findAttrName (char *, char *);
char *kind_string(int);
void searchPreentGroup (char *, char *, char *);
void add_cdf_GROUP_gAttribute (char *);
void checkNetcdfVariableNames(int);
int  FindUnlimitedDim (int);
void CollectUnlimitedDims (int);
int  getFormatAttributeLength (int, int);
void add_Depends (int, int);
void add_VarType (int, int);
void write_cdf_DIMENSION_variables (int, int);
void write_cdf_DIMENSION_depends (int, int);
void add_cdf_attribute_vartype (int, int);
void get_nc_variable_attributes (int, int);
int  search_variable_attributes (char *);
void add_cdf_dimension_extra (long, char *);
void create_cdf_STRING_Attributes (int);
void add_STRINGS_G (int);
void add_STRINGS_V (int);
void DefaultMaxValue (Int32, Int32, void *);

/**************************************************************************** 
   NOTE:

   - If a variable has an unlimited dimension, it is a record variable.
   - A variable with the same name as a dimension is called a coordinate
     variable (e.g., int lat(lat), lon(lon), etc.).  This translation 
     software determines whether a variable is a coordinate variable or 
     not, but the software doesn't retrieve data. 
   - A variable not being a standard coordinate variable nor a record variable
     (e.g., float lat(20) ) will be processed as a coordinate variable as it
     is a fixed sized variable.
*****************************************************************************/
int main(int argc, char *argv[])
{
    FILE  *inFile;                 /* Pointer to the netCDF-CDF mapping file */
    char  rec[MAX_REC_LENGTH];     /* netCDF-to-CDF mapping record           */ 
    int   i, created, count, idx;

    /************************/
    /*  netCDF declaration  */
    /************************/
    int  status,                  /* netCDF status code  */
         attrId;                  /* netCDF attribute ID */

    char *ptr, fileName[MAX_FILE_NAME_LEN];
    char gAttr[CDF_ATTR_NAME_LEN+1];
    int   ncid;		  /* netCDF file ID  */
    int  nVars, nDims;
    int  dimids[NC_MAX_DIMS];
    char attrName[CDF_ATTR_NAME_LEN+1];   /* attribute name */
    /**********************/
    /*  CDF declaration   */
    /**********************/
    CDFstatus   cstatus;           /* CDF status code */
    long        cdfAttrNum;

    if (argc <= 1) 
        usage();             /* netCDF input file name not specified */
    else {
        strcpy(fileName, argv[argc-1]);       /* Get the input file name */
        strcpy(mappingFile, "UNSPECIFIED");
        parseCmdLine(argc, argv);
        if (strcmp(mappingFile, fileName) == 0)
            errorMsg("** Error - either input file or mapping file is missing");
        if (strcmp(mappingFile,"UNSPECIFIED") == 0)  
            strcpy(mappingFile, DEFAULT_MAPPING_FILE);
    }

    /***********************************************/
    /*  Load the netCDF-to-CDF mapping information */
    /***********************************************/
    if ((inFile = fopen(mappingFile, "r")) != NULL) {
      totAttrsInMappingFile = 0;
      while (fgets(rec, MAX_REC_LENGTH, inFile) != NULL) {
        rec[strlen(rec)-1] = '\0';       

         /* Process the mapping record if it's not a comment */
        if (rec[0] != '#'  &&  rec[0] != ' '&& strlen(rec) > 0) {
            sscanf (rec, "%s %s", 
                   netcdfAttrNames[totAttrsInMappingFile], 
                   cdfAttrNames[totAttrsInMappingFile]);
            ptr = (char *) strstr(rec,"//");
            if (ptr != NULL) {
                ptr += 3;
                strcpy(comments[totAttrsInMappingFile], ptr);
            }

            if (DEBUG)
                printf ("%s => %s // %s \n", 
                       netcdfAttrNames[totAttrsInMappingFile],
                       cdfAttrNames[totAttrsInMappingFile],
                       comments[totAttrsInMappingFile]); 

            totAttrsInMappingFile++;
        }
      }
    } else {
      int ix;
      totAttrsInMappingFile = (int) (sizeof (staticMapping) /
                                     (sizeof (char *) * 2));
      for (ix = 0; ix < totAttrsInMappingFile; ++ix) {
        strcpy (netcdfAttrNames[ix], staticMapping[2*ix]);
        strcpy (cdfAttrNames[ix], staticMapping[2*ix+1]);
        if (DEBUG)
            printf ("%s => %s\n",
                   netcdfAttrNames[ix],
                   cdfAttrNames[ix]);
      }
    }

    /**************************************************
     *  Process the netCDF file and create a CDF file 
     **************************************************/
    status = nc_open(fileName, NC_NOWRITE, &ncid);   /* Open the netCDF file */
    if (status != NC_NOERR) handle_error(1, status);
    status = nc_inq_format(ncid, &nc_kind);
    if (status != NC_NOERR) handle_error(-1, status);
    printf ("\nInput file name: %s\n", fileName);

    if (cdfFile == NULL) {
      /* Remove the file path if it exists (/home/mydata.nc => mydata.nc). */
      removeFilepath(fileName);

      /* Strip off the ncf file extension. */
      if (EndsWith(fileName,".nc")) removeFileExtension(fileName, ".nc"); 
      else if (EndsWith(fileName,".ncf")) removeFileExtension(fileName, ".ncf"); 
      else if (EndsWith(fileName,".NC")) removeFileExtension(fileName, ".NC"); 
      else if (EndsWith(fileName,".NCF")) removeFileExtension(fileName, ".NCF"); 
      else if (EndsWith(fileName,".nc4")) removeFileExtension(fileName, ".nc4"); 
      else if (EndsWith(fileName,".NC4")) removeFileExtension(fileName, ".NC4"); 
    } else 
      strcpy (fileName, cdfFile);
    
    create_cdf_file (fileName);

    if (EndsWithIgCase (fileName, ".cdf"))
      printf ("Output file name: %s\n", fileName);
    else
      printf ("Output file name: %s.cdf\n", fileName);

    status = nc_inq_unlimdims(ncid, &numUnlimdim, unlimitedDimIds);
    if (status != NC_NOERR) handle_error(-3, status);

    totUnlims = 0;
    CollectUnlimitedDims (ncid);

    status = nc_inq(ncid, &nDims, &nVars, &nGattrs, NULL);
    if (status != NC_NOERR) handle_error(2, status);
    status = nc_inq_dimids(ncid, &nDims, dimids, 0); 
    if (status != NC_NOERR) handle_error(4, status);

    if (DEBUG) 
        printf ("\nFORMAT: %s\n",kind_string(nc_kind));

    if (nc_kind == NC_FORMAT_NETCDF4) checkNetcdfVariableNames(ncid);

    /* Get the global attributes. */
    if (DEBUG) printf ("\nGlobal attributes: (total: %d)\n", nGattrs);
    for (attrId = 0; attrId < nGattrs; attrId++) 
         get_attribute(ncid, NC_GLOBAL, NULL, attrId);

    /* Create the placeholder attributes */
    for (i=0; i < totAttrsInMappingFile; i++) {
         if (strcmp(netcdfAttrNames[i], "*") == 0) {
             findAttrName (netcdfAttrNames[i], attrName);
             cdfAttrNum = create_cdf_attribute (attrName, GLOBAL_SCOPE,
                                                &created);
             count = findIndex (cdfAttrNames[i], &idx);
             if (count == 1) 
               create_cdf_gAttribute_entry (cdfAttrNum, idx, CDF_CHAR,
                                            (long) strlen(comments[i]), 
                                            comments[i]);
             else
               create_cdf_gAttribute_entry (cdfAttrNum, 0, CDF_CHAR,
                                            (long) strlen(comments[i]), 
                                            comments[i]);
         }
    }

    getDimensionInfo(ncid, nDims);

    findTextVariables (ncid);

    if (!NOEXTRA)
      create_cdf_KIND_gAttribute (nc_kind);
    if (nVars > 0 && !NOEXTRA)
      create_cdf_DIMENSIONS_gAttribute (ncid, nDims, 0);
    if (toDelimiter == 1 && !NOEXTRA)
      create_cdf_DELIMITER_gAttribute ();

    for (i=0; i<nVars; ++i) processVariable (ncid, nDims, dimids, i, "/");

    if (nc_kind == NC_FORMAT_NETCDF4) doGroups(ncid);
    /* Close the netCDF and CDF files */
    if (ADDdepend) add_Depends (ncid, nVars);
    if (ADDvartype) add_VarType (ncid, nVars);
/*  create_cdf_STRING_Attributes (); */
    add_STRINGS_G (ncid);
    add_STRINGS_V (ncid);
    cstatus = CDFlib (CLOSE_, CDF_, NULL_);
    if (cstatus != CDF_OK) cdf_status_handler (cstatus, "CLOSE_, CDF_");
    status = nc_close(ncid);
    if (status != NC_NOERR) handle_error(6, status);
    if (numTextVars > 0) {
      for (i = 0; i < numTextVars; ++i) free (textVars[i]);
      free (textVars);
      free (varLen);
    }
    if (numAttrs > 0) {
      for (i = 0; i < numAttrs; ++i) free (vAttrs[i]);
      free (vAttrs);
    }
    if (numEpochs > 0) {
      for (i = 0; i < numEpochs; ++i) free (EPOCHnames[i]);
      free (EPOCHnames);
    }
    if (numTimes > 0) {
      for (i = 0; i < numTimes; ++i) free (TIMEnames[i]);
      free (TIMEnames);
    }
    if (cdfFile != NULL) free (cdfFile);
    if (DEBUG) {
        printf ("\nnDims = %d, nVars = %d, nGattrs = %d",
                totDims, totVars, nGattrs);
        if (numUnlimdim > 1) {
          printf (", unlimited dim id(s) = ");
          for (i = 0; i < numUnlimdim; ++i) {
            printf ("%d",unlimitedDimIds[i]);
            if (numUnlimdim > 1 && i != (numUnlimdim-1)) printf(", ");
          }
        }
        printf ("\n\n");
    }
    if (totUnlims > 0) {
      free (totRecsX);
      free (unlimitedDims);
    }
    exit (0);
}


/*------------------------------------------------------------------
 * Get the dimension names and their sizes for the root group in
 * the netCDF file.
 *
 *   Global variables set: dims 
 *   Global variables used: unlimitedDimId (set in main)
 *-----------------------------------------------------------------*/
void getDimensionInfo (int ncid, int nDims)
{
   int dimId, status, maxRecs, i, found;

   if (DEBUG) printf ("\nRoot Dimensions: ncid=%d\n",ncid);
   for (dimId=0; dimId < nDims; dimId++) {
       status = nc_inq_dim(ncid, dimId,            /* in  */
                           dims[dimId].name,       /* out - dimension name */
                           &dims[dimId].size);     /* out - dimension size */
       if (status != NC_NOERR) handle_error(7, status);
       found = 0;
       if (numUnlimdim > 0) {
         for (i = 0; i < numUnlimdim; ++i) {
           if (dimId == unlimitedDimIds[i]) {
             found = 1;
             break;
           }
         }
       }

       if (found == 1) {
           maxRecs = dims[dimId].size;
           if (DEBUG) printf ("\t%s = %s ; // (%d records currently)\n",
                              dims[dimId].name, "UNLIMITED", maxRecs);
       }
       else 
           if (DEBUG) printf ("\t%s = %ld ;\n", dims[dimId].name,
                              (long)dims[dimId].size);
   }
   totDims += nDims;
}

/*------------------------------------------------------------------
 * Get the dimension names and their sizes in a group in a netCDF file.
 *
 *   Global variables set: dims 
 *   Global variables used: unlimitedDimId (set in main)
 *-----------------------------------------------------------------*/
void getGroupDimensionInfo (int ncid, int nDims, int *dimids, int start)
{
   int i, dimId, status, maxRecs, found;
   char *fullName, *grpName, grpId;
   size_t len;

   if (DEBUG) printf ("\nGroup Dimensions: ncid=%d start=%d\n",ncid, start);
   for (dimId=0; dimId < nDims; dimId++) {
       status = nc_inq_dim(ncid, dimids[dimId],       /* in  */
                           dims[dimId+start].name,    /* out - dimension name */
                           &dims[dimId+start].size);  /* out - dimension size */
       if (status != NC_NOERR) handle_error(8, status);
       found = FindUnlimitedDim (dimId);
       if (found > -1) maxRecs = dims[dimId+start].size;
   }

   status = nc_inq_grpname_len(ncid, &len);
   if (status != NC_NOERR) handle_error(9, status);
   grpName = (char *) malloc (len +1);
   status = nc_inq_grpname_full(ncid, &len, grpName);
   if (status != NC_NOERR) handle_error(10, status);
   for (dimId=0; dimId < nDims; dimId++) {
       fullName = (char *) malloc (len + strlen(dims[dimId+start].name)+1);
       strcpy (fullName, grpName+1);
       for (i = 0; i < (int) strlen(fullName); ++i)
         if (*(fullName+i) == (char) '/')
           *(fullName+i) = (char)*delimiters[curDiD];
       add_cdf_GROUP_gAttribute (fullName);
       strcat (fullName, delimiters[curDiD]);
       strcat (fullName, dims[dimId+start].name);
       strcpy (dims[dimId+start].name, fullName);
       found = FindUnlimitedDim (dimId);
       if (found > -1) {
           if (DEBUG) printf ("\t%s = %s ; // (%d records currently)\n",
                              dims[dimId+start].name, "UNLIMITED", maxRecs);
       }
       else 
           if (DEBUG) printf ("\t%s = %ld ;\n", dims[dimId+start].name,
                              (long)dims[dimId+start].size);
       free (fullName);
   }
   free (grpName);
}

/*--------------------------------------------------------------------------
 *  Read and process the netCDF variables in a group.
 *--------------------------------------------------------------------------*/
void doGroups(int ncid)
{
   int status, i, j, grpno, grpno2, *grpids, dimids[NC_MAX_DIMS];
   int nVars, nDims, varids[NC_MAX_DIMS];
   size_t len;
   char grpName[NC_MAX_NAME+1];

   status = nc_inq_grps (ncid, &grpno, NULL);
   if (status != NC_NOERR) handle_error (11, status);
   if (grpno > 0) {
     create_cdf_GROUP_gAttribute();
     grpids = malloc(sizeof(int) * grpno);
     status = nc_inq_grps (ncid, NULL, grpids);
     if (status != NC_NOERR) handle_error (12, status);
     for (i = 0; i < grpno; ++i) {
       status = nc_inq_grpname_full (grpids[i], &len, grpName);
       if (status != NC_NOERR) handle_error (13, status);
       if (len > 1) {
         for (j = 0; j < len; ++j)
           if (*(grpName+j) == (char) '/')
             *(grpName+j) = (char)*delimiters[curDiD];
       }
       status = nc_inq_dimids (grpids[i], &nDims, dimids, 0);
       if (status != NC_NOERR) handle_error (14, status);
       getGroupDimensionInfo (grpids[i], nDims, dimids, totDims);
       if (nDims > 0) {
         if (!NOEXTRA) create_cdf_DIMENSIONS_gAttribute (grpids[i],
                                                         nDims, totDims);
         totDims += nDims;
       }
       /* Find all variables in this group. */
       status = nc_inq_varids (grpids[i], &nVars, varids);
       if (status != NC_NOERR) handle_error (15, status);
       for (j = 0; j < nVars; ++j)
         processVariable (grpids[i], nDims, dimids, varids[j], grpName);
       /* Check if there are any sub-groups. */
       status = nc_inq_grps (grpids[i], &grpno2, NULL);
       if (status != NC_NOERR) handle_error (16, status);
       if (grpno2 > 0) doGroups (grpids[i]); 
     }
     free(grpids);
   }
}

/*--------------------------------------------------------------------------
 *  Check all variable names if there is a group(s). So, the delimiter used
 *  to combine group name(s) and variable name can be determined (into CDF
 *  variable name).
 *--------------------------------------------------------------------------*/
void checkNetcdfVariableNames(int ncid)
{
   int status, i, j, k, grpno, grpno2, *grpids, dimidsX[NC_MAX_DIMS];
   int type, nattsX, nVarsX, nDimsX, nGattrsX, varidsX[NC_MAX_DIMS];
   int numUnlimdimX;
   size_t len;
   char grpName[NC_MAX_NAME+1], name[NC_MAX_NAME+1];
   status = nc_inq(ncid, &nDimsX, &nVarsX, &nGattrsX, &numUnlimdimX);
   if (status != NC_NOERR) handle_error(302, status);
   status = nc_inq_unlimdims(ncid, &numUnlimdim, unlimitedDimIds);
   if (status != NC_NOERR) handle_error(303, status);
   for (i  = 0; i < nVarsX; ++i) {
     status = nc_inq_var(ncid, i, name, &type, &nDimsX, dimidsX, &nattsX);
     if (status != NC_NOERR) handle_error(304, status);
     for (j = curDiD; j < maxDiD; ++j) {
       if (strstr(name, delimiters[j]) == NULL) break;
       ++curDiD;
     }
   }
   status = nc_inq_grps (ncid, &grpno, NULL);
   if (status != NC_NOERR) handle_error (305, status);
   if (grpno > 0) {
     toDelimiter = 1;
     grpids = malloc(sizeof(int) * grpno);
     status = nc_inq_grps (ncid, NULL, grpids);
     if (status != NC_NOERR) handle_error (306, status);
     for (i = 0; i < grpno; ++i) {
       status = nc_inq_grpname_full (grpids[i], &len, grpName);
       if (status != NC_NOERR) handle_error (307, status);
       for (j = curDiD; j < maxDiD; ++j) {
         if (strstr(grpName, delimiters[j]) == NULL) break;
         ++curDiD;
       }
       /* Find all variables in this group. */
       status = nc_inq_varids (grpids[i], &nVarsX, varidsX);
       if (status != NC_NOERR) handle_error (308, status);
       for (j = 0; j < nVarsX; ++j) {
         status = nc_inq_var(grpids[i], j, name, &type, &nDimsX, dimidsX,
                             &nattsX);
         if (status != NC_NOERR) handle_error(309, status);
         for (k = curDiD; k < maxDiD; ++k) {
           if (strstr(name, delimiters[k]) == NULL) break;
           ++curDiD;
         }
       }
       /* Check if there are any sub-groups. */
       status = nc_inq_grps (grpids[i], &grpno2, NULL);
       if (status != NC_NOERR) handle_error (310, status);
       if (grpno2 > 0) checkNetcdfVariableNames (grpids[i]); 
     }
     free(grpids);
   }
}

/*--------------------------------------------------------------------------
 *  Read and process a netCDF variable.  This routine reads one netCDF
 *  variable (including data and variable attributes) and create the
 *  corresponding CDF variable.
 *
 *   Global variables used: dims  (set in getDimensionInfo)
 *--------------------------------------------------------------------------*/
void processVariable (int ncid, int ngdims, int *dimids, int varId,
                      char *grpName)
{
/*  int   varId, attrId, status; */
  int   attrId;
  long  varNum, attrNum;
  CDFstatus status;

  struct ncvar var;           /* Structure for a netCDF variable */
  struct ncatt attr;          /* Structure for a netCDF variable attribute */
  struct cdfvar cdfVar;       /* Structure for a CDF variable */

  if (DEBUG) printf ("\nVariable: ncid=%d varid=%d\n",ncid,varId);

  if (varId == 0 && !NOEXTRA) create_cdf_DIMENSIONS_vAttribute ();

  getVariableInfo (ncid, varId, &var);

  if (nc_kind != NC_FORMAT_CLASSIC && !NOEXTRA) {

    if (var.type == NC_STRING ||
        (var.type == NC_CHAR && (var.ndims > 0 &&
                                 var.dimSizes[var.ndims-1] > 1))) {
      status = CDFlib (CREATE_, ATTR_, "NC_STRING", VARIABLE_SCOPE, &attrNum,
                       NULL_);
      if (status != CDF_OK && status != ATTR_EXISTS) {
        cdf_status_handler (status, "000...CREATE_, ATTR_");
        return;
      }
    }

    status = CDFlib (CREATE_, ATTR_, "NC_DEFLATE", VARIABLE_SCOPE, &attrNum,
                     NULL_);
    if (status != CDF_OK && status != ATTR_EXISTS) {
      cdf_status_handler (status, "111...CREATE_, ATTR_");
      return;
    }

    status = CDFlib (CREATE_, ATTR_, "NC_SHUFFLE", VARIABLE_SCOPE, &attrNum,
                     NULL_);
    if (status != CDF_OK && status != ATTR_EXISTS) {
      cdf_status_handler (status, "222...CREATE_, ATTR_");
      return;
    }

    status = CDFlib (CREATE_, ATTR_, "NC_STORAGE", VARIABLE_SCOPE, &attrNum,
                     NULL_);
    if (status != CDF_OK && status != ATTR_EXISTS) {
      cdf_status_handler (status, "333...CREATE_, ATTR_");
      return;
    }

    status = CDFlib (CREATE_, ATTR_, "NC_CHUNKSIZES", VARIABLE_SCOPE, &attrNum,
                     NULL_);
    if (status != CDF_OK && status != ATTR_EXISTS) {
      cdf_status_handler (status, "444...CREATE_, ATTR_");
      return;
    }
  }

  varNum = create_cdf_var (ncid, grpName, varId, var, &cdfVar);  
  if (varNum == UNKNOWN_DT) {     /* unknown datatype */
    if (FORCE) return;            /* Skip this variable. */
    else {
      printf ("An error has occurred, halting... The output file is invalid.\n");
      exit (1);
    }
  }
  cdfVar.num = varNum;

  if (nc_kind != NC_FORMAT_CLASSIC && !NOEXTRA) {
    if (var.type == NC_STRING) {
      status = CDFlib (SELECT_, ATTR_NAME_, "NC_STRING",
                                zENTRY_, cdfVar.num,
                       PUT_, zENTRY_DATA_, (long) CDF_CHAR, (long) 3, "yes",
                       NULL_);
      if (status != CDF_OK) {
        cdf_status_handler (status, "000...PUT_, ENTRY_");
        return;
      }
    } else if (var.type == NC_CHAR && 
               (var.ndims > 0 && var.dimSizes[var.ndims-1] > 1)) {
      status = CDFlib (SELECT_, ATTR_NAME_, "NC_STRING",
                                zENTRY_, cdfVar.num,
                       PUT_, zENTRY_DATA_, (long) CDF_CHAR, (long) 2, "no",
                       NULL_);
      if (status != CDF_OK) {
        cdf_status_handler (status, "000...PUT_, ENTRY_");
        return;
      }
    }

    status = CDFlib (SELECT_, ATTR_NAME_, "NC_DEFLATE",
                              zENTRY_, cdfVar.num,
                     PUT_, zENTRY_DATA_, (long) CDF_INT4, (long) 1, &var.dLevel,
                     NULL_);
    if (status != CDF_OK) {
      cdf_status_handler (status, "111...PUT_, ENTRY_");
      return;
    }

    if (var.shuffle == 1) {
      status = CDFlib (SELECT_, ATTR_NAME_, "NC_SHUFFLE",
                                zENTRY_, cdfVar.num,
                       PUT_, zENTRY_DATA_, (long) CDF_CHAR, (long) 4, "true",
                       NULL_);
    } else {
      status = CDFlib (SELECT_, ATTR_NAME_, "NC_SHUFFLE",
                                zENTRY_, cdfVar.num,
                       PUT_, zENTRY_DATA_, (long) CDF_CHAR, (long) 5, "false",
                       NULL_);
    }
    if (status != CDF_OK) {
      cdf_status_handler (status, "222...PUT_, ENTRY_");
      return;
    }
  
    if (var.storage == NC_CONTIGUOUS) {
      status = CDFlib (SELECT_, ATTR_NAME_, "NC_STORAGE",
                                zENTRY_, cdfVar.num,
                       PUT_, zENTRY_DATA_, (long) CDF_CHAR, (long) 10,
                                           "contiguous",
                       NULL_);
    } else if (var.storage == NC_CHUNKED) {
      status = CDFlib (SELECT_, ATTR_NAME_, "NC_STORAGE",
                                zENTRY_, cdfVar.num,
                       PUT_, zENTRY_DATA_, (long) CDF_CHAR, (long) 7, "chunked",
                       NULL_);
      if (status != CDF_OK) {
        cdf_status_handler (status, "333...PUT_, ENTRY_");
        return;
      }
      if (var.chunkSizes[0] != 0) {
        int chunks[NC_MAX_DIMS];
        int ix;
        for (ix=0; ix<var.ndims; ++ix) chunks[ix] = (int) var.chunkSizes[ix];
        status = CDFlib (SELECT_, ATTR_NAME_, "NC_CHUNKSIZES",
                                  zENTRY_, cdfVar.num,
                         PUT_, zENTRY_DATA_, (long) CDF_INT4, (long) var.ndims,
                               chunks,
                         NULL_);
      }
    } else
      status = CDF_OK;
    if (status != CDF_OK) {
      cdf_status_handler (status, "555...PUT_, ENTRY_");
      return;
    }
  }

  /******************************************************************/
  /* If the number of records to be written is known, allocate that */
  /* number of records before before writing data to the current    */
  /* zVariable created in cdf_create_cdf_var.  Allocating records   */
  /* ahead of time is optional and often produces an efficient I/O. */
  /******************************************************************/
  if (var.recordVariable && !compression)
      allocate_cdf_var_records ((long)var.dimSizes[0]);  /* # of records */

  /* Get the variable attributes */
  for (attrId=0; attrId < var.natts; attrId++) 
       get_attribute (ncid, varId, &cdfVar, attrId);     /* all 'in' */

  if (!NOEXTRA)
    write_cdf_DIMENSIONS_vAttribute_entry (ncid, ngdims, dimids, grpName,
                                           varId, varNum);

  /***************************************************************/
  /* Get data only if it is not a record variable, or if it is a */
  /* record variable and at least one record has been written.   */
  /***************************************************************/

  if (var.ndims == 0 ||                     /* Variable is a scalar  */
      !var.recordVariable ||
      (dims[0].size != 0)) {     /* Rec var & not empty   */
      getVariableData (ncid,
                       &var,          /* in - netCDF variable structure */
                       varId,         /* in - netCDF variable ID */
                       &cdfVar);      /* in/out - CDF variable structure */
  }

  ++totVars;
}


/*------------------------------------------------------------------------
 *  Get the netCDF variable information pointed by 'varId'.
 *  An netCDF variable should be either a record variable or coordinate
 *  variable. But, for our processing, a coordinate variable as time(time)
 *  while time is UNLIMITED dimension, this variable is also classified as
 *  record variable.
 *------------------------------------------------------------------------*/
void getVariableInfo (int ncid,
                      int varId,             /* in - variable ID    */
                      struct ncvar *var)     /* out - variable info */
{
    int i, j, k, status, dimId, attrId, found; 
    size_t len; char *fullName, *grpName;

    /* Get the variable information. */
    status = nc_inq_var(ncid, varId,    /* in */
                        var->name,       /* out - variable name            */
                        &var->type,      /* out - netCDF datatype          */
                        &var->ndims,     /* out - dimensionality           */
                        var->dimIds,     /* out - dimension IDs            */
                        &var->natts);    /* out - # of variable attributes */
    if (status != NC_NOERR) handle_error(17, status);
    status = nc_inq_var_deflate(ncid, varId, &var->shuffle, &var->deflate,
                                &var->dLevel);
    if (status != NC_NOERR) handle_error(18, status);
    for (i = 0; i < NC_MAX_DIMS; ++i) var->chunkSizes[i] = 0;
    status = nc_inq_var_chunking(ncid, varId, &var->storage, var->chunkSizes);
    if (status != NC_NOERR) handle_error(19, status);
    status = nc_inq_grpname_len(ncid, &len);
    if (status != NC_NOERR) handle_error(20, status);
    grpName = (char *) malloc (len +1);
    status = nc_inq_grpname_full (ncid, &len, grpName);
    if (status != NC_NOERR) handle_error(21, status);
    strcpy (var->group, grpName);
    strcpy (var->ncname, var->name);
    if (len > 1) { /* More than just a "/" */
      fullName = (char *) malloc (strlen(var->name)+len+1); 
      strcpy (fullName, grpName+1);
      strcat (fullName, delimiters[curDiD]);
      strcat (fullName, var->name);
      strcpy (var->name, fullName);
      for (i = 0; i < (int) strlen(var->name); ++i)
        if (*(var->name+i) == (char) '/')
          *(var->name+i) = (char)*delimiters[curDiD];
      free (fullName);
    }
    free (grpName);

    if (DEBUG) {
        printf ("\n\t%d. %s %s", varId, NC_datatype(var->type), var->name);
        if (var->ndims > 0) printf ("(");
    }
    /* Find out if this is a coordinate variable and/or a record variable.  */
    var->recordVariable     = FALSE;
    var->coordinateVariable = FALSE;
    if (var->ndims > 0) {
        if (var->ndims == 1) {
            for (dimId = 0; dimId < totDims; /*nDims */ dimId++) {
                 if (strcmp(dims[dimId].name, var->name) == 0) {
                     var->coordinateVariable = TRUE;
                     break;
                 }
            }
         }
         found = FindUnlimitedDim (var->dimIds[0]);
         if (found > -1) var->recordVariable = TRUE;
    }

    /* if (!var.coordinateVariable) continue; */

    for (i=0; i < var->ndims; i++) {
         var->dimSizes[i] = dims[var->dimIds[i]].size;
         if (DEBUG) printf ("%d%s", (int) dims[var->dimIds[i]].size,
                            i < var->ndims-1 ? ", " : ")");
    }
    if (DEBUG) {
        if (var->recordVariable) printf (" - record variable");
        if (var->coordinateVariable) printf (" - coordinate variable");
        else if (!var->recordVariable) printf (" - fixed-size variable");
        printf ("\n");
    }
}


/*------------------------------------------------------------------------
 *  Find all the NC_STRING variables and determine their length (as the
 *  number of elements in CDF).
 *------------------------------------------------------------------------*/
void findTextVariables (int ncid)
{
    int i, j, k, found, status; 
    struct ncvar var;
    int nDims, nVars, nGattrs;
    int recs; size_t start[NC_MAX_DIMS], count[NC_MAX_DIMS];
    int values, unlimitID, totalValues;
    char **strings;

    status = nc_inq(ncid, &nDims, &nVars, &nGattrs, NULL);
    if (status != NC_NOERR) handle_error(2, status);
    for (i = 0; i < nVars; ++i) {
    /* Get the variable information. */
      status = nc_inq_var(ncid, i,         /* in */
                          var.name,       /* out - variable name            */
                          &var.type,      /* out - netCDF datatype          */
                          &var.ndims,     /* out - dimensionality           */
                          var.dimIds,     /* out - dimension IDs            */
                          &var.natts);    /* out - # of variable attributes */
      if (status != NC_NOERR) handle_error(17, status);
      if (var.type == NC_STRING) ++numTextVars;
    }
    if (numTextVars > 0) {
      varLen = (int *) malloc ((size_t) sizeof(int) * numTextVars);
      textVars = (char **) malloc ((size_t) sizeof(char *) * numTextVars);
      for (j = 0; j < numTextVars; ++j) {
        textVars[j] = (char *) malloc (sizeof(char) * NC_MAX_NAME);
        varLen[j] = 0;
      }
    }
    j = 0;
    for (i = 0; i < nVars; ++i) {
      values = 1;
    /* Get the variable information. */    
      status = nc_inq_var(ncid, i,         /* in */
                          var.name,       /* out - variable name            */
                          &var.type,      /* out - netCDF datatype          */
                          &var.ndims,     /* out - dimensionality           */
                          var.dimIds,     /* out - dimension IDs            */
                          &var.natts);    /* out - # of variable attributes */
      if (status != NC_NOERR) handle_error(17, status);
      if (var.type != NC_STRING) continue;
      for (k = 0; k < var.ndims; ++k) start[k] = 0;
      found = FindUnlimitedDim (var.dimIds[0]);
      if ((var.ndims > 0) && (found > -1)) {
         count[0] = (int) totRecsX[found];
         if (var.ndims > 1) {
           size_t len;
           for (k = 1; k < var.ndims; ++k) {
             status = nc_inq_dimlen (ncid, var.dimIds[k], &len);
             if (status != NC_NOERR) handle_error(22, status);
             count[k] = len;
             values *= len;
           }
         }
         recs = totRecsX[found];
      } else {
         if (var.ndims > 0) {
           size_t len;
           for (k = 0; k < var.ndims; ++k) {
             status = nc_inq_dimlen (ncid, var.dimIds[k], &len);
             if (status != NC_NOERR) handle_error(23, status);
             count[k] = len;
             values *= len;
           }
         }
         recs = 1;
      }
      strcpy(textVars[j], var.name);
      totalValues = recs * values;
      if (totalValues > 0) {
        strings = (char **) malloc (sizeof(char *) * totalValues); 
        status = nc_get_vara_string(ncid, i, start, count, strings);
        if (status != NC_NOERR) handle_error(23, status);
        for (k = 0; k < totalValues; ++k) {
          if (strings[k] != NULL && strlen(strings[k]) > varLen[j])
            varLen[j] = (int) strlen (strings[k]);
        }
        nc_free_string (totalValues, strings);
/*      if (strings != NULL) free (strings); */
      }
      if (varLen[j] == 0) {
        varLen[j] = getFormatAttributeLength (ncid, i);
      }
      ++j;
    }
}


/*----------------------------------------------------------------------------
 *  This routine specifies the number of records to allocate for the 
 *  current zVariable.  The records are allocated beginning at record
 *  number 0 (zero).  This operation is optional and only applicable to
 *  uncompressed zVariables in single-file CDFs.
 *---------------------------------------------------------------------------*/
void allocate_cdf_var_records (long nRecs)   /* # of records to be allocated */
{
   long status;
   status = CDFlib (PUT_, zVAR_ALLOCATERECS_, nRecs,
                    NULL_);
   if (status != CDF_OK) 
       cdf_status_handler (status, "PUT_, zVAR_ALLOCATERECS_");
}

 
/*----------------------------------------------------------------------------
 *  This routine retrieves the netCDF attribute (global or variable) name
 *  and its data for the given netCDF variable ID and the attribute ID.  
 *  The retrieved attribute information is then used to create either a 
 *  CDF global or variable attribute.
 *
 *   Global variables set and used: cdfAttrs, totAttrs
 *---------------------------------------------------------------------------*/
void get_attribute(int ncid,             /* in - netCDF id            */
                   int varId,            /* in - netCDF variable ID   */
                   struct cdfvar *var,   /* in - cdf variable struct  */
                   int attrId)           /* in - attribute ID         */
{
    int           status, i, j, attrAlreadyCreated, emptyAttr;
    long          cdfAttrNum, cdfAttrEntryNum,
                  cdfAttrScope, cdfDatatype;
    
    char          *cPtr, attrName[CDF_ATTR_NAME_LEN+1];
    signed char   *scPtr;
    unsigned char *uscPtr;
    short         *sPtr;
    unsigned short *usPtr;
    int           *iPtr;
    unsigned int  *uiPtr;
    long long     *jPtr;
    float         *fPtr;
    double        *dPtr;
    struct ncatt  attr;           /* attribute */
    int           created, idx, count, varType;


    /*****************************************************************/
    /* Get the attribute name for the given ncid, varId, and attrId  */ 
    /* If varId = NC_GLOBAL, attrId represents a global attribute ID */
    /*****************************************************************/
    status = nc_inq_attname(ncid,           /* In - netCDF ID               */
                            varId,          /* In - netCDF variable ID      */
                            attrId,         /* In - netCDF var attribute ID */
                            attr.name);     /* Out - netCDF attribute name  */
    if (status != NC_NOERR) handle_error(24, status);

    if (varId != NC_GLOBAL) {
        if (DEBUG) printf ("\t"); 
        cdfAttrScope = VARIABLE_SCOPE;        /* Variable attribute */
    }
    else
        cdfAttrScope = GLOBAL_SCOPE;          /* Global attribute   */
    if (DEBUG) {
        if (varId != NC_GLOBAL)
            printf ("\t%s:%s ", var->name, attr.name);
        else
            printf ("\t%s ", attr.name);
    }

    /**************************************************************/
    /* Get the attribute information for the given attribute name */
    /**************************************************************/
    status = nc_inq_att(ncid,                /* in - netCDF file ID  */
                        varId,               /* in - variable ID     */
                        attr.name,           /* in - attribute name  */
                        &attr.type,          /* out - Attribute type */
                        &attr.len);          /* out - # of attribute values */
                                             /*       or the length of string */
    if (status != NC_NOERR) handle_error(25, status);
    /***********************************************************/
    /* Map the netCDF datatype to an appropriate CDF datatype. */
    /***********************************************************/
    /* Fix an error for int64 value with ****L for netCDF v4.3.0 */ 
    status = nc_inq_vartype(ncid, varId, &varType);
    if (nc_kind == NC_FORMAT_NETCDF4 && varType == NC_INT64 &&
        attr.type == NC_UINT64)
      attr.type = NC_INT64;
    cdfDatatype = get_cdf_datatype (attr.type);
    if (cdfDatatype == UNKNOWN_DT) {
        printf ("*** ERROR: netCDF datatype:%d for attribute \"%s\" is "
                "unsupported\n", attr.type, attr.name);
        return;
    }

    /************************************************************************
     * IMPORTANT NOTE:
     *
     * CDF does not allow duplicate attributes.  If the attribute is already
     * created, do not create one.  netCDF's variable attribute name can be 
     * same as one of the global attributes that is contrary to CDF.  
     ************************************************************************/
    attrAlreadyCreated = FALSE; 

    map_netCDF_attr_to_CDF(attr.name,     /* in - netCDF attribute name */
                           attrName);     /* out - mapped CDF attribute name */

    if (totAttrs > 0) {
        for (i=0; i < totAttrs; i++) {
             if (strcmp(cdfAttrs[i].name, attr.name) == 0) {
                 if (cdfAttrs[i].scope == VARIABLE_SCOPE) {
                     attrAlreadyCreated = TRUE;
                 }
                 else {   /* GLOBAL_SCOPE - this varAttr name = gAttr name */
                          /* add '_' to make variable attr name unique */
                     strcat(attrName, "_"); 
 
                     /*****************************************************/
                     /* Check whether the current variable attribute name */
                     /* has been already created.                         */
                     /*****************************************************/
                     for (j=0; j < totAttrs; j++) {
                          if ( (strcmp(cdfAttrs[j].name, attr.name) == 0) && 
                               (cdfAttrs[j].scope == VARIABLE_SCOPE)) {
                              attrAlreadyCreated = TRUE;
                              break;
                          }
                     }
                 }
                 break;
             }
        }
    }

    if (!attrAlreadyCreated) {
        cdfAttrNum = create_cdf_attribute (attrName, cdfAttrScope, &created);
        if (created) {
          cdfAttrs[totAttrs].scope = cdfAttrScope;
          strcpy (cdfAttrs[totAttrs++].name, attr.name);
        }
    }

    emptyAttr = FALSE;
    if (attr.len == 0) {
      if (attr.type == NC_CHAR ||
          attr.type == NC_STRING) { /* show 0-length entry as empty string */
        CDFstatus status;
        long attrNum;
        if (cdfAttrScope == GLOBAL_SCOPE) {
          status = CDFlib (CREATE_, ATTR_, attr.name, GLOBAL_SCOPE, &attrNum,
                           NULL_);
          if (status != CDF_OK && status != ATTR_EXISTS)
            cdf_status_handler (status, "333...CREATE_, ATTR_");
          return;
        } else {
          status = CDFlib (CREATE_, ATTR_, attr.name, VARIABLE_SCOPE, &attrNum,
                           NULL_);
          if (status != CDF_OK && status != ATTR_EXISTS)
            cdf_status_handler (status, "444...CREATE_, ATTR_");
          return;
        }
      } else {
        attr.type = NC_CHAR;
        attr.len = 1;
        emptyAttr = TRUE;
      }
    }

    /****************************************/
    /* Get the attribute value(s).          */
    /****************************************/
    switch (attr.type) {
        case NC_CHAR:
            if (cdfAttrScope == GLOBAL_SCOPE && setAttributeType == -1) {
                setAttributeType = NC_CHAR;
            }
            cPtr = (char *) malloc(attr.len * sizeof(char) + 1);
            if (cPtr == NULL) memory_error(1, "get_attribute", "NC_CHAR");
            if (!emptyAttr) {
                status =nc_get_att_text(ncid, varId, attr.name, cPtr);
                if (status != NC_NOERR) handle_error(26, status);
            }
            else
                *cPtr = (char) '\0';
            *(cPtr+attr.len) = '\0';               /* End of string mark */
            if (DEBUG) {
               if (!emptyAttr) printf (" = \"%s\"\n", cPtr); /* xxx */
               else printf (" = \"\"\n"); /* xxx */
            }
            if (cdfAttrScope == GLOBAL_SCOPE) {
                count = findIndex (attrName, &idx);
                if (count == 1)
                  create_cdf_gAttribute_entry (cdfAttrNum, idx, cdfDatatype, 
                                               (long) attr.len, cPtr);
                else
                  create_cdf_gAttribute_entry (cdfAttrNum, 0, cdfDatatype, 
                                               (long) attr.len, cPtr);
/*                  create_cdf_gAttribute_entry (cdfAttrNum, 0, cdfDatatype, */
/*                                               (long) 0, cPtr); */
            } else
                create_cdf_vAttribute_entry (attrName, var, cdfDatatype, 
                                             (long) attr.len, cPtr);
            free(cPtr);
            break;

        case NC_BYTE:
            scPtr = (signed char *) malloc (sizeof(signed char) * attr.len);
            if (scPtr == NULL) memory_error(2, "get_attribute", "NC_BYTE");
            status = nc_get_att_schar(ncid, varId, attr.name, scPtr);
            if (status != NC_NOERR) handle_error(27, status);
            if (DEBUG) {
                printf (" = ");
                for (i=0; i < attr.len; i++) {
                     if (i > 0) printf (", ");
                     printf ("%d", *(scPtr+i));
                }
            }
            if (cdfAttrScope == GLOBAL_SCOPE) {
                count = findIndex (attrName, &idx);
                if (count == 1)
                  create_cdf_gAttribute_entry (cdfAttrNum, idx, cdfDatatype,
                                               (long) attr.len, scPtr);
                else
                  create_cdf_gAttribute_entry (cdfAttrNum, 0, cdfDatatype, 
                                               (long) attr.len, scPtr);
            } else
                create_cdf_vAttribute_entry (attrName, var, cdfDatatype, 
                                             (long) attr.len, scPtr);
            free (scPtr);
            break;

        case NC_UBYTE:
            uscPtr = (unsigned char *) malloc (sizeof(signed char) * attr.len);
            if (uscPtr == NULL) memory_error(3, "get_attribute", "NC_UBYTE");
            status = nc_get_att_uchar(ncid, varId, attr.name, uscPtr);
            if (status != NC_NOERR) handle_error(28, status);
            if (DEBUG) {
                printf (" = ");
                for (i=0; i < attr.len; i++) {
                     if (i > 0) printf (", ");
                     printf ("%ud", *(uscPtr+i));
                }
            }
            if (cdfAttrScope == GLOBAL_SCOPE) {
                count = findIndex (attrName, &idx);
                if (count == 1)
                  create_cdf_gAttribute_entry (cdfAttrNum, idx, cdfDatatype,
                                               (long) attr.len, uscPtr);
                else
                  create_cdf_gAttribute_entry (cdfAttrNum, 0, cdfDatatype, 
                                               (long) attr.len, uscPtr);
            } else
                create_cdf_vAttribute_entry (attrName, var, cdfDatatype, 
                                             (long) attr.len, uscPtr);
            free (uscPtr);
            break;

        case NC_SHORT:
            sPtr = (short *) malloc (sizeof(short) * attr.len);
            if (sPtr == NULL) memory_error(4, "get_attribute", "NC_SHORT");
            status = nc_get_att_short(ncid, varId, attr.name, sPtr);
            if (status != NC_NOERR) handle_error(29, status);
            if (DEBUG) {
                printf (" = ");
                for (i=0; i < attr.len; i++) {
                     if (i > 0) printf (", ");
                     printf ("%d", *(sPtr+i));
                }
            }
            if (cdfAttrScope == GLOBAL_SCOPE) {
                count = findIndex (attrName, &idx);
                if (count == 1)
                  create_cdf_gAttribute_entry (cdfAttrNum, idx, cdfDatatype,
                                               (long) attr.len, sPtr);
                else
                  create_cdf_gAttribute_entry (cdfAttrNum, 0, cdfDatatype, 
                                               (long) attr.len, sPtr);
            } else
                create_cdf_vAttribute_entry (attrName, var, cdfDatatype, 
                                             (long) attr.len, sPtr);
            free (sPtr);
            break;

        case NC_USHORT:
            usPtr = (unsigned short *) malloc (sizeof(short) * attr.len);
            if (usPtr == NULL) memory_error(5, "get_attribute", "NC_USHORT");
            status = nc_get_att_ushort(ncid, varId, attr.name, usPtr);
            if (status != NC_NOERR) handle_error(30, status);
            if (DEBUG) {
                printf (" = ");
                for (i=0; i < attr.len; i++) {
                     if (i > 0) printf (", ");
                     printf ("%ud", *(usPtr+i));
                }
            }
            if (cdfAttrScope == GLOBAL_SCOPE) {
                count = findIndex (attrName, &idx);
                if (count == 1)
                  create_cdf_gAttribute_entry (cdfAttrNum, idx, cdfDatatype,
                                               (long) attr.len, usPtr);
                else
                  create_cdf_gAttribute_entry (cdfAttrNum, 0, cdfDatatype, 
                                               (long) attr.len, usPtr);
            } else
                create_cdf_vAttribute_entry (attrName, var, cdfDatatype, 
                                             (long) attr.len, usPtr);
            free (usPtr);
            break;

        case NC_INT:
            iPtr = (int *) malloc (sizeof(int) * attr.len);
            if (iPtr == NULL) memory_error(6, "get_attribute", "NC_INT");
            status = nc_get_att_int(ncid, varId, attr.name, iPtr);
            if (status != NC_NOERR) handle_error(31, status);
            if (DEBUG) {
                printf (" = ");
                for (i=0; i < attr.len; i++) {
                     if (i > 0) printf (", ");
                     printf ("%d", *(iPtr+i));
                }
            }
            if (cdfAttrScope == GLOBAL_SCOPE) {
                count = findIndex (attrName, &idx);
                if (count == 1)
                  create_cdf_gAttribute_entry (cdfAttrNum, idx, cdfDatatype,
                                               (long) attr.len, iPtr);
                else
                  create_cdf_gAttribute_entry (cdfAttrNum, 0, cdfDatatype, 
                                               (long) attr.len, iPtr);
            } else
                create_cdf_vAttribute_entry (attrName, var, cdfDatatype, 
                                             (long) attr.len, iPtr);
            free (iPtr);
            break;

        case NC_UINT:
            uiPtr = (unsigned int *) malloc (sizeof(int) * attr.len);
            if (uiPtr == NULL) memory_error(7, "get_attribute", "NC_UINT");
            status = nc_get_att_uint(ncid, varId, attr.name, uiPtr);
            if (status != NC_NOERR) handle_error(32, status);
            if (DEBUG) {
                printf (" = ");
                for (i=0; i < attr.len; i++) {
                     if (i > 0) printf (", ");
                     printf ("%ud", *(uiPtr+i));
                }
            }
            if (cdfAttrScope == GLOBAL_SCOPE) {
                count = findIndex (attrName, &idx);
                if (count == 1)
                  create_cdf_gAttribute_entry (cdfAttrNum, idx, cdfDatatype,
                                               (long) attr.len, uiPtr);
                else
                  create_cdf_gAttribute_entry (cdfAttrNum, 0, cdfDatatype, 
                                               (long) attr.len, uiPtr);
            } else
                create_cdf_vAttribute_entry (attrName, var, cdfDatatype, 
                                             (long) attr.len, uiPtr);
            free (uiPtr);
            break;

        case NC_INT64:
            jPtr = (long long *) malloc (sizeof(long long) * attr.len);
            if (jPtr == NULL) memory_error(8, "get_attribute", "NC_INT64");
            status = nc_get_att_longlong(ncid, varId, attr.name, jPtr);
            if (status != NC_NOERR) handle_error(33, status);
            if (DEBUG) {
                printf (" = ");
                for (i=0; i < attr.len; i++) {
                     if (i > 0) printf (", ");
                     printf ("%lld", *(jPtr+i));
                }
            }
            if (cdfAttrScope == GLOBAL_SCOPE) {
                count = findIndex (attrName, &idx);
                if (count == 1)
                  create_cdf_gAttribute_entry (cdfAttrNum, idx, cdfDatatype,
                                               (long) attr.len, jPtr);
                else
                  create_cdf_gAttribute_entry (cdfAttrNum, 0, cdfDatatype, 
                                               (long) attr.len, jPtr);
            } else
                create_cdf_vAttribute_entry (attrName, var, cdfDatatype, 
                                             (long) attr.len, jPtr);
            free (jPtr);
            break;

        case NC_FLOAT:
            fPtr = (float *) malloc (sizeof(float) * attr.len);
            if (fPtr == NULL) memory_error(9, "get_attribute", "NC_FLOAT");
            status = nc_get_att_float(ncid, varId, attr.name, fPtr);
            if (status != NC_NOERR) handle_error(34, status);
            if (DEBUG) {
                printf (" = ");
                for (i=0; i < attr.len; i++) {
                     if (i > 0) printf (", ");
                     printf ("%g", *(fPtr+i));
                }
            }
            if (cdfAttrScope == GLOBAL_SCOPE) {
                count = findIndex (attrName, &idx);
                if (count == 1)
                  create_cdf_gAttribute_entry (cdfAttrNum, idx, cdfDatatype,
                                               (long) attr.len, fPtr);
                else
                  create_cdf_gAttribute_entry (cdfAttrNum, 0, cdfDatatype, 
                                               (long) attr.len, fPtr);
            } else
                create_cdf_vAttribute_entry (attrName, var, cdfDatatype, 
                                             (long) attr.len, fPtr);
            free (fPtr);
            break;

        case NC_DOUBLE:
            dPtr = (double *) malloc (sizeof(double) * attr.len);
            if (dPtr == NULL) memory_error(10, "get_attribute", "NC_DOUBLE");
            status = nc_get_att_double(ncid, varId, attr.name, dPtr);
            if (status != NC_NOERR) handle_error(35, status);
            if (DEBUG) {
                printf (" = ");
                for (i=0; i < attr.len; i++) {
                     if (i > 0) printf (", ");
                     printf ("%g", *(dPtr+i));
                }
            }
            if (cdfAttrScope == GLOBAL_SCOPE) {
                count = findIndex (attrName, &idx);
                if (count == 1)
                  create_cdf_gAttribute_entry (cdfAttrNum, idx, cdfDatatype,
                                               (long) attr.len, dPtr);
                else
                  create_cdf_gAttribute_entry (cdfAttrNum, 0, cdfDatatype, 
                                               (long) attr.len, dPtr);
            } else
                create_cdf_vAttribute_entry (attrName, var, cdfDatatype, 
                                             (long) attr.len, dPtr);
            free (dPtr);
            break;

        case NC_STRING: {
            int ilen;
            char **strings;
            strings = (char **)malloc (sizeof(char *) * attr.len );
            if (strings == NULL) memory_error(11, "get_attribute", "NC_STRING");
            if (cdfAttrScope == GLOBAL_SCOPE && setAttributeType == -1) {
              setAttributeType = NC_STRING;
            }
            status = nc_get_att_string(ncid, varId, attr.name, strings);
            if (status != NC_NOERR) handle_error(36, status);
            if (DEBUG) {
                printf (" = ");
                for (i=0; i < attr.len; i++) {
                     if (i > 0) printf (", ");
                     printf ("%s", strings[i]);
                }
            }
            if (cdfAttrScope == GLOBAL_SCOPE) {
                count = findIndex (attrName, &idx);
                for (i = 0; i < attr.len; ++i) {
                  if (strings[i] != NULL)
                    ilen = (int) strlen(strings[i]);
                  else
                    ilen = 0;
                  cPtr = (char *) malloc (sizeof(char) * ilen + 1);
                  if (cPtr == NULL) memory_error(12, "get_attribute",
                                                 "NC_STRING");
                  if (strings[i] != NULL) strcpy (cPtr, strings[i]);
                  if (count == 1)
                    create_cdf_gAttribute_entry (cdfAttrNum, idx, cdfDatatype,
                                                 (long) ilen, cPtr);
                  else
                    create_cdf_gAttribute_entry (cdfAttrNum, i, cdfDatatype, 
                                                 (long) ilen, cPtr);
                }
                free (cPtr);
            } else {
                char *newString;
                size_t tmpLen;
                ilen = 0;
                for (i = 0; i < attr.len; ++i) {
                  if (i > 0) ilen += strlen(STRINGDELIMITER);
                  if (strings[i] != NULL) {
                    tmpLen=strlen(strings[i]);
                    ilen += tmpLen;
                  }
                }
                newString=(char *) malloc(ilen+1);
                ilen = 0;
                for (i = 0; i < attr.len; ++i) {
                  if (i > 0) {
                    memcpy(newString+ilen, STRINGDELIMITER, 
                          strlen(STRINGDELIMITER));
                    ilen += strlen(STRINGDELIMITER);
                  }
                  if (strings[i] != NULL) {
                    tmpLen=strlen(strings[i]);
                    memcpy(newString+ilen, strings[i], tmpLen);
                    ilen += tmpLen;
                  }
                }
                *(newString+ilen) = '\0';
                create_cdf_vAttribute_entry (attrName, var, cdfDatatype, 
                                             (long) ilen, newString);
                free (newString);
            }
            if (attr.len > 0)
                nc_free_string (attr.len, strings);
/*            if (strings != NULL) free (strings); */
            break;
        }

        default:
            printf ("** Error in get_attribute: bad/unknown data type: %s",
                    NC_datatype (attr.type));
    }
    if (DEBUG) printf (" ;\n");
}


/*----------------------------------------------------------------------------
 *  This routine checks where a netcdf's variable is an ISTP-based epoch. If 
 *  so, its data type, while converted in CDF, will be set accordingly. 
 *---------------------------------------------------------------------------*/
int get_ISTPepoch(int ncid,             /* in - netCDF id            */
                  int varId)            /* in - netCDF variable ID   */
{
    int           status;
    long          cdfDatatype;
    
    char          *cPtr1 = NULL, *cPtr2 = NULL;
    nc_type       type1, type2;
    size_t        len1, len2;
    int           found;

    found = 0;
    status = nc_inq_att(ncid,                /* in - netCDF file ID  */
                        varId,               /* in - variable ID     */
                        "TIME_BASE",         /* in - attribute name  */
                        &type1,              /* out - Attribute type */
                        &len1);              /* out - # of attribute values */
                                             /*       or the length of string */
    if (status != NC_NOERR) return found;

    status = nc_inq_att(ncid,                /* in - netCDF file ID  */
                        varId,               /* in - variable ID     */
                        "UNITS",             /* in - attribute name  */
                        &type2,              /* out - Attribute type */
                        &len2);              /* out - # of attribute values */
                                             /*       or the length of string */
    if (status != NC_NOERR) return found;

    /***********************************************************/
    /* Map the netCDF datatype to an appropriate CDF datatype. */
    /***********************************************************/
    cdfDatatype = get_cdf_datatype (type1);
    if (cdfDatatype == UNKNOWN_DT) {
        printf ("*** ERROR: netCDF datatype:%d for attribute \"TIME_BASE\" is "
                "unsupported\n", type1);
        return found;
    }
    cdfDatatype = get_cdf_datatype (type2);
    if (cdfDatatype == UNKNOWN_DT) {
        printf ("*** ERROR: netCDF datatype:%d for attribute \"UNITS\" is "
                "unsupported\n", type2);
        return found;
    }

    /****************************************/
    /* Get the attribute value(s).          */
    /****************************************/
    cPtr1 = cPtr2 = NULL;
    switch (type1) {
        case NC_CHAR:
            cPtr1 = (char *) malloc(len1 * sizeof(char) + 1);
            if (cPtr1 == NULL) memory_error(1, "get_attribute", "NC_CHAR");
            status = nc_get_att_text(ncid, varId, "TIME_BASE", cPtr1);
            if (status != NC_NOERR) handle_error(37, status);
            *(cPtr1+len1) = '\0';               /* End of string mark */
            if (DEBUG) printf (" = \"%s\"", cPtr1);
            break;

        case NC_STRING: {
            char **strings = (char **)malloc (sizeof(char *) * 1 /* len */);
            if (strings == NULL) memory_error(11, "get_attribute", "NC_STRING");
            status = nc_get_att_string(ncid, varId, "TIME_BASE", strings);
            if (status != NC_NOERR) handle_error(38, status);
            cPtr1 = (char *) malloc (sizeof(char) * strlen(strings[0])+1);
            if (cPtr1 == NULL) memory_error(12, "get_attribute", "NC_STRING");
            strcpy (cPtr1, strings[0]);
            nc_free_string (1, strings);
/*          if (strings != NULL) free (strings); */
            break;
        }

        default:
           ;
    }
    switch (type2) {
        case NC_CHAR:
            cPtr2 = (char *) malloc(len2 * sizeof(char) + 1);
            if (cPtr2 == NULL) memory_error(2, "get_attribute", "NC_CHAR");
            status =nc_get_att_text(ncid, varId, "UNITS", cPtr2);
            if (status != NC_NOERR) handle_error(39, status);
            *(cPtr2+len2) = '\0';               /* End of string mark */
            if (DEBUG) printf (" = \"%s\"", cPtr2);
            break;

        case NC_STRING: {
            char **strings = (char **)malloc (sizeof(char *) * 1 /* len */);
            if (strings == NULL) memory_error(11, "get_attribute", "NC_STRING");
            status = nc_get_att_string(ncid, varId, "UNITS", strings);
            if (status != NC_NOERR) handle_error(40, status);
            cPtr2 = (char *) malloc (sizeof(char) * strlen(strings[0])+1);
            if (cPtr2 == NULL) memory_error(12, "get_attribute", "NC_STRING");
            strcpy (cPtr2, strings[0]);
            nc_free_string (1, strings);
/*          if (strings != NULL) free (strings); */
            break;
        }

        default:
           ;
    }
    if (cPtr1 != NULL && StrStrIgCaseX(cPtr1, "0 ad")) {
      if (cPtr2 != NULL && (StrStrIgCaseX(cPtr2, "ms") ||
                            StrStrIgCaseX(cPtr2, "milliseconds"))) {
        found = 1;
      } else if (cPtr2 != NULL && (StrStrIgCaseX(cPtr2, "ps") ||
                                   StrStrIgCaseX(cPtr2, "picoseconds"))) {
        found = 2;
      }
    } else if(cPtr1 != NULL && StrStrIgCaseX(cPtr1, "j2000")) {
      if (cPtr2 != NULL && (StrStrIgCaseX(cPtr2, "ns") ||
                            StrStrIgCaseX(cPtr2, "nanoseconds"))) {
        found = 3;
      }
    }
    if (cPtr1 != NULL) free (cPtr1);
    if (cPtr2 != NULL) free (cPtr2);
    if (DEBUG) printf (" ;\n");
    return found;
}


/*--------------------------------------------------------------------------
 *  Get the netCDF variable data and put that value into the CDF variable
 *  created in create_cdf_var.
 *--------------------------------------------------------------------------*/
void getVariableData (int ncid,
                      struct ncvar *var,      /* in - netCDF var structure  */
                      int varId,              /* in - netCDF var ID         */
                      struct cdfvar *cdfVar)  /* in/out - CDF var structure */
{
    size_t start[NC_MAX_DIMS],        /* index for where to read first */ 
           count[NC_MAX_DIMS];        /* # of values to read           */

    int     i, j, divide, status, 
                  numValues;                /* # of values on each record */

    char          *cPtr;
    signed char   *scPtr;
    unsigned char *ucPtr;
    short         *sPtr;
    unsigned short *usPtr;
    int           *iPtr;
    unsigned int  *uiPtr;
    long long     *jPtr;
    float         *fPtr;
    double        *dPtr;
    long          recNo, 
                  allRecs,                  /* total # of records to read */
                  numRecsLeftToRead,
                  numRecs;       /* # of records to read and write at a time */
    char epString[EPOCH_STRING_LEN+1];
    char ep16String[EPOCH16_STRING_LEN+1];
    char tt2000String[TT2000_3_STRING_LEN+1];

    if (var->recordVariable && var->dimSizes[0] == 0) return; /* No records */
    numValues = cdfVar->numValues;          /* # of values on each record */
    for (i=0; i <= var->ndims; i++) {
         start[i] = 0;
         count[i] = var->dimSizes[i];
    }

    if (var->recordVariable) {
        allRecs = var->dimSizes[0];    /* Get the total # of records to read */
        count[0] = allRecs;            /* Try to read all records at once    */
    }
    else { 
        allRecs  = 1;              /* total # of records to read and write */
    }

    recNo = 0;                      /* Record number to which write data */
    numRecsLeftToRead = allRecs;
    while (numRecsLeftToRead > 0) {
        recNo   = start[0];
        divide  = 1;
        numRecs = numRecsLeftToRead;     /* # of records to read */
    
        switch (var->type) {

        case NC_CHAR:
            cPtr = NULL;
            while (cPtr == NULL) {
               cPtr = (char *) malloc (sizeof(char) * numValues * numRecs);
               if (cPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = allRecs / divide;
               }
            }
            if (var->recordVariable) 
                count[0] = numRecs;
            status = nc_get_vara_text(ncid, varId, start, count, cPtr);
            if (status != NC_NOERR) handle_error(41, status);
            if (cdfVar->datatype == CDF_EPOCH) {
              dPtr = (double *) malloc (sizeof(double) *
                                        cdfVar->numElements * numRecs);
              if (dPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   memory_error(13, "Get var data", "NC_CHAR->EPOCH");
              } 
              for (i = 0; i < cdfVar->numElements * numRecs; ++i) {
                memcpy (epString, cPtr+i*EPOCH_STRING_LEN, EPOCH_STRING_LEN);
                *(epString+EPOCH_STRING_LEN) = '\0';
                *(dPtr+i) = parseEPOCH (epString);
              }
              write_cdf_vardata (cdfVar->num, recNo, numRecs,
                                 cdfVar->dimensionality, cdfVar->dimSizes,
                                 dPtr);
              free (dPtr);
            } else if (cdfVar->datatype == CDF_EPOCH16) {
              double tmp;
              dPtr = (double *) malloc (sizeof(double) * 2 *
                                        cdfVar->numElements * numRecs);
              if (dPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   memory_error(14, "Get var data", "NC_CHAR->EPOCH16");
              } 
              for (i = 0; i < cdfVar->numElements * numRecs; ++i)
              {
                char mmm[EPOCH16_STRING_LEN+1];
                memcpy(ep16String, cPtr+i*EPOCH16_STRING_LEN,
                       EPOCH16_STRING_LEN);
                ep16String[EPOCH16_STRING_LEN] = (char) '\0';
                tmp = parseEPOCH16 (ep16String, dPtr+2*i);
              }
              write_cdf_vardata (cdfVar->num, recNo, numRecs,
                                 cdfVar->dimensionality, cdfVar->dimSizes,
                                 dPtr);
              free (dPtr);
            } else if (cdfVar->datatype == CDF_TIME_TT2000) {
              jPtr = (long long *) malloc (sizeof(long long) *
                                           cdfVar->numElements * numRecs);
              if (jPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   memory_error(15, "Get var data", "NC_CHAR->TT2000");
              } 
              for (i = 0; i < cdfVar->numElements * numRecs; ++i) {
                memcpy(tt2000String, cPtr+i*TT2000_3_STRING_LEN,
                       TT2000_3_STRING_LEN);
                tt2000String[TT2000_3_STRING_LEN] = (char) '\0';
                *(jPtr+i) = parseTT2000 (tt2000String);
              }
              write_cdf_vardata (cdfVar->num, recNo, numRecs,
                                 cdfVar->dimensionality, cdfVar->dimSizes,
                                 jPtr);
              free (jPtr);
            } else
              write_cdf_vardata (cdfVar->num,
                                 recNo,
                                 numRecs,            /* # of records to write */
                                 cdfVar->dimensionality,
                                 cdfVar->dimSizes,   /* dimension sizes */
                                 cPtr);              /* data to be written */
            free (cPtr);
            break;

        case NC_BYTE:
            scPtr = NULL;
            while (scPtr == NULL) {
               scPtr = (signed char *) malloc (sizeof(signed char)*numValues*
                                               numRecs);
               if (scPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = allRecs / divide;
               }
            }
            if (var->recordVariable) 
                count[0] = numRecs;
            status = nc_get_vara_schar(ncid, varId, start, count, scPtr);
            if (status != NC_NOERR) handle_error(42, status);
/* NC_FILL_BYTE == DEFAULT_INT1_PADVALUE
            if (fillvalue) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (*((signed char *)(scPtr)+i) == NC_FILL_BYTE)
                  *((signed char *)(scPtr)+i) =
                                       (signed char) DEFAULT_INT1_PADVALUE;
              }
            }
*/
            write_cdf_vardata (cdfVar->num,
                               recNo,
                               numRecs,          /* # of records to write */
                               cdfVar->dimensionality,
                               cdfVar->dimSizes,   /* dimension sizes */
                               scPtr);             /* data to be written */
            free (scPtr);
            break;

        case NC_UBYTE:
            ucPtr = NULL;
            while (ucPtr == NULL) {
               ucPtr = (unsigned char *) malloc (sizeof(signed char)*numValues*
                                                 numRecs);
               if (ucPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = allRecs / divide;
               }
            }
            if (var->recordVariable) 
                count[0] = numRecs;
            status = nc_get_vara_uchar(ncid, varId, start, count, ucPtr);
            if (status != NC_NOERR) handle_error(43, status);
            if (fillvalue) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (*((unsigned char *)(ucPtr)+i) == NC_FILL_UBYTE)
                  *((unsigned char *)(ucPtr)+i) =
                                      (unsigned char) DEFAULT_UINT1_PADVALUE;
              }
            }
            write_cdf_vardata (cdfVar->num,
                               recNo,
                               numRecs,          /* # of records to write */
                               cdfVar->dimensionality,
                               cdfVar->dimSizes,   /* dimension sizes */
                               ucPtr);             /* data to be written */
            free (ucPtr);
            break;

        case NC_SHORT:
            sPtr = NULL;
            while (sPtr == NULL) {
               sPtr = (short *) malloc (sizeof(short) * numValues * numRecs);
               if (sPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = allRecs / divide;
               }
            }
            if (var->recordVariable) 
                count[0] = numRecs;
            status = nc_get_vara_short(ncid, varId, start, count, sPtr);
            if (status != NC_NOERR) handle_error(44, status);
/* NC_FILL_SHORT == DEFAULT_INT2_PADVALUE
            if (fillvalue) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (*((short *)(sPtr)+i) == NC_FILL_SHORT)
                  *((short *)(sPtr)+i) = (short) DEFAULT_INT2_PADVALUE;
              }
            }
*/
            write_cdf_vardata (cdfVar->num,
                               recNo,  
                               numRecs,            /* # of records to write */
                               cdfVar->dimensionality,
                               cdfVar->dimSizes,   /* dimension sizes */
                               sPtr);              /* data to be written */
            free (sPtr);
            break;

        case NC_USHORT:
            usPtr = NULL;
            while (usPtr == NULL) {
               usPtr = (unsigned short *) malloc (sizeof(short) * numValues *
                                                  numRecs);
               if (usPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = allRecs / divide;
               }
            }
            if (var->recordVariable) 
                count[0] = numRecs;
            status = nc_get_vara_ushort(ncid, varId, start, count, usPtr);
            if (status != NC_NOERR) handle_error(45, status);
            if (fillvalue) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (*((unsigned short *)(usPtr)+i) == NC_FILL_USHORT)
                  *((unsigned short *)(usPtr)+i) =
                                      (unsigned short) DEFAULT_UINT2_PADVALUE;
              }
            }
            write_cdf_vardata (cdfVar->num,
                               recNo,  
                               numRecs,            /* # of records to write */
                               cdfVar->dimensionality,
                               cdfVar->dimSizes,   /* dimension sizes */
                               usPtr);              /* data to be written */
            free (usPtr);
            break;

        case NC_INT:
            iPtr = NULL;
            while (iPtr == NULL) {
               iPtr = (int *) malloc (sizeof(int) * numValues * numRecs);
               if (iPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = allRecs / divide;
               }
            }
            if (var->recordVariable) 
                count[0] = numRecs;
            status = nc_get_vara_int(ncid, varId, start, count, iPtr);
            if (status != NC_NOERR) handle_error(46, status);
/* NC_FILL_INT == DEFAULT_INT4_PADVALUE
            if (fillvalue) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (*((int *)(iPtr)+i) == NC_FILL_INT)
                  *((int *)(iPtr)+i) = (int) DEFAULT_INT4_PADVALUE;
              }
            }
*/
            write_cdf_vardata (cdfVar->num,
                               recNo,
                               numRecs,            /* # of records to write */
                               cdfVar->dimensionality,
                               cdfVar->dimSizes,   /* dimension sizes */
                               iPtr);              /* data to be written */
            free (iPtr);
            break;

        case NC_UINT:
            uiPtr = NULL;
            while (uiPtr == NULL) {
               uiPtr = (unsigned int *) malloc (sizeof(int) * numValues *
                                                numRecs);
               if (uiPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = allRecs / divide;
               }
            }
            if (var->recordVariable) 
                count[0] = numRecs;
            status = nc_get_vara_uint(ncid, varId, start, count, uiPtr);
            if (status != NC_NOERR) handle_error(47, status);
            if (fillvalue) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (*((unsigned int *)(uiPtr)+i) == NC_FILL_UINT)
                  *((unsigned int *)(uiPtr)+i) =
                                        (unsigned int) DEFAULT_UINT4_PADVALUE;
              }
            }
            write_cdf_vardata (cdfVar->num,
                               recNo,
                               numRecs,            /* # of records to write */
                               cdfVar->dimensionality,
                               cdfVar->dimSizes,   /* dimension sizes */
                               uiPtr);             /* data to be written */
            free (uiPtr);
            break;

        case NC_INT64:
            jPtr = NULL;
            while (jPtr == NULL) {
               jPtr = (long long *) malloc (sizeof(long long) * numValues *
                                            numRecs);
               if (jPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = allRecs / divide;
               }
            }
            if (var->recordVariable) 
                count[0] = numRecs;
            status = nc_get_vara_longlong(ncid, varId, start, count, jPtr);
            if (status != NC_NOERR) handle_error(48, status);
            if (fillvalue) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (*((long long *)(jPtr)+i) == NC_FILL_INT64) {
                  if (cdfVar->datatype == CDF_INT8) 
                    *((long long *)(jPtr)+i) =
                                       (long long) DEFAULT_INT8_PADVALUE;
                  else
                    *((long long *)(jPtr)+i) =
                                       (long long) DEFAULT_TT2000_PADVALUE;
                }
              }
            }
            write_cdf_vardata (cdfVar->num,
                               recNo,
                               numRecs,            /* # of records to write */
                               cdfVar->dimensionality,
                               cdfVar->dimSizes,       /* dimension sizes */
                               jPtr);                  /* data to be written */
            free (jPtr);
            break;

        case NC_FLOAT:
            fPtr = NULL;
            while (fPtr == NULL) {
               fPtr = (float *) malloc (sizeof(float) * numValues * numRecs);
               if (fPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = allRecs / divide;
               }
            }
            if (var->recordVariable) 
                count[0] = numRecs;
            status = nc_get_vara_float(ncid, varId, start, count, fPtr);
            if (status != NC_NOERR) handle_error(49, status);
            if (fillvalue) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (*((float *)(fPtr)+i) == NC_FILL_FLOAT)
                  *((float *)(fPtr)+i) = (float) DEFAULT_FLOAT_PADVALUE;
              }
            }
            write_cdf_vardata (cdfVar->num,
                               recNo,
                               numRecs,            /* # of records to write */
                               cdfVar->dimensionality,
                               cdfVar->dimSizes,   /* dimension sizes */
                               fPtr);              /* data to be written */
            free (fPtr);
            break;

        case NC_DOUBLE:
            dPtr = NULL;
            while (dPtr == NULL) {
               dPtr = (double *) malloc (sizeof(double) * numValues * numRecs);
               if (dPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = allRecs / divide;
               }
            }
            if (var->recordVariable) 
                count[0] = numRecs;
            status = nc_get_vara_double(ncid, varId, start, count, dPtr);
            if (status != NC_NOERR) handle_error(50, status);
            if (fillvalue) {
              if (cdfVar->datatype == CDF_DOUBLE ||
                  cdfVar->datatype == CDF_EPOCH) {
                for (i = 0; i < numValues*numRecs; ++i) {
                  if (*((double *)(dPtr)+i) == NC_FILL_DOUBLE) {
                    if (cdfVar->datatype == CDF_DOUBLE) 
                      *((double *)(dPtr)+i) = (double) DEFAULT_DOUBLE_PADVALUE;
                    else
                      *((double *)(dPtr)+i) = (double) DEFAULT_EPOCH_PADVALUE;
                  }
                }
              } else { /* CDF_EPOCH16 */
                for (i = 0; i < numValues*numRecs; i=i+2) {
                  if (*((double *)(dPtr)+i) == NC_FILL_DOUBLE)
                    *((double *)(dPtr)+i) = (double) DEFAULT_EPOCH16_PADVALUE;
                  if (*((double *)(dPtr)+i+1) == NC_FILL_DOUBLE)
                    *((double *)(dPtr)+i+1) = (double) DEFAULT_EPOCH16_PADVALUE;
                }
              }
            }
            write_cdf_vardata (cdfVar->num,
                               recNo,
                               numRecs,            /* # of records to write */
                               cdfVar->dimensionality,
                               cdfVar->dimSizes,   /* dimension sizes */
                               dPtr);              /* data to be written */
            free (dPtr);
            break;

        case NC_STRING: {
            char **strings;
            int recs, values;
            cPtr = NULL;
            if (var->recordVariable) {
                count[0] = numRecs;
                recs = numRecs;
            } else
                recs = 1;
            values = 1;
            for (i = 0; i < cdfVar->dimensionality; ++i)
              values *= cdfVar->dimSizes[i];
            if (cPtr == NULL) {
               for (i = 0; i < numTextVars; ++i) {
                 if (strcmp (cdfVar->name, textVars[i]) == 0) {
                   cPtr = (char *) malloc (sizeof(char)*(varLen[i]+1)*values*
                                                         recs);
                   if (cPtr == NULL) {
                       if (DEBUG) insufficient_memory();
                   }
                   break;
                 }
               }
            }
            strings = (char **) malloc (sizeof (char *) * recs * values);
            status = nc_get_vara_string(ncid, varId, start, count, strings);
            if (status != NC_NOERR) handle_error(51, status);
            if (strings[0] != NULL && strlen(strings[0]) > 0) {
              for (i = 0; i < recs; ++i) {
                for (j = 0; j < values; ++j)
                  memcpy (cPtr+j*cdfVar->numElements,
                          strings[i*values+j],
                          cdfVar->numElements);
                write_cdf_vardata (cdfVar->num,
                                   recNo+i,
                                   1,            /* # of records to write */
                                   cdfVar->dimensionality,
                                   cdfVar->dimSizes,   /* dimension sizes */
                                   (void *) cPtr);     /* data to be written */
              }
              free (cPtr);
              nc_free_string (recs*values, strings);
/*            if (strings != NULL) free (strings); */
           }
/*
           } else
              if (strings != NULL) free (strings);
*/
            break;
        }

        default:
            printf ("** Error in getVarData: bad/unknown type: %s\n",
                    NC_datatype (var->type));
        }    

        numRecsLeftToRead = allRecs - numRecs;
        if (numRecsLeftToRead > 0) {
            allRecs  = numRecsLeftToRead;
            start[0] = numRecs;
        }
    }       /* end of the 'while' loop */
}


/*--------------------------------------------------------------------------*/
char *  NC_datatype (nc_type type)
{
    switch (type) {
      case NC_BYTE:
        return "NC_BYTE";
      case NC_UBYTE:
        return "NC_UBYTE";
      case NC_CHAR:
        return "NC_CHAR";
      case NC_SHORT:
        return "NC_SHORT";
      case NC_USHORT:
        return "NC_USHORT";
      case NC_INT:
        return "NC_INT";
      case NC_UINT:
        return "NC_UINT";
      case NC_INT64:
        return "NC_INT64";
      case NC_FLOAT:
        return "NC_FLOAT";
      case NC_DOUBLE:
        return "NC_DOUBLE";
      case NC_STRING:
        return "NC_STRING";
      case NC_UINT64:
        return "NC_UINT64";
      case NC_COMPOUND:
        return "NC_COMPOUND";
      case NC_FIRSTUSERTYPEID:
        return "NC_FIRSTUSERTYPEID";
      default:
        return "UNKNOWN";
    }
}


/*--------------------------------------------------------------------------*/
long get_cdf_datatype (nc_type type)
{
    switch (type) {
      case NC_BYTE:
        return CDF_INT1;

      case NC_UBYTE:
        return CDF_UINT1;

      case NC_CHAR:
      case NC_STRING:
        return CDF_CHAR;

      case NC_SHORT:
        return CDF_INT2;

      case NC_USHORT:
        return CDF_UINT2;

      case NC_INT:
        return CDF_INT4;

      case NC_UINT:
        return CDF_UINT4;

      case NC_INT64:
        return CDF_INT8;

      case NC_FLOAT:
        return CDF_FLOAT;

      case NC_DOUBLE:
        return CDF_DOUBLE;

      default:
        return UNKNOWN_DT;     /* Unknown netCDF datatype */
    }
}


/*------------------------------------------------------------------
 *   This routine creates a CDF file.
 *------------------------------------------------------------------*/
void create_cdf_file (char *fileName)
{
    CDFstatus status;

    int   i;
    long  dim,                          /* Array dimensionality      */
          dimSizes[CDF_MAX_DIMS],       /* Array dimension sizes     */
          dimVarys[CDF_MAX_DIMS];       /* Array dimension variances */

    /********************************************************************/
    /*  The following variables need some dummy values for the creation */
    /*  of a CDF file and CDF variables.                                */
    /********************************************************************/
    dim = 0L;
    for (i=0; i < CDF_MAX_DIMS; i++) {
        dimSizes[i] = 1;
        dimVarys[i] = VARY;
    }

    if (DELETE) {
      status = CDFlib (OPEN_, CDF_, fileName, &id,
                       NULL_);
      if (status == CDF_OK) {
        status = CDFlib (DELETE_, CDF_,
                         NULL_);
        cdf_status_handler (status, "DELETE_, CDF_");
      }
    }
    /* The CDF library adds .cdf at the end of the file name */
    status = CDFlib (CREATE_, CDF_, fileName, dim, dimSizes, &id,
                     PUT_, CDF_MAJORITY_, (COLUMN?COLUMN_MAJOR:ROW_MAJOR),
/*                           CDF_CHECKSUM_, MD5_CHECKSUM, */
                     NULL_);
    if (status != CDF_OK) cdf_status_handler (status, "CREATE_, CDF_");
}


/*--------------------------------------------------------------------------*/
/* Create a CDF zVariable.                                                  */
/*--------------------------------------------------------------------------*/
long create_cdf_var (int ncid,                /* in - netcdf id */
                     char *group,             /* in - group name */
                     int varId,               /* in - netCDF var ID        */
                     struct ncvar var,        /* in - netCDF var structure */
                     struct cdfvar *cdfVar)   /* out - CDF var structure   */
{
   CDFstatus status;
   int       i;
   long      numElements,
             recVariance,                  /* Record variance */
             dimVariance[CDF_MAX_DIMS],    /* Dimension variance */
             varNum;                       /* zVariable number */
   long cParms[] = {0};

   cdfVar->numValues = 1;
  
   /***********************************************************************/
   /* If the netCDF variable is a record variable, the first dimension of */
   /* the variable is a record number and it needs to be skipped.         */
   /***********************************************************************/
   if (var.recordVariable) {
       recVariance = VARY;             
       if (var.ndims > 0)       /* variable is not scalar */
           cdfVar->dimensionality = var.ndims - 1;
   }
   else {
       recVariance    = NOVARY;            
       cdfVar->dimensionality = var.ndims;
   }

   for (i=0; i < var.ndims; i++) {
        if (var.recordVariable) {
            if (i != 0) {      /* Ignore the record number */
                cdfVar->numValues *= var.dimSizes[i];
                cdfVar->dimSizes[i-1] = var.dimSizes[i];
                dimVariance[i-1] = VARY;
            }
        }
        else {
            dimVariance[i] = VARY;
            cdfVar->numValues *= var.dimSizes[i];
            cdfVar->dimSizes[i] = var.dimSizes[i];
        }
   }

   cdfVar->datatype = get_cdf_datatype(var.type);
   if (cdfVar->datatype == UNKNOWN_DT) {
       if (FORCE) 
         printf ("*** warning: netCDF datatype: %s for variable \"%s\" "
                 "in group:\"%s\""
                 " is not supported... skipped\n", NC_datatype(var.type),
                 var.ncname, var.group);
       else
         printf ("*** ERROR: netCDF datatype: %s for variable \"%s\" in "
                 "group:\"%s\""
                 " is not supported... \n", NC_datatype(var.type),
                 var.ncname, var.group);
       return (long) UNKNOWN_DT;
   }

   /**************************************************************/
   /*  In CDF, a string of 200 characters should be defined as   */
   /*    CHAR/200 0:[],  not CHAR/1 1:[200]                      */
   /*                                                            */
   /*  A string is treated as 1-D in netCDF while it is treated  */
   /*  as 0-D in CDF.                                            */
   /**************************************************************/
   if (var.type == NC_CHAR) {
     if (EPOCH) {
       if (!FindEpoch (var.name)) {
         if (var.ndims > 1) {
           cdfVar->numElements = var.dimSizes[var.ndims-1]; /* String length */
           cdfVar->dimensionality--;
         }
       } else {
         if (var.ndims > 1) {
           if (var.dimSizes[1] == EPOCH_STRING_LEN)
             cdfVar->datatype = CDF_EPOCH;
           else if (var.dimSizes[1] == EPOCH16_STRING_LEN)
             cdfVar->datatype = CDF_EPOCH16;
           else if (var.dimSizes[1] == TT2000_3_STRING_LEN)
             cdfVar->datatype = CDF_TIME_TT2000;
         } else {
           if (var.dimSizes[0] == EPOCH_STRING_LEN)
             cdfVar->datatype = CDF_EPOCH;
           else if (var.dimSizes[0] == EPOCH16_STRING_LEN)
             cdfVar->datatype = CDF_EPOCH16;
           else if (var.dimSizes[0] == TT2000_3_STRING_LEN)
             cdfVar->datatype = CDF_TIME_TT2000;
         }
         cdfVar->numElements = 1L;
         cdfVar->dimensionality--;
       }
     } else if (ISTPepoch) {
        int found = get_ISTPepoch (ncid, varId);
        if (found > 0) {
          if (found == 1) cdfVar->datatype = CDF_EPOCH;
          if (found == 2) {
            cdfVar->datatype = CDF_EPOCH16;
          }
          if (found == 3) cdfVar->datatype = CDF_TIME_TT2000;
          cdfVar->numElements = 1L;
          cdfVar->dimensionality--;
        } else {
          cdfVar->numElements = var.dimSizes[var.ndims-1];
          cdfVar->dimensionality--;
        }
     } else {
           cdfVar->numElements = var.dimSizes[var.ndims-1]; /* String length */
           cdfVar->dimensionality--;
     }
   }
   else
       cdfVar->numElements = 1L;

   strcpy (cdfVar->name, var.name);

   if (var.type == NC_STRING) {
      for (i = 0; i < numTextVars; ++i) {
        if (strcmp (cdfVar->name, textVars[i]) == 0) {
          cdfVar->numElements = varLen[i];
          break;
        }
      }
   }
   if ((EPOCH && FindEpoch (cdfVar->name)) || 
       (ISTPepoch && (get_ISTPepoch (ncid, varId) > 0))) {
     if (cdfVar->datatype == CDF_DOUBLE) {
       if (cdfVar->dimensionality == 0 ||
           cdfVar->dimSizes[cdfVar->dimensionality-1] != 2)
         cdfVar->datatype = CDF_EPOCH;
       else {
         cdfVar->datatype = CDF_EPOCH16;
         cdfVar->dimensionality = cdfVar->dimensionality - 1;
       }
     } else if (cdfVar->datatype == CDF_INT8) {
         cdfVar->datatype = CDF_TIME_TT2000;
     } else if ((cdfVar->datatype == CDF_CHAR || cdfVar->datatype == CDF_UCHAR)
                && cdfVar->dimensionality > 0) {
         if (var.dimSizes[cdfVar->dimensionality] == EPOCH_STRING_LEN) {
           cdfVar->dimensionality = cdfVar->dimensionality - 1;
           cdfVar->datatype = CDF_EPOCH;
         } else if (var.dimSizes[cdfVar->dimensionality] ==
                    EPOCH16_STRING_LEN) {
           cdfVar->datatype = CDF_EPOCH16;
           cdfVar->dimensionality = cdfVar->dimensionality - 1;
         } else if (var.dimSizes[cdfVar->dimensionality] ==
                    TT2000_3_STRING_LEN) {
           cdfVar->datatype = CDF_TIME_TT2000;
           cdfVar->dimensionality = cdfVar->dimensionality - 1;
         }
         if (cdfVar->dimensionality == 0) cdfVar->numElements = 1;
     }
   }
   if (cdfVar->datatype == CDF_CHAR && cdfVar->numElements == 0) {
     cdfVar->numElements = getFormatAttributeLength (ncid, varId);
   }
   if (cdfVar->numElements == -1) {
     char **strings; char name[NC_MAX_NAME+1];
     int ii, ncStatus; size_t start = 0, len;
     if (var.ndims > 0) {
       ncStatus = nc_inq_dim(ncid, var.dimIds[0], name, &len);
       if (len > 0) {
         strings = (char **) malloc (sizeof(char *) * len);
         ncStatus = nc_get_vara_string(ncid, varId, &start, &len, strings);
         for (ii=0; ii<(int)len;++ii) {
            if (strlen(strings[ii]) > cdfVar->numElements)
              cdfVar->numElements = (long) strlen(strings[ii]);
         }
         nc_free_string(len, strings);
       } else
         cdfVar->numElements = 100;
     } else
       cdfVar->numElements = 100;
   }
   status = CDFlib (CREATE_, zVAR_, cdfVar->name, 
                                    cdfVar->datatype, 
                                    cdfVar->numElements,
                                    cdfVar->dimensionality,
                                    cdfVar->dimSizes,
                                    recVariance,
                                    dimVariance,
                                    &varNum,
                    NULL_);
   if (status != CDF_OK) cdf_status_handler (status, "CREATE_, zVAR_");
   if (DEBUG) {
       printf("            Creating var: %s type=%ld nelems=%ld "
              "dimensionality=%ld recvary=%ld\n",cdfVar->name,cdfVar->datatype,
              cdfVar->numElements,cdfVar->dimensionality,recVariance);
       if (cdfVar->dimensionality> 0) {
         int iyy;
         printf("dimension: ");
         for (iyy=0; iyy < cdfVar->dimensionality; ++iyy)
           printf("%ld ", cdfVar->dimSizes[iyy]);
         printf("\n");
       }
   }
/*   if (compression && cdfVar->datatype != CDF_EPOCH &&
       cdfVar->datatype != CDF_EPOCH && cdfVar->datatype != CDF_TIME_TT2000) { */
   if (compression) {
     int size = 1; long dtSize;
     for (i = 0; i < cdfVar->dimensionality; ++i)
        if (dimVariance[i] == VARY) size *= (int) cdfVar->dimSizes[i];
     if (var.recordVariable) size *= var.dimSizes[0];
     status = CDFgetDataTypeSize(cdfVar->datatype, &dtSize);
     size *= cdfVar->numElements;
     size *=  (int) dtSize;
     if (dLevel != -1) cParms[0] = (long) dLevel;
     else if (var.deflate != 0 && var.dLevel > 0)
        cParms[0] = (long) var.dLevel;
     else if (size >= 1024) cParms[0] = 6L;
     if (cParms[0] > 0) {
       if ((cdfVar->datatype != CDF_EPOCH && cdfVar->datatype != CDF_EPOCH &&
            cdfVar->datatype != CDF_TIME_TT2000) ||
           (compressionE && (cdfVar->datatype == CDF_EPOCH ||
                             cdfVar->datatype == CDF_EPOCH ||
                             cdfVar->datatype == CDF_TIME_TT2000))) {
         status = CDFlib (PUT_, zVAR_COMPRESSION_, GZIP_COMPRESSION,
                                                   cParms,
                          NULL_);
         if (status != CDF_OK) cdf_status_handler (status, "PUT_, COMPRESSION");
       }
/*
       if (var.recordVariable) {
         status = CDFlib (PUT_, zVAR_BLOCKINGFACTOR_, (long) var.dimSizes[0],
                          NULL_);
         if (status != CDF_OK)
                cdf_status_handler (status, "PUT_, BLOCKINGFACTOR");
       }
*/
     } else {
       if (var.recordVariable && var.dimSizes[0] != 0) {
         status = CDFlib (PUT_, zVAR_ALLOCATERECS_, (long) var.dimSizes[0],
                          NULL_);
         if (status != CDF_OK) cdf_status_handler (status, "PUT_, ALLOCATERECS");
       }
       if (!var.recordVariable) {
         status = CDFlib (PUT_, zVAR_BLOCKINGFACTOR_, (long) 1,
                          NULL_);
         if (status != CDF_OK)
                cdf_status_handler (status, "PUT_, BLOCKINGFACTOR");
       }
     }
   } else {
     if (var.recordVariable && var.dimSizes[0] != 0) {
       status = CDFlib (PUT_, zVAR_ALLOCATERECS_, (long) var.dimSizes[0],
                        NULL_);
       if (status != CDF_OK) cdf_status_handler (status, "PUT_, ALLOCATERECS");
     }
     if (!var.recordVariable) {
         status = CDFlib (PUT_, zVAR_BLOCKINGFACTOR_, (long) 1,
                          NULL_);
         if (status != CDF_OK)
                cdf_status_handler (status, "PUT_, BLOCKINGFACTOR");
     }
   }

   return varNum;
}


/*---------------------------------------------------------------------------
 *  Write the netCDF variable data to the given CDF variable number.
 *---------------------------------------------------------------------------*/
void write_cdf_vardata (long varNum,     /* CDF variable number */ 
                        long recNo,      /* Record # from which data is added */
                        long recCount,   /* # of records to be written */
                        long dim,        /* CDF dimensionality */
                        long *dimSizes,  /* CDF dimension sizes */
                        void *data)      /* data to be written */
{
   CDFstatus status;
   int       i;
   long      recInterval,                /* No. of records to skip           */
             dimIndices[CDF_MAX_DIMS],   /* Beginning location of the array  */
                                         /* to be dumped                     */
             dimIntervals[CDF_MAX_DIMS]; /* No. of elements to skip          */

   recInterval = 1L;             /* # of records to skip between writes   */

   for (i=0; i < dim; i++) {
        dimIndices[i]   = 0;     /* Dump data from the beginning of array */
        dimIntervals[i] = 1;     /* Dump all the elements in the array    */
   }
   status = CDFlib (SELECT_,zVAR_, varNum,
                            zVAR_RECNUMBER_, recNo,
                            zVAR_RECCOUNT_, recCount,
                            zVAR_RECINTERVAL_, recInterval,
                            zVAR_DIMINDICES_, dimIndices,
                            zVAR_DIMCOUNTS_, dimSizes,
                            zVAR_DIMINTERVALS_, dimIntervals,
                    NULL_);
   if (status != CDF_OK) cdf_status_handler (status, "PUT_, zVAR_HYPERDATA_");
   if (!COLUMN || dim < 2) {
     status = CDFlib (PUT_, zVAR_HYPERDATA_, data,
                      NULL_);
   } else {
     long dataType, numElms, size, recSize = 1;
     void *tmpData;
     for (i=0; i < dim; i++) recSize *= dimSizes[i];
     status = CDFlib (GET_,   zVAR_DATATYPE_, &dataType,
                              zVAR_NUMELEMS_, &numElms,
                      NULL_);
     CDFgetDataTypeSize(dataType,&size);
     recSize *= size;
     recSize *= numElms;
     tmpData = (void *) malloc ((size_t) (recSize * recCount));
     if (tmpData == NULL) memory_error(16, "get_data", "row->col");;
     for (i = 0; i < (int) recCount; ++i) {
       ROWtoCOL ((void *) (((char *)data)+(int)recSize*i),
                 (void *) (((char *)tmpData)+(int)recSize*i), dim, dimSizes,
                 size*numElms); 
     }
     status = CDFlib (PUT_, zVAR_HYPERDATA_, tmpData,
                      NULL_);
     free (tmpData);
   }
   if (status != CDF_OK) cdf_status_handler (status, "PUT_, zVAR_HYPERDATA_");
}


/*--------------------------------------------------------------------------
 *  This routine creates a CDF global or variable attribute.
 *-------------------------------------------------------------------------*/
long create_cdf_attribute (char *attrName,   /* Attribute name to be created */
                           long attrScope,   /* Attribute scope */
                           int *created)     /* Indicator whether an attribute
                                                is really created */
{                                        
   CDFstatus status;
   long      attrNum;
   char      msg[80], *ptr;
   char      gAttr[CDF_ATTR_NAME_LEN+1];
   int       count, idx;

   ptr = strrchr (attrName, (char ) *delimiters[curDiD]);
   count = 0;
   if (ptr != NULL) count = sscanf (ptr+1,"%d",&idx);
   if (attrScope == GLOBAL_SCOPE && count == 1) {
     strncpy(gAttr, attrName, ptr-attrName);
     gAttr[ptr-attrName] = (char) '\0';
     status = CDFlib (CREATE_, ATTR_, gAttr,       /* in - attribute name    */
                                      attrScope,   /* in - attribute scope   */
                                      &attrNum,    /* out - attribute number */
                      NULL_);
     if (status != CDF_OK && status != ATTR_EXISTS)  {
       strcpy(msg, "1...CREATE_, ATTR_, ");
       strcat(msg, gAttr);
       cdf_status_handler (status, msg);
     } else if (status == ATTR_EXISTS) {
       status = CDFlib (GET_, ATTR_NUMBER_, gAttr, &attrNum,
                        NULL_);
       if (status != CDF_OK)  {
         strcpy(msg, "GET_, ATTR_NUM_, ");
         strcat(msg, gAttr);
         cdf_status_handler (status, msg);
       }
       *created = FALSE;
     } else
       *created = TRUE; 
   } else {
     status = CDFlib (CREATE_, ATTR_, attrName,    /* in - attribute name    */
                                      attrScope,   /* in - attribute scope   */
                                      &attrNum,    /* out - attribute number */
                      NULL_);
     if (status != CDF_OK && status != ATTR_EXISTS)  {
       strcpy(msg, "2...CREATE_, ATTR_, ");
       strcat(msg, attrName);
       cdf_status_handler (status, msg);
     }
     if (attrNum > -1) 
       *created = TRUE;
     else {
       status = CDFlib (GET_, ATTR_NUMBER_, attrName, &attrNum,
                        NULL_);
       *created = FALSE;
     }
   }

   return attrNum;
}


/*--------------------------------------------------------------------------
 *  This routine creates a CDF global attribute entry.  An attribute
 *  can have one or more entries.
 *--------------------------------------------------------------------------*/
void create_cdf_gAttribute_entry (long attrNum,     /* Attr number          */
                                  long entryNum,    /* Attr entry number    */
                                  long datatype,    /* Attr entry datatype  */
                                  long numElements, /* # of attr entries    */
                                  void *data)       /* Attr entry data      */
{
   CDFstatus status; int i;

   /**************************************************************************
    * If the datatype is a string (i.e. CDF_CHAR), the value of 'numElements'
    * must be the length of the string.  For non-string data, the value of
    * 'numElements' represents the number of values in 'data'.
    **************************************************************************/
   if (datatype == CDF_CHAR) {
/*       if (deleteNULL) numElements = strlen( (char *) data); */
       if (numElements > 0) {
          if (deleteNULL) numElements = strlen( (char *) data);
       }
/*
           numElements = 1;
           strcpy((char *) data, " ");
           *(char *) data = '\0';
*/
   }

   if (fillvalue) {
     if (datatype == CDF_INT1) {
/*
       for (i =0; i < (int) numElements; ++i)
         if ((*(signed char *)data+i) == NC_FILL_BYTE)
           *((signed char *)data+i) = (signed char) DEFAULT_INT1_PADVALUE;
*/
     } else if (datatype == CDF_UINT1) {
       for (i =0; i < (int) numElements; ++i)
         if ((*(unsigned char *)data+i) == NC_FILL_UBYTE)
           *((unsigned char *)data+i) = (unsigned char) DEFAULT_UINT1_PADVALUE;
     } else if (datatype == CDF_INT2) {
/*
         for (i =0; i < (int) numElements; ++i)
         if ((*(short *)data+i) == NC_FILL_SHORT)
           *((short *)data+i) = (short) DEFAULT_INT2_PADVALUE;
*/
     } else if (datatype == CDF_UINT2) {
       for (i =0; i < (int) numElements; ++i)
         if ((*(unsigned short *)data+i) == NC_FILL_USHORT)
           *((unsigned short *)data+i) = (unsigned short) DEFAULT_UINT2_PADVALUE;
     } else if (datatype == CDF_INT4) {
/*
       for (i =0; i < (int) numElements; ++i)
         if ((*(int *)data+i) == NC_FILL_INT)
           *((int *)data+i) = (unsigned int) DEFAULT_INT4_PADVALUE;
*/
     } else if (datatype == CDF_UINT4) {
       for (i =0; i < (int) numElements; ++i)
         if ((*(unsigned int *)data+i) == NC_FILL_UINT)
           *((unsigned int *)data+i) = (unsigned int) DEFAULT_UINT4_PADVALUE;
     } else if (datatype == CDF_INT8) {
       for (i =0; i < (int) numElements; ++i)
         if ((*(long long *)data+i) == NC_FILL_INT64)
           *((long long *)data+i) = (long long) DEFAULT_INT8_PADVALUE;
     } else if (datatype == CDF_FLOAT) {
       for (i =0; i < (int) numElements; ++i)
         if ((*(float *)data+i) == NC_FILL_FLOAT)
           *((float *)data+i) = (float) DEFAULT_FLOAT_PADVALUE;
     } else if (datatype == CDF_DOUBLE) {
       for (i =0; i < (int) numElements; ++i)
         if ((*(double *)data+i) == NC_FILL_DOUBLE)
           *((double *)data+i) = (double) DEFAULT_DOUBLE_PADVALUE;
     } else if (datatype == CDF_EPOCH) {
       for (i =0; i < (int) numElements; ++i)
         if ((*(double *)data+i) == NC_FILL_DOUBLE)
           *((double *)data+i) = (double) DEFAULT_EPOCH_PADVALUE;
     } else if (datatype == CDF_EPOCH16) {
       for (i =0; i < (int) numElements; ++i) {
         if (*((double *)data+2*i) == NC_FILL_DOUBLE)
           *((double *)data+2*i) = (double) DEFAULT_EPOCH16_PADVALUE;
         if (*((double *)data+2*i+1) == NC_FILL_DOUBLE)
           *((double *)data+2*i+1) = (double) DEFAULT_EPOCH16_PADVALUE;
       }
     } else if (datatype == CDF_TIME_TT2000) {
       for (i =0; i < (int) numElements; ++i)
         if (*((long long *)data+i) == NC_FILL_INT64)
           *((long long *)data+i) = (long long) DEFAULT_TT2000_PADVALUE;
     }
   }
   if ((datatype != 51 && datatype != 52) || numElements > 0)
     status = CDFlib (SELECT_, ATTR_, attrNum,
                               gENTRY_, entryNum,
                      PUT_, gENTRY_DATA_, datatype,
                                          numElements,
                                          data,
                      NULL_);
   else {
     char dataX[1];
     dataX[0] = (char) '\0';
     status = CDFlib (SELECT_, ATTR_, attrNum,
                               gENTRY_, entryNum,
                      PUT_, gENTRY_DATA_, datatype,
                                          1L,
                                          dataX,
                      NULL_);
   }
   if (status != CDF_OK) cdf_status_handler (status, "PUT_, gENTRY_DATA_");
}

/*--------------------------------------------------------------------------
 *  This routine creates the "NC_KIND" global attribute and entry.
 *--------------------------------------------------------------------------*/
void create_cdf_KIND_gAttribute (int kind)
{
   CDFstatus status;
   long attrNum;

   status = CDFlib (CREATE_, ATTR_, "NC_KIND", GLOBAL_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS)
     cdf_status_handler (status, "3.1 ...CREATE_, ATTR_");
   status = CDFlib (SELECT_, ATTR_, attrNum,
                             gENTRY_, 0L,
                    PUT_, gENTRY_DATA_, (long) CDF_CHAR,
                                        (long) strlen (kind_string(kind)),
                                        kind_string(kind),
/*
                    SELECT_, gENTRY_, 1L,
                    PUT_, gENTRY_DATA_, (long) CDF_CHAR,
                                        (long) strlen (setAttributeType==NC_CHAR?"NC_CHAR":"NC_STRING"),
                                        (setAttributeType==NC_CHAR?"NC_CHAR":"NC_STRING"),
*/
                    NULL_);
   if (status != CDF_OK)
     cdf_status_handler (status, "3.1c ...PUT_, gENTRY_DATA_");
}


/*--------------------------------------------------------------------------
 *  This routine creates the "NC_DELIMITER" global attribute and entry.
 *--------------------------------------------------------------------------*/
void create_cdf_DELIMITER_gAttribute ()
{
   CDFstatus status;
   long attrNum;

   status = CDFlib (CREATE_, ATTR_, "NC_DELIMITER", GLOBAL_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK)
     cdf_status_handler (status, "3.1d ...CREATE_, ATTR_");
   status = CDFlib (SELECT_, ATTR_, attrNum,
                             gENTRY_, 0L,
                    PUT_, gENTRY_DATA_, (long) CDF_CHAR, 1L,
                                        delimiters[curDiD],
                    NULL_);
   if (status != CDF_OK)
     cdf_status_handler (status, "3.1e ...PUT_, gENTRY_DATA_");
}


/*--------------------------------------------------------------------------
 *  This routine creates the "NC_DIMENSIONS_V" variable attribute.
 *--------------------------------------------------------------------------*/
void create_cdf_DIMENSIONS_vAttribute ()
{
   CDFstatus status;
   long attrNum;

   status = CDFlib (CREATE_, ATTR_, "NC_DIMENSIONS_V", VARIABLE_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS)
     cdf_status_handler (status, "3.2 ...CREATE_, ATTR_");
}


/*--------------------------------------------------------------------------
 *  This routine creates and loads netCDF's dimensions to "NC_DIMENSIONS_G"
 *  global attribute.
 *--------------------------------------------------------------------------*/
void create_cdf_DIMENSIONS_gAttribute (int ncid, int nDims, int start)
{
   CDFstatus status;
   int statusN;
   char entry[NC_MAX_NAME + 20 + 1];
   int  id, len, i, found;
   long attrNum;

   if (start == 0) {
     status = CDFlib (CREATE_, ATTR_, "NC_DIMENSIONS_G", GLOBAL_SCOPE, &attrNum,
                      NULL_);
     if (status != CDF_OK) cdf_status_handler (status, "4...CREATE_, ATTR_");
   }
   for (id = 0; id < nDims; id++) {
      strcpy(entry, dims[start+id].name);
      strcat(entry,"=");
      found = FindUnlimitedDim (id);
      if (found > -1) {
          strcat(entry, "UNLIMITED");
      } else {
          char nums[21];
          sprintf (nums,"%u",(unsigned int) dims[start+id].size);
          strcat(entry, nums);
      }
      len = strlen(entry);
      status = CDFlib (SELECT_, ATTR_NAME_, "NC_DIMENSIONS_G",
                                gENTRY_, (long) (id+start),
                       PUT_, gENTRY_DATA_, (long) CDF_CHAR, (long) len, entry,
                       NULL_);
      if (status != CDF_OK) cdf_status_handler (status, "PUT_, gENTRY_DATA");
   }  

}

/*--------------------------------------------------------------------------
 *  This routine creates CDF attributes: "NC_STRINGS_G" or "NC_STRINGS_V"
 *  to keep records of strings defined in netCDF's attributes.
 *--------------------------------------------------------------------------*/
void create_cdf_STRING_Attribute (int which)
{
   CDFstatus status;
   long attrNum;

   if (which == 1) {
     status = CDFlib (CREATE_, ATTR_, "NC_STRINGS_G", GLOBAL_SCOPE, &attrNum,
                      NULL_);
     if (status != CDF_OK)
       cdf_status_handler (status, "5...CREATE_, NC_STRINGS_G");
   } else {
     status = CDFlib (CREATE_, ATTR_, "NC_STRINGS_V", VARIABLE_SCOPE, &attrNum,
                      NULL_);
     if (status != CDF_OK)
       cdf_status_handler (status, "5...CREATE_, NC_STRINGS_V");
   }
}

/*--------------------------------------------------------------------------
 *  This routine adds an entry to global attribute "NC_STRINGS_G".
 *--------------------------------------------------------------------------*/
void add_STRINGS_G (int ncid)
{
   CDFstatus status;
   char attrName[NC_MAX_NAME+1];
   int  j, natts, nStatus, type;
   char *entry = NULL;

   nStatus = nc_inq_natts (ncid, &natts);
   if (nStatus != NC_NOERR) handle_error(560, nStatus);
   for (j = 0; j < natts; ++j) {
     nStatus = nc_inq_attname (ncid, NC_GLOBAL, j, attrName);
     if (nStatus != NC_NOERR) handle_error(561, nStatus);
     nStatus = nc_inq_atttype (ncid, NC_GLOBAL, attrName, &type);
     if (nStatus != NC_NOERR) handle_error(562, nStatus);
     if (type == NC_STRING) {
       if (entry == NULL) {
         entry = (char *) malloc (strlen(attrName)+1);
         strcpy (entry, attrName);
       } else {
         entry = (char *) realloc (entry, strlen(entry)+strlen(attrName)+2);
         strcat (entry, " ");
         strcat (entry, attrName);
       }
     }
   }
   if (entry != NULL) {
     status = CDFlib (SELECT_, ATTR_NAME_, "NC_STRINGS_G",
                      NULL_);
     if (status != CDF_OK) create_cdf_STRING_Attribute (1);
     status = CDFlib (SELECT_, ATTR_NAME_, "NC_STRINGS_G",
                               gENTRY_, 0L,
                      PUT_, gENTRY_DATA_, CDF_CHAR, (long) strlen(entry),
                                                    entry,
                      NULL_);
     free (entry);
     if (status != CDF_OK) cdf_status_handler (status, "PUT_, GENTRY_DATA_");
   }
}

/*--------------------------------------------------------------------------
 *  This routine adds an entry to variable attribute "NC_STRINGS_V".
 *--------------------------------------------------------------------------*/
void add_STRINGS_V (int ncid)
{
   CDFstatus status;
   char varName[NC_MAX_NAME+1], attrName[NC_MAX_NAME+1];
   int  i, j, natts, nvars, nStatus, type;
   char *entry;

   nStatus = nc_inq_varids (ncid, &nvars, NULL);
   if (nStatus != NC_NOERR) handle_error(660, nStatus);
   for (j = 0; j < nvars; ++j) {
     entry = NULL;
     nStatus = nc_inq_var (ncid, j, varName, NULL, NULL, NULL, &natts);
     if (nStatus != NC_NOERR) handle_error(661, nStatus);
     for (i = 0; i < natts; ++i) {
       nStatus = nc_inq_attname (ncid, j, i, attrName);
       if (nStatus != NC_NOERR) handle_error(662, nStatus);
       nStatus = nc_inq_atttype (ncid, j, attrName, &type);
       if (nStatus != NC_NOERR) handle_error(663, nStatus);
       if (type == NC_STRING) {
         if (entry == NULL) {
           entry = (char *) malloc (strlen(attrName)+1);
           strcpy (entry, attrName);
         } else {
           entry = (char *) realloc (entry, strlen(entry)+strlen(attrName)+2);
           strcat (entry, " ");
           strcat (entry, attrName);
         }
       }
     }
     if (entry != NULL) {
       status = CDFlib (SELECT_, ATTR_NAME_, "NC_STRINGS_V",
                        NULL_);
       if (status != CDF_OK) create_cdf_STRING_Attribute (2);
       status = CDFlib (SELECT_, ATTR_NAME_, "NC_STRINGS_V",
                                 zENTRY_, (long)j,
                        PUT_, zENTRY_DATA_, CDF_CHAR, (long) strlen(entry),
                                                      entry,
                        NULL_);
       free (entry);
       if (status != CDF_OK) cdf_status_handler (status, "PUT_, gENTRY_DATA");
     }
   }
}

/*--------------------------------------------------------------------------
 *  This routine adds non-unlimited dimension variables and DEPEND_* 
 *  attributes and entries if they don't already exist.
 *--------------------------------------------------------------------------*/
void add_Depends (int ncid, int nVars)
{
   CDFstatus status;
   long numgAtts, numvAtts, scope;
   char attrName[CDF_ATTR_NAME_LEN256+1];
   int  j, toskip;

   toskip = 0;
   status = CDFlib (GET_, CDF_NUMvATTRS_, &numvAtts,
                          CDF_NUMgATTRS_, &numgAtts,
                    NULL_);
   if (status != CDF_OK) cdf_status_handler (status, "GET_, CDF_NUMvATTRS");
   for (j = 0; j < (int)(numvAtts+numgAtts); ++j) {
     status = CDFlib (SELECT_, ATTR_, (long) j,
                      GET_, ATTR_SCOPE_, &scope,
                      NULL_);
     if (scope == GLOBAL_SCOPE) continue;
     status = CDFlib (GET_, ATTR_NAME_, attrName,
                      NULL_);
     if (strcmpIgCase(attrName, "depend_0") == 1) {
        toskip = 1;
        break;
     }
   }
   write_cdf_DIMENSION_variables (ncid, nVars);
   if (toskip == 0) {
     /* "depend_*" not defined originally.... */
/*     write_cdf_DIMENSION_variables (ncid, nVars); */
     write_cdf_DIMENSION_depends (ncid, nVars);
   }
}

/*--------------------------------------------------------------------------
 *  This routine adds VAR_TYPE variable attribute and entries if they don't
 *  already exist.
 *--------------------------------------------------------------------------*/
void add_VarType (int ncid, int nVars)
{
   CDFstatus status;
   long numgAtts, numvAtts, scope;
   char attrName[CDF_ATTR_NAME_LEN256+1];
   int  j, toskip;

   toskip = 0;
   status = CDFlib (GET_, CDF_NUMvATTRS_, &numvAtts,
                          CDF_NUMgATTRS_, &numgAtts,
                    NULL_);
   if (status != CDF_OK) cdf_status_handler (status, "GET_, CDF_NUMvATTRS");
   for (j = 0; j < (int)(numvAtts+numgAtts); ++j) {
     status = CDFlib (SELECT_, ATTR_, (long) j,
                      GET_, ATTR_SCOPE_, &scope,
                      NULL_);
     if (scope == GLOBAL_SCOPE) continue;
     status = CDFlib (GET_, ATTR_NAME_, attrName,
                      NULL_);
     if (strcmpIgCase(attrName, "var_type") == 1) {
        toskip = 1;
        break;
     }
   }
   if (toskip == 0) {
     /* "var_type" not defined originally.... */
     add_cdf_attribute_vartype (ncid, nVars);
   }
}

/*--------------------------------------------------------------------------
 *  This routine writes netCDF's dimensions as CDF variables if they are 
 *  not defined as variables in netCDF.
 *--------------------------------------------------------------------------*/
void write_cdf_DIMENSION_variables (int ncid, int nVars)
{
   CDFstatus status;
   int nStatus;
   int  id, j, found;
   nc_type rh_type;
   int  rh_id, rh_ndims, rh_dimids[NC_MAX_VAR_DIMS], rh_natts;
   size_t recs;
   char dimName[NC_MAX_NAME+1];
   long attrNum, varNum;

   for (id = 0; id < nVars; id++) {
       nStatus = nc_inq_var (ncid, id, 0, &rh_type, &rh_ndims, rh_dimids,
                             &rh_natts);
       if (nStatus != NC_NOERR) handle_error(690, nStatus);
       for (j = 0; j < rh_ndims; ++j) {
         addedAttrs = (char *) malloc(201);
         strcpy (addedAttrs, "");
         nStatus = nc_inq_dim(ncid, rh_dimids[j], dimName, &recs);
         if (nStatus != NC_NOERR) handle_error(691, nStatus);
         nStatus = nc_inq_varid (ncid, dimName, &rh_id);
         if (nStatus == NC_NOERR) {
           /* This dimension is a netCDF variable. So it is already in CDF. */ 
           free (addedAttrs);
           continue;
         }
         /* Not an netCDF variable... We need to add it as a variable to CDF. */
         found = FindUnlimitedDim (rh_dimids[j]);
         if (found < 0) {
           /* Non-unlimited dimension... Create NRV variable. */
           long dimSizes[] = { (long) recs};
           long dimVariance[] = { (long) VARY};
           long dimIndices[] = {0L}, dimIntervals[] = {1L};
           void  *data;
           status = CDFlib (CREATE_, zVAR_, dimName, CDF_INT1, 1L,
                                            1L, dimSizes, NOVARY,
                                            dimVariance, &varNum,
                            NULL_);
           if (status != CDF_OK && status != VAR_EXISTS)
             cdf_status_handler (status, "CREATE_, zVAR_1");
           if (status == CDF_OK) {
             /* Just create the variable. */
             data = (void *) malloc(sizeof(char) * recs);
             for (j = 0; j < recs; ++j) ((char *)data)[j] = (char) (j + 1);
             status = CDFlib (SELECT_,zVAR_RECNUMBER_, 0L,
                                      zVAR_RECCOUNT_, 1L,
                                      zVAR_RECINTERVAL_, 1L,
                                      zVAR_DIMINDICES_, dimIndices,
                                      zVAR_DIMCOUNTS_, dimSizes,
                                      zVAR_DIMINTERVALS_, dimIntervals,
                              PUT_, zVAR_HYPERDATA_, data,
                              NULL_);
              free (data);
              if (status != CDF_OK) cdf_status_handler (status, "PUT_, zVAR_1");
           }
         } else {
           /* Unlimited dimension... Create RV variable. */
           long dimSizes[] = { (long) recs};
           long dimVariance[] = { (long) NOVARY};
           long dimIndices[] = {0L}, dimIntervals[] = {1L};
           status = CDFlib (CREATE_, zVAR_, dimName, CDF_INT1, 1L,
                                            0L, dimSizes, VARY,
                                            dimVariance, &varNum,
                            NULL_);
           if (status != CDF_OK && status != VAR_EXISTS)
             cdf_status_handler (status, "CREATE_, zVAR_2");
           if (status == CDF_OK) {
             status = CDFlib (SELECT_,zVAR_RECNUMBER_, 0L,
                                      zVAR_RECCOUNT_, 1L,
                                      zVAR_RECINTERVAL_, 1L,
                                      zVAR_DIMINDICES_, dimIndices,
                                      zVAR_DIMCOUNTS_, dimSizes,
                                      zVAR_DIMINTERVALS_, dimIntervals,
                              PUT_, zVAR_HYPERDATA_, &recs,
                              NULL_);
             if (status != CDF_OK) cdf_status_handler (status, "PUT_, zVAR_2");
           }
         }
         /* Add an attribute "NC_ADDED". */
         status = CDFlib (CREATE_, ATTR_, "NC_ADDED", VARIABLE_SCOPE, &attrNum,
                          NULL_);
         if (status != CDF_OK && status != ATTR_EXISTS)
            cdf_status_handler (status, "CREATE_, ATTR_");
         strcat (addedAttrs, "variable ");
         add_cdf_dimension_extra (varNum, dimName);
         free (addedAttrs);
       }
   }

}

/*--------------------------------------------------------------------------
 *  This routine writes variables' DEPEND_ attributes/entries based on
 *  netCDF's non-unlimited dimensions.
 *--------------------------------------------------------------------------*/
void write_cdf_DIMENSION_depends (int ncid, int nVars)
{
   CDFstatus status;
   char ext[3];
   int  id, j;
   long dimSizes[CDF_MAX_DIMS], numDims, recVariance, attrNum;
   char depend[9];
   int  nStatus, idx;
   nc_type rh_type;
   int  rh_id, rh_ndims, rh_dimids[NC_MAX_VAR_DIMS], rh_natts;
   size_t recs;
   char dimName[NC_MAX_NAME+1];

   for (id = 0; id < nVars; id++) {
       addedAttrs = (char *) malloc (201);
       strcpy (addedAttrs, "");
       nStatus = nc_inq_var (ncid, id, 0, &rh_type, &rh_ndims, rh_dimids,
                             &rh_natts);
       if (nStatus != NC_NOERR) handle_error(692, nStatus);
       get_nc_variable_attributes (ncid, id);
       for (j = 0; j < rh_ndims; ++j) {
         strcpy(depend, "DEPEND_");
         sprintf(ext, "%d", j);
         strcat(depend, ext);
         idx = search_variable_attributes (depend);
         if (idx > -1) continue;
         nStatus = nc_inq_dim(ncid, rh_dimids[j], dimName, &recs);
         if (nStatus != NC_NOERR) handle_error(693, nStatus);
         /* Do we have a netCDF variable by that dimension name? */
         nStatus = nc_inq_varid (ncid, dimName, &rh_id);
         if (nStatus != NC_NOERR) {
/*
           status = CDFlib (SELECT_, ATTR_NAME_, "NC_DEPEND",
                                     zENTRY_, (long) id,
                            PUT_, zENTRY_DATA_, CDF_CHAR, 9L, "attribute",
                            NULL_);
           if (status != CDF_OK) cdf_status_handler (status, "PUT_, zENTRY_DATA");
*/
         }
         strcat (addedAttrs, depend);
         strcat (addedAttrs, " ");
         status = CDFlib (CREATE_, ATTR_, depend, VARIABLE_SCOPE, &attrNum,
                          NULL_);
         if (status != CDF_OK && status != ATTR_EXISTS)
            cdf_status_handler (status, "CREATE_, ATTR_");
         if (status == ATTR_EXISTS)
           status = CDFlib (GET_, ATTR_NUMBER_, depend, &attrNum,
                            NULL_);
         status = CDFlib (SELECT_, ATTR_, attrNum,
                                   zENTRY_, (long) id,
                          PUT_, zENTRY_DATA_, CDF_CHAR, (long)strlen(dimName),
                                dimName,
                          NULL_);
         if (status != CDF_OK) cdf_status_handler (status, "PUT_, zENTRY_DATA");
       }
       status = CDFlib (CREATE_, ATTR_, "NC_ADDED", VARIABLE_SCOPE, &attrNum,
                        NULL_);
       if (status != CDF_OK && status != ATTR_EXISTS) {
         cdf_status_handler (status, "018...CREATE_, ATTR_");
         return;
       }
       status = CDFlib (SELECT_, ATTR_NAME_, "NC_ADDED",
                                 zENTRY_, (long) id,
                        PUT_, zENTRY_DATA_, CDF_CHAR, (long) strlen(addedAttrs)
                                                      - 1, addedAttrs,
                        NULL_);
       if (status != CDF_OK) cdf_status_handler (status, "019...PUT_, ENTRY_");
       free (addedAttrs);
   }
}

/*--------------------------------------------------------------------------
 *  This routine gets a netCDF variable's attributes.
 *--------------------------------------------------------------------------*/
void get_nc_variable_attributes (int ncid, int varid)
{
   char name[NC_MAX_NAME + 1];
   int  type, j; size_t len;
   int  nStatus;
   nc_type rh_type;
   int  rh_id, rh_ndims, rh_dimids[NC_MAX_VAR_DIMS], rh_natts;

   nStatus = nc_inq_var (ncid, varid, 0, &rh_type, &rh_ndims, rh_dimids,
                         &rh_natts);
   if (nStatus != NC_NOERR) handle_error(694, nStatus);
   if (numAttrs > 0) {
     for (j = 0; j < numAttrs; ++j) free (vAttrs[j]);
     free (vAttrs);
   }
   if (rh_natts > 0) {
     vAttrs = (char **) malloc (sizeof(char *) * rh_natts);
     for (j = 0; j < rh_natts; ++j) {
       nStatus = nc_inq_attname(ncid, varid, j, name);
       nStatus = nc_inq_att(ncid, varid, name, &type, &len);
       if (nStatus != NC_NOERR) handle_error(695, nStatus);
       vAttrs[j] = (char *) malloc (sizeof(char) * len + 1);
       strcpy (vAttrs[j], name);
     }
   }
   numAttrs = rh_natts;
}

/*--------------------------------------------------------------------------
 *  This routine searches for variable's attributes to see if a given one
 *  is there (case in-sensitive). if found, the index is returned. -1,
 *  otherwise. 
 *--------------------------------------------------------------------------*/
int search_variable_attributes (char *attr)
{
   int  j;

   if (numAttrs > 0) {
     for (j = 0; j < numAttrs; ++j) {
       if (strcmpIgCase (attr, vAttrs[j]) == 1) return j;
     }
     return -1;
   } else
     return -1;
}

/*--------------------------------------------------------------------------
 *  This routine creates the "NC_GROUP" global attribute.
 *--------------------------------------------------------------------------*/
void create_cdf_GROUP_gAttribute ()
{
   CDFstatus status;
   long attrNum;

   status = CDFlib (CREATE_, ATTR_, "NC_GROUP", GLOBAL_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS)
     cdf_status_handler (status, "3.2 ...CREATE_, ATTR_");
}


/*--------------------------------------------------------------------------
 *  This routine loads netCDF's group information to "NC_GROUP"
 *  global attribute.
 *--------------------------------------------------------------------------*/
void add_cdf_GROUP_gAttribute (char *grpName)
{
   CDFstatus status;
   int  i;
   long numElems, numEntries;
   char *entry;

   status = CDFlib (SELECT_, ATTR_NAME_, "NC_GROUP",
                    NULL_);
   if (status != CDF_OK) cdf_status_handler (status, "SELECT, NC_GROUP");

   status = CDFlib (GET_, ATTR_NUMgENTRIES_, &numEntries,
                    NULL_);
   if (status != CDF_OK) cdf_status_handler (status, "GET, NC_GROUP");
   if (numEntries == 0) {
     status = CDFlib (SELECT_, gENTRY_, 0L,
                      PUT_, gENTRY_DATA_, (long) CDF_CHAR,
                            (long) strlen(grpName), grpName,
                      NULL_);
     if (status != CDF_OK) cdf_status_handler (status, "PUT, NC_GROUP");
   } else {
     for (i = 0; i < (int) numEntries; i++) {
       status = CDFlib (SELECT_, gENTRY_, (long) i,
                        GET_, gENTRY_NUMELEMS_, &numElems,
                        NULL_);
       if (status != CDF_OK) cdf_status_handler (status, "GET, NC_GROUP");
       entry = malloc ((size_t) numElems + 1);
       status = CDFlib (GET_, gENTRY_DATA_, entry,
                        NULL_);
       if (status != CDF_OK) {
         free (entry);
         cdf_status_handler (status, "GET, NC_GROUP");
       }
       *(entry+numElems) = '\0';
       if (strcmp(entry, grpName) == 0) {
         free (entry);
         return;
       }
       free (entry);
     }
     status = CDFlib (SELECT_, gENTRY_, (long) numEntries,
                      PUT_, gENTRY_DATA_, (long) CDF_CHAR,
                            (long) strlen(grpName), grpName,
                      NULL_);
     if (status != CDF_OK) cdf_status_handler (status, "PUT_, gENTRY_DATA");
   }  

}

/*--------------------------------------------------------------------------
 *  This routine find the dimension name from the global attribute: 
 *  NC_DIMENSIONS_G.
 *--------------------------------------------------------------------------*/
void searchPreentGroup (char *grp, char *dimname, char *entry)
{
   CDFstatus status;
   int statusN;
   char tmp[NC_MAX_NAME + 1], tmp2[NC_MAX_NAME + 1];
   int  id, len, togo;
   long numEntries, numElems;
   char **entries, *ptr, *ptr2, *value0, *ldimname = NULL;
 
   status = CDFlib (SELECT_, ATTR_NAME_, "NC_DIMENSIONS_G",
                    GET_,  ATTR_NUMgENTRIES_, &numEntries,
                    NULL_);
   if (status != CDF_OK)
     cdf_status_handler (status, "SELECT/GET NC_DIMENSIONS_G");
   if (numEntries == 0) return;
   entries = (char **) malloc (sizeof(char *) * numEntries);
   for (id = 0; id < (int) numEntries; id++) {
      status = CDFlib (SELECT_, gENTRY_, (long) id,
                       GET_, gENTRY_NUMELEMS_, &numElems,
                       NULL_);
      if (status != CDF_OK) cdf_status_handler (status, "GET, gENTRY_NUMELEMS");
      value0 = malloc ((size_t)numElems+1);
      entries[id] = (char *) malloc ((size_t)numElems+1);
      ldimname = malloc ((size_t)numElems+1);
      status = CDFlib (GET_, gENTRY_DATA_, value0,
                       NULL_);
      if (status != CDF_OK) cdf_status_handler (status, "GET, gENTRY_NUMELEMS");
      value0[(int)numElems] = (char) '\0';
      ptr = strstr (value0, "=");
      *(ptr) = (char) '\0';
      strcpy (entries[id], value0);
      free (value0);
   }
   if (strlen(grp) == 1) 
     strcpy (tmp, dimname);
   else {
     strcpy (tmp, grp+1);
     strcat (tmp, delimiters[curDiD]);
     strcat (tmp, dimname);
   }
   togo = 1;
   while (togo) {
      for (id = 0; id < (int) numEntries; id++) {
        if (strcmp(entries[id], tmp) == 0) {
          strcpy (entry, tmp);
          togo = 0;
          break;
        }
      }
      if (togo) {
        strcpy (tmp2, tmp);
        ptr = strrchr (tmp2, (char)*delimiters[curDiD]);
        *ptr = (char) '\0';
        ptr2 = strrchr (tmp2, (char)*delimiters[curDiD]);
        if (ptr2 == NULL) strcpy (tmp, dimname);
        else {
          *ptr2 = (char ) '\0';
          strcpy (tmp, tmp2);
          strcat (tmp, delimiters[curDiD]);
          strcat (tmp, dimname);
        }
      }
   }  
   for (id = 0; id < numEntries; ++id) free (entries[id]);
   free (entries); 
}


/*--------------------------------------------------------------------------
 *  This routine writes the entry to "NC_DIMENSIONS_V" variable attribute.
 *--------------------------------------------------------------------------*/
void write_cdf_DIMENSIONS_vAttribute_entry (int ncid,   /* in - netCDF id   */
                                            int numgdims, /* in - grp dim cnt */
                                            int *gdimids, /* in - grp dim ids */
                                            char *grpName, /* in - group name */
                                            int varId, /* in - netCDF var ID */
                                            long varNum) /* in - CDF var num */
{
   CDFstatus status;
   int statusN;
   char dim_name[NC_MAX_NAME + 1], varName[NC_MAX_NAME + 1];
   char *entry, *value;
   int  isLocal, j, id, ndims, natts, len, dimsIds[NC_MAX_DIMS];
   nc_type  type;
   long dataType;

   statusN = nc_inq_var(ncid, varId, varName, &type, &ndims, dimsIds, &natts);
   if (statusN != NC_NOERR) handle_error(70, statusN);
   if (ndims == 0) return;
   entry = (char *) malloc ((size_t) ndims * NC_MAX_NAME + ndims + 1 );
   entry[0] = (char) '\0';
   value = (char *) malloc ((size_t) ndims * NC_MAX_NAME + ndims + 1 );
   len = 0;
   status = CDFlib (SELECT_, zVAR_, varNum,
                    GET_, zVAR_DATATYPE_, &dataType,
                    NULL_);
   if (dataType == CDF_EPOCH16) ndims--;
   for (id = 0; id < ndims; id++) {
      statusN = nc_inq_dimname(ncid, dimsIds[id], dim_name);
      if (statusN != NC_NOERR) handle_error(71, statusN);
      isLocal = 0;
      for (j = 0; j < numgdims; ++j) {
        if (dimsIds[id] == gdimids[j]) {
          isLocal = 1;
          break;
        }
      }
      if (isLocal == 1) {
        strncat (entry, grpName+1, strlen(grpName)-1);
        if (strlen(grpName) > 1) strcat (entry, delimiters[curDiD]);
        strncat (entry, dim_name, strlen(dim_name));
      } else {
        searchPreentGroup (grpName, dim_name, value);
        strcat (entry, value);
      }
      if (id < (ndims-1)) strncat (entry, ",", 2);
      len = strlen(entry);
   }
   if (len == 0) return;
   for (id = 0; id < len; id++) {
     if (*(entry+id) == (char) '/')
       *(entry+id) = (char)*delimiters[curDiD];
   }
   status = CDFlib (SELECT_, ATTR_NAME_, "NC_DIMENSIONS_V",
                             zENTRY_, varNum,
                    PUT_, zENTRY_DATA_, (long) CDF_CHAR, (long) len, entry,
                    NULL_);
   free (entry);
   if (status != CDF_OK) cdf_status_handler (status, "1... PUT_, zENTRY_DATA");
}


/*--------------------------------------------------------------------------
 *  This routine creates a variable attribute entry for the given 
 *  variable name and attribute name.  Unlike global attribute, the
 *  variable attribute can't have more than one entry.  
 *--------------------------------------------------------------------------*/
void create_cdf_vAttribute_entry (char *attrName,   /* CDF attribute name   */
                                  struct cdfvar *var,/* CDF variable struct */
                                  long datatype,    /* Attr entry datatype  */
                                  long numElements, /* # of attr entries    */
                                  void *data)       /* Attr entry data      */
{
   CDFstatus status;
   int toChange, cvtStatus;
   void *dataNew;
   /**************************************************************************
    * If the datatype is a string (i.e. CDF_CHAR), the value of 'numElements'
    * must be the length of the string.  For non-string data, the value of
    * 'numElements' represents the number of values in 'data'.
    **************************************************************************/
   if (datatype == CDF_CHAR) {
       if (deleteNULL) numElements = strlen( (char *) data);
       if (numElements == 0) {                   /* Empty data */
           numElements = 1; 
/*           strcpy((char *) data, " "); */
           *(char *) data = '\0';
       }
   }
   status = CDFlib (SELECT_, ATTR_NAME_, attrName,
                             zENTRY_NAME_, var->name,
                    NULL_);
   if (status != CDF_OK) cdf_status_handler (status, "SELECT_, zENTRY_DATA_");
   toChange = 0;
   cvtStatus = 1;
   if (ENFORCE && findSpecialAttributes(attrName) > 0) {
     if (datatype != var->datatype) {
       toChange = 1;
       dataNew = malloc(CDFelemSize(var->datatype) * numElements);
       cvtStatus = ConvertData (dataNew, var->datatype, data, datatype,
                                (int)numElements);
     }
   }
   if (cvtStatus == 0) {
     printf("**** Variable attributes for "
            "valid_min/validmin/valid_max/validmax\n");
     printf("     and/or fill_val are invalid.... likely outside the valid "
            "range\n");
     printf("     for the variable's defined data type.\n");
     printf("**** Stop...\n\n");
     status = CDFlib (CLOSE_, CDF_,
                      NULL_);
     status = CDFlib (OPEN_, CDF_, cdfFile, &id,
                      DELETE_, CDF_,
                      NULL_);
     free (cdfFile);
     free (dataNew);
     if (numAttrs > 0) {
       int j;
       for (j = 0; j < numAttrs; ++j) free (vAttrs[j]);
       free (vAttrs);
     }
     exit(1);
   }
   if (((EPOCH && FindEpoch (var->name)) || ISTPepoch) &&
       (findSpecialAttributes(attrName) > 0)) {
     if (var->datatype == CDF_EPOCH) {
       if (numElements == EPOCH_STRING_LEN) {
         double epoch;
         epoch = parseEPOCH((char *)data);
         status = CDFlib (PUT_, zENTRY_DATA_, CDF_EPOCH, 1L, &epoch,
                          NULL_);
       } else {
         if (numElements == 1 && datatype == CDF_CHAR)
           *(double *)data = DEFAULT_EPOCH_PADVALUE;
         if (fillvalue && *(double *)data == NC_FILL_DOUBLE) 
           *(double *)data = (double) DEFAULT_EPOCH_PADVALUE;
         status = CDFlib (PUT_, zENTRY_DATA_, CDF_EPOCH, 1L, data,
                          NULL_);
       }
     } else if (var->datatype == CDF_EPOCH16) {
       if (numElements == EPOCH16_STRING_LEN) {
         double epoch[2], tmp;
         tmp = parseEPOCH16((char *)data, epoch);
         status = CDFlib (PUT_, zENTRY_DATA_, CDF_EPOCH16, 1L, epoch,
                          NULL_);
       } else {
         if (numElements == 1 && datatype == CDF_CHAR) {
           *(double *)data = DEFAULT_EPOCH16_PADVALUE;
           *((double *)data+1) = DEFAULT_EPOCH16_PADVALUE;
         }
         if (fillvalue && *(double *)data == NC_FILL_DOUBLE) {
           *(double *)data = DEFAULT_EPOCH16_PADVALUE;
           *((double *)data+1) = DEFAULT_EPOCH16_PADVALUE;
         }
         status = CDFlib (PUT_, zENTRY_DATA_, CDF_EPOCH16, 1L, data,
                          NULL_);
       }
     } else if (var->datatype == CDF_TIME_TT2000) {
       if (numElements == TT2000_3_STRING_LEN) {
         long long tt2000;
         tt2000 = parseTT2000((char *)data);
         status = CDFlib (PUT_, zENTRY_DATA_, CDF_TIME_TT2000, 1L, &tt2000,
                          NULL_);
       } else {
         if (numElements == 1 && datatype == CDF_CHAR)
           *(long long *)data = DEFAULT_TT2000_PADVALUE;
         if (fillvalue && *(long long *)data == NC_FILL_INT64)
           *(long long *)data = (long long) DEFAULT_TT2000_PADVALUE;
         status = CDFlib (PUT_, zENTRY_DATA_, CDF_TIME_TT2000, 1L, data,
                          NULL_);
       }
     } else {
       if (var->datatype == CDF_EPOCH) datatype = CDF_EPOCH;
       else if (var->datatype == CDF_EPOCH16) datatype = CDF_EPOCH16;
       else if (var->datatype == CDF_TIME_TT2000) datatype = CDF_TIME_TT2000;
       status = CDFlib (PUT_, zENTRY_DATA_, datatype, numElements, data,
                        NULL_);
     }
   } else {
     if (fillvalue) {
       if (datatype == CDF_INT1 && *(char *)data == NC_FILL_BYTE)
         *(char *)data = (signed char) DEFAULT_INT1_PADVALUE;
       else if (datatype == CDF_UINT1 &&
                *(unsigned char *)data == NC_FILL_UBYTE)
         *(unsigned char *)data = (unsigned char) DEFAULT_UINT1_PADVALUE;
       else if (datatype == CDF_INT2 && *(short *)data == NC_FILL_SHORT)
         *(short *)data = (short) DEFAULT_INT2_PADVALUE;
       else if (datatype == CDF_UINT2 &&
                *(unsigned short *)data == NC_FILL_USHORT)
         *(unsigned short *)data = (unsigned short) DEFAULT_UINT2_PADVALUE;
       else if (datatype == CDF_INT4 && *(int *)data == NC_FILL_INT)
         *(int *)data = (int) DEFAULT_INT4_PADVALUE;
       else if (datatype == CDF_UINT4 && *(unsigned int *)data == NC_FILL_UINT)
         *(unsigned int *)data = (unsigned int) DEFAULT_UINT4_PADVALUE;
       else if (datatype == CDF_INT8 && *(long long *)data == NC_FILL_INT64)
         *(long long *)data = (long long) DEFAULT_INT8_PADVALUE;
       else if (datatype == CDF_FLOAT && *(float *)data == NC_FILL_FLOAT)
         *(float *)data = (float) DEFAULT_FLOAT_PADVALUE;
       else if (datatype == CDF_DOUBLE && *(double *)data == NC_FILL_DOUBLE)
         *(double *)data = (double) DEFAULT_DOUBLE_PADVALUE;
       status = CDFlib (PUT_, zENTRY_DATA_, datatype, numElements, data,
                        NULL_);
     } else {
       if (toChange == 0) {
         status = CDFlib (PUT_, zENTRY_DATA_, datatype, numElements, data,
                          NULL_);
       } else {
         status = CDFlib (PUT_, zENTRY_DATA_, var->datatype, numElements,
                                              dataNew,
                          NULL_);
         free (dataNew);
       }
     }
     if (status != CDF_OK)
       cdf_status_handler (status, "2... PUT_, zENTRY_DATA_");
   }
}


/*--------------------------------------------------------------------------
 *  Remove the filepath from a file name. *
 *  Example:
 *      /home/data/mydata.txt => mydata.txt
 *--------------------------------------------------------------------------*/
void removeFilepath (char *fileName) {
    char *ptr;

    ptr = (char *) strrchr(fileName, '/');        /* Unix file path separator */
    if (ptr != NULL) {
        ptr++;
        strcpy(fileName, ptr);
    }
    else {
       ptr = (char *) strrchr(fileName, '\\');     /* Windows file separator */
       if (ptr != NULL) {
           ptr++;
           strcpy(fileName, ptr);
       }
    }
}


/*--------------------------------------------------------------------------
 *  Handles a CDF error.
 *--------------------------------------------------------------------------*/
void cdf_status_handler (CDFstatus status, char *source)
{
   char  message[CDF_STATUSTEXT_LEN+1];

   CDFerror (status, message);              /* Get the appropriate message */

   if (status < CDF_WARN) {
       printf ("An error has occurred, halting...\n");
       printf ("%s\n", message);
       printf ("** Error source: %s\n", source);
       exit (1);
   }
   else if (status < CDF_OK) {
       printf ("Warning, function may not have completed as expected...\n");
       printf ("%s\n", message);
   }
   else if (status > CDF_OK && status != DID_NOT_COMPRESS) {
       printf ("Function completed successfully, but be advised that...\n");
       printf ("%s\n", message);
   }
}


/*--------------------------------------------------------------------------
 *  Handles a netCDF error.
 *--------------------------------------------------------------------------*/
void handle_error(int loc, int status) {
  if (NC_ISSYSERR(status)) {
    fprintf (stderr, "System error (%d at %d)... Data conversion is halted "
                     "and the output is either not created or bad..\n",
             status, loc);
    exit (1);
  } else if (status <= -1) {  
    fprintf (stderr, "%s (at %d)... Data conversion is halted and the output "
                     "is either not created or bad..\n", nc_strerror(status),
             loc);
    exit (1);
  }
}


/*--------------------------------------------------------------------------
 *  Handles a memory allocation error.
 *--------------------------------------------------------------------------*/
void memory_error(int i, char *source, char *datatype) {
   printf ("** %d: Memory allocation error occurred in %s. data type = %s\n", 
           i, source, datatype);
   exit (1);
}

void insufficient_memory () {
   printf ("Insufficient memory to load all the records at once\n");
}


/*--------------------------------------------------------------------------*/
void usage()
{
   printf ("\nDescription:\n");
   printf ("    This program converts a netCDF file into a CDF file.\n");
   printf ("    It will exit 0 if the conversion is successful. If the \n");
   printf ("    conversion fails, it exits 1 and the output CDF, if created,\n");
   printf ("    will be bad.\n");
   printf ("\n");
   printf ("Usage: netCDF-to-cdf [OPTION] <netCDF file name>\n");
   printf ("\n");
   printf ("OPTION:\n");
   printf ("  -o <cdffile>     Optionally, assign the output CDF file name. "
           "Without it,\n");
   printf ("                   the CDF will use the netCDF name with \".cdf\" "
           "extension.\n");
   printf ("  -map <mapfile>   Use the supplied mapping file, instead of the "
           "default.\n");
   printf ("                   (located at the same directory as the "
           "translator). If\n");
   printf ("                   the mapping file can't be found, the hard-coded"
           " internal\n");
   printf ("                   mapping (same contents as the default) is used."
           "\n");
   printf ("                   Use \"-debug\" to show its contents.\n");

   printf ("  -epoch <var(s)>  Optionally, convert variable(s) <epoch(s)> to a"
           " CDF's\n");
   printf ("                   epoch data type (CDF_EPOCH, CDF_EPOCH16 or\n");
   printf ("                   CDF_TIME_TT2000) if it is presented as an "
           "equivalent data\n");
   printf ("                   type or string of pre-defined length (likely "
           "from a\n");
   printf ("                   CDF-converted netCDF).\n");
/*   printf ("  -time <var(s)>   Optionally, specify the time variable(s) in the netCDF file.\n"); */
   printf ("  -istpepoch       Optionally, variable(s), having ISTP-compliant "
           "entry for\n");
   printf ("                   epoch attributes \"TIME_BASE\" and \"UNITS\", "
           "will be\n");
   printf ("                   converted to CDF's epoch type: CDF_EPOCH, "
           "CDF_EPOCH16 or\n");
   printf ("                   CDF_TIME_TT2000, based on their values: \n");
   printf ("                     \"TIME_BASE\" = \"0 AD\" and \"UNITS\" = "
           "\"ms\" ==> CDF_EPOCH \n");
   printf ("                     \"TIME_BASE\" = \"0 AD\" and \"UNITS\" = "
           "\"ps\" ==> CDF_EPOCH16 \n");
   printf ("                     \"TIME_BASE\" = \"J2000\" and \"UNITS\" = "
           "\"ns\" ==> \n");
   printf ("                                                             "
           "CDF_TIME_TT2000 \n");
   printf ("  -cdfdepend       Optionally, add netCDF\'s defined dimension(s) "
           "as\n");
   printf ("                   variable(s) and insert an extra \"DEPEND_*\" "
           "attribute(s)\n");
   printf ("                   in the CDF if neither the dimension(s) nor "
           "the \"DEPEND_*\"\n");
   printf ("                   (in upper/lower case) is defined in netCDF. An "
           "extra\n");
   printf ("                   \"NC_ADDED\" attribute is also added to record "
           "the items\n");
   printf ("                   being added to the CDF.\n");
   printf ("  -d <level>       Set the deflate level for variables. The "
           "default is to\n");
   printf ("                   follow the source file's. If the source is not "
           "set, as CDF\n");
   printf ("                   always tries to compress the variable data, "
           "the deflate\n");
   printf ("                   will automatically set to 6 if the variable's "
           "total data\n");
   printf ("                   size is > 1K. Level 0 means no compression.\n");
   printf ("                   [All CDF epoch based data types: CDF_EPOCH, "
           "CDF_EPOCH16\n");
   printf ("                    and CDF_TIME_TT2000 are not compressed, by "
           "default.\n");
   printf ("                    To enforce compression for these epoch data, "
           "add an \"e\"\n");
   printf ("                    to the compression level, e.g., \"-d 6e\".]\n");
   printf ("  -f               Force the conversion to continue when an "
           "unsupported\n");
   printf ("                   data type(s) from netCDF-4 is encountered "
           "(bypassing this\n");
   printf ("                   variable). Without this, the conversion will "
           "be terminated\n");
   printf ("                   immediately and the CDF becomes useless or "
           "gets deleted.\n");
   printf ("  -convertfill     To replace the data defined in the netCDF's "
           "fill\n");
   printf ("                   values by CDF's pad values. Will slow down "
           "the\n");
   printf ("                   program. The default is to use the data as "
           "is.\n");
   printf ("  -nodeletenul     Do not delete the NUL char embedded in NC_CHAR "
           "value\n");
   printf ("                   for global or variable attribute. Sometimes, "
           "this \n");
   printf ("                   option is needed when doing this "
           "netcdf-cdf-netcdf\n");
   printf ("                   loop conversion to restore the file to its "
           "original\n");
   printf ("                   form. The default is to remove it.\n");
   printf ("  -column          Create a column-major CDF and transpose data "
           "from\n");
   printf ("                   row-major based netCDF for all "
           "multi-dimensional\n");
   printf ("                   (> 1) variables. The default cdf is row major "
           "without\n");
   printf ("                   data transpose.\n");
   printf ("  -delete          Optionally, delete the CDF if it already "
           "exists.\n");
   printf ("  -noextra         By default, a set of netCDF specific "
           "information is \n");
   printf ("                   preserved in the CDF as attributes (global and "
           "variable)\n");
   printf ("                   so the CDF can be in the ISTP-compilant form "
           "or/and\n");
   printf ("                   reversed back to netCDF precisely.\n");
   printf ("                   The extra information added in the CDF, "
           "identified as\n");
   printf ("                   \"NC_xxxxx\", includes netCDF's type (kind) and"
           " its\n");
   printf ("                   defined dimensions.\n");
   printf ("                   For netCDF-4 regular and classic files, the CDF"
           " also keeps\n");
   printf ("                   the deflate/chunking information. The preserved"
           " information\n");
   printf ("                   is added as global and variable attributes in "
           "the CDF.\n");
   printf ("                   The added global attributes are:\n");
   printf ("                     NC_KIND: the kind of netCDF file.\n");
   printf ("                     NC_DIMENSIONS_G: netCDF dimension information"
           " (name\n");
   printf ("                                      and length).\n");
   printf ("                     NC_GROUP: the group(s) from a netCDF-4 if it"
           " exists.\n");
   printf ("                     NC_DELIMITER: the delimiter used to set up "
           "the CDF\n");
   printf ("                                   variable name (prepending group"
           " name(s)\n");
   printf ("                                   to NC variable name) if the "
           "netCDF-4\n");
   printf ("                                   has group(s).\n");
   printf ("                   The potential added global attributes are:\n");
   printf ("                     NC_STRINGS_G: keep a record of string global "
           "attributes\n");
   printf ("                     of NC_STRING data type. All others are "
           "NC_CHAR type.\n");
   printf ("                   The potential added variable attributes are:\n");
   printf ("                     NC_STRING: Variable is in NC_STRING form.\n");
   printf ("                     NC_SHUFFLE: Variable's shuffle filter.\n");
   printf ("                     NC_DEFLATE: Variable's deflate level (0-9)."
           "\n");
   printf ("                     NC_STORAGE: Variable's storage property.\n");
   printf ("                     NC_CHUNKSIZES: Variable's array list of "
           "chunk sizes.\n");
   printf ("                     NC_DEPEND: Indicates DEPEND_* attributes/"
           "entries\n");
   printf ("                                were added to the CDF.\n");
   printf ("                     DEPEND_*: Depend dimension(s).\n");
   printf ("                     NC_STRINGS_V: keep a record of string "
           "variable attributes\n");
   printf ("                     of NC_STRING data type. All others are "
           "NC_CHAR type.\n");
   printf ("                   Entries in NC_DEFLATE and NC_CHUNKSIZES are of "
           "CDF_INT4.\n");
   printf ("                   type. All others are of CDF_CHAR type. "
           "NC_SHUFFLE,\n");
   printf ("                   NC_DEFLATE, NC_STORAGE and NC_CHUNKSIZES are "
           "created only\n");
   printf ("                   from netCDF-4 regular or classic files. "
           "NC_STRING is made\n");
   printf ("                   only from netCDF-4 regular file only. If a "
           "variable uses\n");
   printf ("                   contiguous storage, then its chunking sizes "
           "entry will not\n");
   printf ("                   be written as they are irrelevant. This option "
           "indicates\n");
   printf ("                   that no such attributes are to be saved in the "
           "CDF file.\n");
   printf ("  -fixit           Replace the data types for variable "
           "attributes:\n");
   printf ("                   valid_min/valid_max/_FillValue with variable's "
           "if they\n");
   printf ("                   are different. Note: if the original attribute "
           "value is\n");
   printf ("                   outside of the valid range for the variable's "
           "data type,\n");
   printf ("                   the data conversion will stop with an error "
           "message\n");
   printf ("                   and no CDF is created.\n");
   printf ("  -addvartype      Add a variable attribute VAR_TYPE to the CDF. "
           "\n");
   printf ("                   Assign value \"data\" or \"support_data\" to "
           "the entry.\n");
   printf ("  -v               Show the converter and CDF versions.\n");
   printf ("  -debug           Show debug information.\n");
   printf ("\n");
   printf ("Note: 1. The following netCDF-4 types: NC_UINT64, NC_COMPOUND, "
           "NC_VLEN,\n");
   printf ("         NC_OPAQUE, NC_ENUM and NC_FIRSTUSERTYPEID are not "
           "supported.\n");
   printf ("      2. Variables in a group in netCDF-4 are renamed to\n");
   printf ("         groupname(s)_variable form in the CDF. They can be "
           "convered back\n");
   printf ("         to their original form by cdf-to-netcdf converter. \n");
   printf ("      3. Data type conversion for netCDF-3 or netCDF-4 Classic:\n");
   printf ("         NC_BYTE => CDF_INT1\n");
   printf ("         NC_SHORT => CDF_SHORT\n");
   printf ("         NC_INT => CDF_INT4\n");
   printf ("         NC_FLOAT => CDF_FLOAT\n");
   printf ("         NC_DOUBLE => CDF_DOUBLE\n");
   printf ("                      or\n");
   printf ("                   => CDF_EPOCH|CDF_EPOCH16|CDF_TIME_TT2000\n");
   printf ("                      if \"-epoch\" is specified\n");
   printf ("                      or \"-istpepoch\" is specified and "
           "attribute\n");
   printf ("                        \"TIME_BASE\" has an entry of \"0 AD\" or "
           "\"J2000\"\n");
   printf ("         NC_CHAR|NC_STRING => CDF_CHAR\n");
   printf ("                           or\n");
   printf ("                          => CDF_EPOCH|CDF_EPOCH16|"
           "CDF_TIME_TT2000\n");
   printf ("                            if \"-epoch\" is specified\n");
   printf ("                            or \"-istpepoch\" is specified and a "
           "variable's  attribute\n");
   printf ("                            \"TIME_BASE\" has an entry of \"0 "
           "AD\" or \"J2000\"\n");
   printf ("         Data type conversion for netCDF-4:\n");
   printf ("         NC_BYTE => CDF_INT1\n");
   printf ("         NC_UBYTE => CDF_UINT1\n");
   printf ("         NC_SHORT => CDF_SHORT\n");
   printf ("         NC_USHORT => CDF_UINT2\n");
   printf ("         NC_INT => CDF_INT4\n");
   printf ("         NC_UINT => CDF_UINT4\n");
   printf ("         NC_FLOAT => CDF_FLOAT\n");
   printf ("         NC_DOUBLE => CDF_DOUBLE \n");
   printf ("                      or\n");
   printf ("                   => CDF_EPOCH|CDF_EPOCH16\n");
   printf ("                      if \"-epoch\" is specified\n");
   printf ("                      or \"-istpepoch\" is specified and "
           "attribute\n");
   printf ("                         \"TIME_BASE\" has an entry of \"0 "
           "AD\".\n");
   printf ("         NC_INT64 => CDF_INT8\n");
   printf ("                     or\n");
   printf ("                  => CDF_TIME_TT2000\n");
   printf ("                     if \"-epoch\" is specified\n");
   printf ("                     or \"-istpepoch\" is specified and "
           "attribute\n");
   printf ("                        \"TIME_BASE\" has an entry of "
           "\"J2000\".\n");
   printf ("         NC_CHAR|NC_STRING => CDF_CHAR\n");
   printf ("                           or\n");
   printf ("                           => CDF_EPOCH|CDF_EPOCH16|"
           "CDF_TIME_TT2000\n");
   printf ("                              if \"-epoch\" is specified\n");
   printf ("                              or \"-istpepoch\" is specified "
           "and\n");
   printf ("                                 attribute \"TIME_BASE\" has an "
           "entry of\n");
   printf ("                                 \"0 AD\" or \"J2000\" \n");
   printf ("\n");
   printf ("Examples of usage:\n\n");
   printf ("1) netCDF-to-cdf test.nc\n");
   printf ("   Convert test.nc to test.cdf file.\n\n");
   printf ("2) netCDF-to-cdf -map mymap.dat test.nc\n");
   printf ("   Convert test.nc to test.cdf file using\n");
   printf ("   the user-supplied mapping file called 'mymap.dat', instead\n");
   printf ("   of the default netCDF-to-CDF mapping file.\n\n");
   printf ("3) netCDF-to-cdf -o mycdf.cdf -istpepoch source.nc\n");
   printf ("   Convert source.nc to mycdf.cdf. If the netCDF has the "
           "ISTP-compilant\n");
   printf ("   attributes \"TIME_BASE\" and \"UNITS\", they are used to "
           "convert epoch data\n");
   printf ("   into CDF epoch type. Otherwise, specifies the epoch "
           "variable(s)\n");
   printf ("     netCDF-to-cdf -o mycdf.cdf -epoch \"Epoch\" source.nc\n");
   printf ("   This CDF can be reversed back to its original netCDF:\n");
   printf ("     netCDF-to-cdf -o back.nc mycdf.cdf\n\n");
   printf ("\n");
   exit (0);
}

void parseCmdLine (int argc, char *argv[])
{
   int  i;
   char *nextArg;

   for (i=1; i < argc; i++) {
        if (strcmp(argv[i],"-v") == 0) {
            long version, release, increment; char subIncrement;
            CDFgetLibraryVersion (&version, &release, &increment, 
                                  &subIncrement);
            printf("\nnetCDF-to-cdf version: 2.2.0\n");
            printf("  (Based on CDF version %ld.%ld.%ld.%c)\n",version, release,
                   increment, subIncrement);
            printf("  (Based on netCDF version %s)\n",nc_inq_libvers());
            printf("Converter last modified on May 24, 2019.\n");
            printf("Submit problems/questions/suggestions to "
                   "NASA-CDF-Support@nasa.onmicrosoft.com\n\n");
            exit (0);
        }
        else if (strncmpIgCase(argv[i],"-map",4) > 0) {
            nextArg = getNextArgument(argc, argv, i);
/*            if (strcmp(nextArg, "No next argument") == 0 || */
            if (nextArg == NULL ||
                *(nextArg+0) == '-')
                errorMsg("** Error - mapping file is not defined");
            else {
                strcpy(mappingFile, nextArg);
                ++i;
            }
        }
        else if (strncmpIgCase(argv[i],"-o",2) > 0) {
            nextArg = getNextArgument(argc, argv, i);
/*            if (strcmp(nextArg, "No next argument") == 0 || */
            if (nextArg == NULL ||
                *(nextArg+0) == '-')
                errorMsg("** Error - mapping file is not defined");
            else {
                cdfFile = (char *) malloc (MAX_FILE_NAME_LEN+1);
                strcpy(cdfFile, nextArg);
                ++i;
            }
        }
        else if (strcmp(argv[i],"-d") == 0) {
            nextArg = getNextArgument(argc, argv, i);
            if (strlen(nextArg) == 1) {
              if (sscanf (nextArg, "%d", &dLevel) != 1)
                errorMsg("** Error - invalid deflation level");
              if (dLevel < 0 || dLevel > 9)
                errorMsg("** Error - invalid deflation level");
              if (dLevel == 0) compression = FALSE;
            } else {
              if (strlen(nextArg) > 2) 
                errorMsg("** Error - invalid deflation level combination");
              else {
                char epochC[2];
                if (sscanf (nextArg, "%d%s", &dLevel, epochC) != 2)
                  errorMsg("** Error - invalid deflation level");
                if (strcmp(epochC, "e") != 0 && strcmp (epochC, "E") != 0)
                  errorMsg("** Error - invalid deflation for epoch");
                if (dLevel < 0 || dLevel > 9)
                  errorMsg("** Error - invalid deflation level");
                if (dLevel == 0) compression = FALSE;
                else compressionE = TRUE;
              }
            }
        }
        else if (strncmpIgCase(argv[i],"-debug",6) > 0)
            DEBUG = TRUE;
        else if (strncmpIgCase(argv[i],"-delete",7) > 0)
            DELETE = TRUE;
        else if (strncmpIgCase(argv[i],"-column",7) > 0)
            COLUMN = TRUE;
        else if (strncmpIgCase(argv[i],"-istpepoch",10) > 0)
            ISTPepoch = TRUE;
        else if (strncmpIgCase(argv[i],"-cdfdepend",10) > 0)
            ADDdepend = TRUE;
        else if (strncmpIgCase(argv[i],"-addvartype",10) > 0)
            ADDvartype = TRUE;
        else if (strncmpIgCase(argv[i],"-f",2) > 0)
            FORCE = TRUE;
        else if (strncmpIgCase(argv[i],"-fixit",6) > 0)
            ENFORCE = TRUE;
        else if (strncmpIgCase(argv[i],"-epoch",6) > 0) {
            EPOCH = TRUE;
            nextArg = getNextArgument(argc, argv, i);
/*            if (strcmp(nextArg, "No next argument") == 0 || */
            if (nextArg == NULL ||
                *(nextArg+0) == '-')
                errorMsg("** Error - Epoch name is not defined");
            else {
                int j; char tmpName[CDF_VAR_NAME_LEN256+1];
                char *p, *q;
                numEpochs = 1;
                p = nextArg;
                strcpy(tmpName, nextArg);
                for (j = 0; nextArg[j] != (char) '\0'; j++) {
                  if (nextArg[j] == (char) ',') ++numEpochs;
                }
                EPOCHnames = (char **) malloc (sizeof (char *) * numEpochs);
                for (j = 0; j < numEpochs; ++j) {
                  EPOCHnames[j] = malloc (sizeof(char) *
                                          (CDF_VAR_NAME_LEN256+1));
                  q = strchr (p, ',');
                  if (q != NULL) {
                    strncpy(EPOCHnames[j], p, q-p);
                    *(EPOCHnames[j]+(q-p)) = '\0';
                    p = q + 1;
                  } else
                    strcpy(EPOCHnames[j], p);
                }
                ++i;
            }
        } else if (strncmpIgCase(argv[i],"-time",5) > 0) {
            TIME = TRUE;
            nextArg = getNextArgument(argc, argv, i);
/*            if (strcmp(nextArg, "No next argument") == 0 || */
            if (nextArg == NULL ||
                *(nextArg+0) == '-')
                errorMsg("** Error - Epoch name is not defined");
            else {
                int i; char tmpName[CDF_VAR_NAME_LEN256+1];
                char *p, *q;
                numTimes = 1;
                p = nextArg;
                strcpy(tmpName, nextArg);
                for (i = 0; nextArg[i] != (char) '\0'; i++) {
                  if (nextArg[i] == (char) ',') ++numTimes;
                }
                TIMEnames = (char **) malloc (sizeof (char *) * numTimes);
                for (i = 0; i < numTimes; ++i) {
                  TIMEnames[i] = malloc (sizeof(char) *
                                         (CDF_VAR_NAME_LEN256+1));
                  q = strchr (p, ',');
                  if (q != NULL) {
                    strncpy(TIMEnames[i], p, q-p);
                    p = q + 1;
                  } else
                    strcpy(TIMEnames[i], p);
                }
                ++i;
            }
        } else if (strncmpIgCase(argv[i],"-convertfill",12) > 0)
            fillvalue = TRUE;
        else if (strncmpIgCase(argv[i],"-nodeletenul",12) > 0)
            deleteNULL = FALSE;
        else if (strncmpIgCase(argv[i],"-noextra",8) > 0)
            NOEXTRA = TRUE;
   }
   if (EPOCH && ISTPepoch)
      errorMsg("** Error - \"-epoch\" and \"-istpepoch\" can't be specified "
               "both");
   if (NOEXTRA && (ADDdepend || ADDvartype))
      errorMsg("** Error - \"-noextra\" and \"-adddepend\" or \"-addvartype\" "
               "can't be specified both");
}


char * getNextArgument (int argc, char *argv[], int currentIndex)
{
    char *nextArg;
    int  nextIndex;

    nextIndex  = currentIndex+1;
    if (nextIndex == argc) {
/*        strcpy (msg, "No next argument"); */
/*        nextArg = msg; */
	nextArg = NULL;
    }
    else
        nextArg = argv[nextIndex];

    return nextArg;
}

void errorMsg (char *msg)
{
   printf ("%s\n", msg);
   exit (1);
}


void map_netCDF_attr_to_CDF (char *searchAttr,      /* in - netCDF attr name */
                             char *mappedAttrName)  /* out - CDF attr name   */
{
   int i;

   strcpy(mappedAttrName, searchAttr);

   for (i=0; i < totAttrsInMappingFile; i++) {
        if (strcmpIgCase(netcdfAttrNames[i], searchAttr) == 1) {
            strcpy(mappedAttrName, cdfAttrNames[i]);
            break;
        }          
   }
}


/*--------------------------------------------------------------------------
 *  Remove the file extension from the given file name.
 *
 *  Example:
 *      mydata.txt => mydata
 *--------------------------------------------------------------------------*/
char * removeFileExtension (char *fileName, char *fileExtension) {
    char *ptr;

    ptr = (char *) strlaststr(fileName, fileExtension);
    if (ptr != NULL) *ptr = '\0';

   return ptr;
}

/*-----------------------------------------------------------------------------
 *  Find the last occurrence of s2 in s1.  If s21 is not found, s1 is returned.
 *----------------------------------------------------------------------------*/
char * strlaststr (char *s1, char *s2)
{
    char *sc2, *psc1, *ps1;

    if (*s2 == '\0')
        return((char *)s1);

    ps1 = s1 + strlen(s1);

    while(ps1 != s1) {
        --ps1;
        for (psc1 = ps1, sc2 = s2; ; )
                if (*(psc1++) != *(sc2++))
                        break;
                else if (*sc2 == '\0')
                        return ((char *)ps1);
    }
    return ((char *)NULL);
}

/*-----------------------------------------------------------------------------
 *  Find if the string is one of the "FILLVAL", "VALIDMIN" or "VALIDMAX". If
 *  so, return a number of 1 for "FILLVAL", 2 for "VALIDMIN", 3 for 
 *  "VALIDMAX", 4 for "scalemin" and 5 for "scalemax". Otherwise, return 0.
 *----------------------------------------------------------------------------*/
int findSpecialAttributes (char *s)
{

    if (strncmpIgCase(s, "fillval", 7) > 0) return 1;
    else if (strncmpIgCase(s, "validmin", 8) > 0) return 2;
    else if (strncmpIgCase(s, "validmax", 8) > 0) return 3;
    else if (strncmpIgCase(s, "scalemin", 8) > 0) return 4;
    else if (strncmpIgCase(s, "scalemax", 8) > 0) return 5;
    else return 0;
}

/*-----------------------------------------------------------------------------
 *  Find the CDF Epoch type, CDF_EPOCH, CDF_EPOCH16 or CDF_TIME_TT2000, based
 *  on the encoded epoch string.
 *----------------------------------------------------------------------------*/
long findEpochType (char *s)
{
    int len = (int) strlen (s);
    if (len == EPOCH_STRING_LEN) return (long) CDF_EPOCH;
    else if (len == EPOCH16_STRING_LEN) return (long) CDF_EPOCH16;
    else if (len == TT2000_3_STRING_LEN) return (long) CDF_TIME_TT2000;
    else return (long) 0;
}

/******************************************************************************
* strncmpIgCase... Assuming count is the length from one of passed strings.
*                  It compares characters upto the passed count.
* It returns  1 if compared strings are identical with same length
*             0 if compared (sub)strings are not same
******************************************************************************/

int strncmpIgCase (string1, string2, count)
char *string1;
char *string2;
size_t count;
{
  int string1L = (int) strlen(string1);
  int string2L = (int) strlen(string2);
  int maxL, i;
  maxL = MINIMUM(string2L,(int)count);
  maxL = MINIMUM(string1L,maxL);
  for (i = 0; i < maxL; i++) {
     if (i == string1L || i == string2L) break;
     if (MakeLower(string1[i]) != MakeLower(string2[i])) return 0;
  }
  if (string1L == string2L) return 1;
  else return 0;
}

/******************************************************************************
 * findIndex... Find the index number from an attribute name, which has the
 * pattern of XXXXX_nnn where XXXXX is the actual attribute name and nnn is
 * the entry number. A netCDF converted from CDF may have global attributes of
 * this pattern. CDF allows multiple entries for an attribute, but netCDF does
 * not. So, a global attribute with multiple entries will be converted into
 * multiple attributes of XXXXX, XXXXX_2, XXXXX_3, etc. While converting the
 * netCDF back to CDF, we consolidate these attributes, if we can find them.
 *
 * It returns  1 if the attribute name has a such pattern.
 *             0 otherwise
 ******************************************************************************/

int findIndex (char *string, int *idx)
{
  char *ptr;
  int count = 0;

  ptr = strrchr(string, (char)*delimiters[curDiD]);
  if (ptr != NULL) {
    count = sscanf (ptr+1, "%d",  idx);
  }
  return count;
}

void findAttrName (char *string, char *string2)
{
  char *ptr;

  ptr = strrchr(string, (char)*delimiters[curDiD]);
  if (ptr == NULL) strcpy (string2, string);
  else {
    int count, idx;
    int len1, len2;
    count = sscanf (ptr+1, "%d",  &idx);
    if (count == 1) {
      len1 = strlen (string);
      len2 = strlen (ptr);
      strncpy (string2, string, len1-len2);
      return;
    }
  }
}

/* Return file type string */
char *kind_string(int kind)
{
    switch (kind) {
    case NC_FORMAT_CLASSIC:
        return "classic";
    case NC_FORMAT_64BIT:
        return "64-bit offset";
    case NC_FORMAT_NETCDF4:
        return "netCDF-4";
    case NC_FORMAT_NETCDF4_CLASSIC:
        return "netCDF-4 classic model";
    default:
        printf ("unrecognized file format: %d", kind);
        return "unrecognized";
    }
}

static int FindEpoch (char *name)
{
    int i;
    for (i = 0; i < numEpochs; ++i) {
      if (strncmpIgCase(name, EPOCHnames[i], (int)strlen(EPOCHnames[i])) == 1)
        return 1;
    }
    return 0;
}

static int FindTime (char *name)
{
    int i;
    for (i = 0; i < numTimes; ++i) {
      if (strncmpIgCase(name, TIMEnames[i], (int)strlen(TIMEnames[i])) == 1)
        return 1;
    }
    return 0;
}

/******************************************************************************
 * strcmpIgCase.
 * It returns  1 if compared strings are identical with the same length
 *             0 if compared (sub)strings are not same 
 ******************************************************************************/

int strcmpIgCase (string1, string2)
char *string1;
char *string2;
{
  int string1L = strlen(string1);
  int string2L = strlen(string2);
  int i;
  if (string1L != string2L) return 0;
  for (i = 0; i < string1L; i++) {
     if (MakeLower(string1[i]) != MakeLower(string2[i])) return 0;
  }
  return 1;
}

/******************************************************************************
 * FindUnlimitedDim.
 * It returns  >= 0 if the dimid is a unlimited dimension in this netcdf file
 *             -1 if a non_unlimited dimension
 ******************************************************************************/
int FindUnlimitedDim (dimid)
int dimid;
{
  int i;

  for (i = 0; i < totUnlims; ++i) {
    if (dimid == unlimitedDims[i]) return i;
  }
  return -1;
}

/******************************************************************************
 * CollectUnlimitedDims.
 * This function searches thru the entire netcdf for all unlimited dimensions
 * and their current sizes. If id has a sub-group(s), it will go down the 
 * sub-group(s).
 ******************************************************************************/

void CollectUnlimitedDims (ncid)
int ncid;
{
  int i, status, ngrps, numUnlims, *grpids;

  grpids = NULL;
  status = nc_inq_unlimdims (ncid, &numUnlims, NULL);
  if (status != NC_NOERR) handle_error(500, status);
  if (numUnlims > 0) {
    if (totUnlims == 0) {
      unlimitedDims = (int *) malloc ((size_t) sizeof(int) * numUnlims);
      totRecsX = (size_t *) malloc ((size_t) sizeof(size_t) * numUnlims);
    } else {
      unlimitedDims = (int *) realloc (unlimitedDims, (size_t) sizeof(int) *
                                       (totUnlims + numUnlims));
      totRecsX = (size_t *) realloc (totRecsX, (size_t) sizeof(size_t) *
                                     (totUnlims + numUnlims));
    }
    status = nc_inq_unlimdims (ncid, &numUnlims, &(unlimitedDims[totUnlims]));
    if (status != NC_NOERR) handle_error(501, status);
    for (i = 0; i < numUnlims; ++i) {
      status = nc_inq_dimlen(ncid, unlimitedDims[totUnlims+i],
                             &(totRecsX[totUnlims+i]));
      if (status != NC_NOERR && status != NC_EBADDIM) handle_error(502, status);
    }
    totUnlims += numUnlims;
  }
  status = nc_inq_grps (ncid, &ngrps, NULL);
  if (status != NC_NOERR) handle_error(503, status);
  grpids = NULL;
  if (ngrps > 0) {
    grpids = (int *) malloc ((size_t) sizeof(int) * numUnlims);
    status = nc_inq_grps (ncid, NULL, grpids);
    if (status != NC_ENOGRP) handle_error(504, status);
    for (i = 0; i < ngrps; ++i) {
      CollectUnlimitedDims (grpids[i]);
    }
  }
  if (grpids != NULL) free (grpids);
}

/******************************************************************************
 * ConvertData.
 * This function converts the data from their original data type to the
 * detstination data type. It everything is converted fine, it returns 1.
 * Otherwise, 0 is returned.
 ******************************************************************************/

int ConvertData (odata, odatatype, idata, idatatype, num)
void *odata;
long odatatype;
void *idata;
int  idatatype;
int  num;
{
  int i;
  
  if (odatatype == CDF_INT1) {
      if (idatatype == NC_BYTE) {
         memcpy(odata, idata, num);
      } else if (idatatype == NC_UBYTE) {
         for (i = 0; i < num; ++i) {
           if ((int)*(((unsigned char *)idata)+i) > 127) return 0;
           *(((signed char *)odata)+i) = *(((unsigned char *)idata)+i);
         }
      } else if (idatatype == NC_SHORT) {
         for (i = 0; i < num; ++i) {
           if (*(((short *)idata)+i) > 127 ||
               *(((short *)idata)+i) < -128) return 0;
           *(((signed char *)odata)+i) = *(((short *)idata)+i);
         }
      } else if (idatatype == NC_USHORT) {
         for (i = 0; i < num; ++i) {
           if (*(((unsigned short *)idata)+i) > 127) return 0;
           *(((signed char *)odata)+i) = *(((unsigned short *)idata)+i);
         }
      } else if (idatatype == NC_INT || idatatype == NC_LONG) {
         for (i = 0; i < num; ++i) {
           if (*(((int *)idata)+i) > 127 ||
               *(((int *)idata)+i) < -128) return 0;
           *(((signed char *)odata)+i) = *(((int *)idata)+i);
         }
      } else if (idatatype == NC_UINT) {
         for (i = 0; i < num; ++i) {
           if (*(((unsigned int *)idata)+i) > 127) return 0;
           *(((signed char *)odata)+i) = *(((unsigned int *)idata)+i);
         }
      } else if (idatatype == NC_INT64) {
         for (i = 0; i < num; ++i) {
           if (*(((long long *)idata)+i) > 127 ||
               *(((long long *)idata)+i) < -128) return 0;
           *(((signed char *)odata)+i) = *(((long long *)idata)+i);
         }
      } else if (idatatype == NC_FLOAT) {
         for (i = 0; i < num; ++i) {
           if ((int)*(((float *)idata)+i) > 127 ||
               (int)*(((float *)idata)+i) < -128) return 0;
           *(((signed char *)odata)+i) = (signed char) *(((float *)idata)+i);
         }
      } else if (idatatype == NC_DOUBLE) {
         for (i = 0; i < num; ++i) {
           if ((int)*(((double *)idata)+i) > 127 ||
               (int)*(((double *)idata)+i) < -128) return 0;
           *(((signed char *)odata)+i) = (signed char) *(((double *)idata)+i);
         }
      }
  } else if (odatatype == CDF_UINT1) {
      if (idatatype == NC_BYTE) {
         for (i = 0; i < num; ++i) {
           if (*(((signed char *)idata)+i) < 0) return 0;
           *(((unsigned char *)odata)+i) = (unsigned char ) *(((signed char *)idata)+i);
         }
      } else if (idatatype == NC_UBYTE) {
         memcpy(odata, idata, num);
      } else if (idatatype == NC_SHORT) {
         for (i = 0; i < num; ++i) {
           if (*(((short *)idata)+i) > 255 ||
               *(((short *)idata)+i) < 0) return 0;
           *(((unsigned char *)odata)+i) = *(((short *)idata)+i);
         }
      } else if (idatatype == NC_USHORT) {
         for (i = 0; i < num; ++i) {
           if (*(((unsigned short *)idata)+i) > 255) return 0;
           *(((unsigned char *)odata)+i) = *(((unsigned short *)idata)+i);
         }
      } else if (idatatype == NC_INT || idatatype == NC_LONG) {
         for (i = 0; i < num; ++i) {
           if (*(((int *)idata)+i) > 255 ||
               *(((int *)idata)+i) < 0) return 0;
           *(((unsigned char *)odata)+i) = (unsigned char) *(((int *)idata)+i);
         }
      } else if (idatatype == NC_UINT) {
         for (i = 0; i < num; ++i) {
           if (*(((unsigned int *)idata)+i) > 255) return 0;
           *(((unsigned char *)odata)+i) = 
                                 (unsigned char) *(((unsigned int *)idata)+i);
         }
      } else if (idatatype == NC_INT64) {
         for (i = 0; i < num; ++i) {
           if (*(((long long *)idata)+i) > 255 ||
               *(((long long *)idata)+i) < 0) return 0;
           *(((unsigned char *)odata)+i) = 
                                 (unsigned char) *(((long long *)idata)+i);
         }
      } else if (idatatype == NC_FLOAT) {
         for (i = 0; i < num; ++i) {
           if ((int)*(((float *)idata)+i) > 255 ||
               (int)*(((float *)idata)+i) < 0) return 0;
           *(((unsigned char *)odata)+i) = 
                                 (unsigned char) *(((float *)idata)+i);
         }
      } else if (idatatype == NC_DOUBLE) {
         for (i = 0; i < num; ++i) {
           if ((int)*(((double *)idata)+i) > 255 ||
               (int)*(((double *)idata)+i) < 0) return 0;
           *(((unsigned char *)odata)+i) = 
                                 (unsigned char) *(((double *)idata)+i);       
         }
      }
  } else if (odatatype == CDF_INT2) {
      if (idatatype == NC_BYTE) {
         for (i = 0; i < num; ++i) { 
           *(((short *)odata)+i) = (short) *(((signed char *)idata)+i);
         }
      } else if (idatatype == NC_UBYTE) {
         for (i = 0; i < num; ++i) { 
           *(((short *)odata)+i) = (short) *(((unsigned char *)idata)+i);
         }
      } else if (idatatype == NC_SHORT) {
         memcpy(odata, idata, 2*num);
      } else if (idatatype == NC_USHORT) {
         for (i = 0; i < num; ++i) {
           if (*(((unsigned short *)idata)+i) > 32767) return 0;
           *(((short *)odata)+i) = (short) *(((unsigned short *)idata)+i);
         }
      } else if (idatatype == NC_INT || idatatype == NC_LONG) {
         for (i = 0; i < num; ++i) {
           if (*(((int *)idata)+i) > 32767 ||
               *(((int *)idata)+i) < -32768) return 0;
           *(((short *)odata)+i) = (short) *(((int *)idata)+i);
         }
      } else if (idatatype == NC_UINT) {
         for (i = 0; i < num; ++i) {
           if (*(((unsigned int *)idata)+i) > 32767) return 0;
           *(((short *)odata)+i) = (short) *(((unsigned int *)idata)+i);
         }
      } else if (idatatype == NC_INT64) {
         for (i = 0; i < num; ++i) {
           if (*(((long long *)idata)+i) > 32767 ||
               *(((long long *)idata)+i) < -32768) return 0;
           *(((short *)odata)+i) = (short) *(((long long *)idata)+i);
         }
      } else if (idatatype == NC_FLOAT) {
         for (i = 0; i < num; ++i) {
           if ((int)*(((float *)idata)+i) > 32767 ||
               (int)*(((float *)idata)+i) < -32768) return 0;
           *(((short *)odata)+i) = (short) *(((float *)idata)+i);
         }
      } else if (idatatype == NC_DOUBLE) {
         for (i = 0; i < num; ++i) {
           if ((int)*(((double *)idata)+i) > 32767 ||
               (int)*(((double *)idata)+i) < -32768) return 0;
           *(((short *)odata)+i) = (short) *(((double *)idata)+i);
         }
      }
  } else if (odatatype == CDF_UINT2) {
      if (idatatype == NC_BYTE) {
         for (i = 0; i < num; ++i) { 
           *(((unsigned short *)odata)+i) = 
                                 (unsigned short) *(((signed char *)idata)+i);
         }
      } else if (idatatype == NC_UBYTE) {
         for (i = 0; i < num; ++i) { 
           *(((unsigned short *)odata)+i) = 
                                 (unsigned short) *(((unsigned char *)idata)+i);
         }
      } else if (idatatype == NC_SHORT) {
         for (i = 0; i < num; ++i) {
           if (*(((short *)idata)+i) < 0) return 0;
           *(((unsigned short *)odata)+i) = 
                                 (unsigned short) *(((short *)idata)+i);
         }
      } else if (idatatype == NC_USHORT) {
         memcpy(odata, idata, 2*num);
      } else if (idatatype == NC_INT || idatatype == NC_LONG) {
         for (i = 0; i < num; ++i) {
           if (*(((int *)idata)+i) > 65535 || 
               *(((int *)idata)+i) < 0) return 0;
           *(((unsigned short *)odata)+i) = (unsigned short) *(((int *)idata)+i);
         }
      } else if (idatatype == NC_UINT) {
         for (i = 0; i < num; ++i) {
           if (*(((unsigned int *)idata)+i) > 65535) return 0;
           *(((unsigned short *)odata)+i) = 
                                 (unsigned short) *(((unsigned int *)idata)+i);
         }
      } else if (idatatype == NC_INT64) {
         for (i = 0; i < num; ++i) {
           if (*(((long long *)idata)+i) > 65535 ||
               *(((long long *)idata)+i) < 0) return 0;
           *(((unsigned short *)odata)+i) = 
                                 (unsigned short) *(((long long *)idata)+i);
         }
      } else if (idatatype == NC_FLOAT) {
         for (i = 0; i < num; ++i) {
           if ((int)*(((float *)idata)+i) > 65535 || 
               (int)*(((float *)idata)+i) < 0) return 0;
           *(((unsigned short *)odata)+i) = 
                                 (unsigned short) *(((float *)idata)+i);
         }
      } else if (idatatype == NC_DOUBLE) {
         for (i = 0; i < num; ++i) {
           if ((int)*(((double *)idata)+i) > 65535 ||
               (int)*(((double *)idata)+i) < 0) return 0;
           *(((unsigned short *)odata)+i) = 
                                 (unsigned short) *(((double *)idata)+i);
         }
      }
  } else if (odatatype == CDF_INT4) {
      if (idatatype == NC_BYTE) {
         for (i = 0; i < num; ++i) {
           *(((int *)odata)+i) = (int) *(((signed char *)idata)+i);
         }
      } else if (idatatype == NC_UBYTE) {
         for (i = 0; i < num; ++i) {
           *(((int *)odata)+i) = (int) *(((unsigned char *)idata)+i);
         }
      } else if (idatatype == NC_SHORT) {
         for (i = 0; i < num; ++i) {
           if (*(((short *)idata)+i) < 0) return 0;
           *(((int *)odata)+i) = (int) *(((short *)idata)+i);
         }
      } else if (idatatype == NC_USHORT) {
         for (i = 0; i < num; ++i) {
           *(((int *)odata)+i) = (int) *(((unsigned short *)idata)+i);
         }
      } else if (idatatype == NC_INT || idatatype == NC_LONG) {
         memcpy(odata, idata, 4*num);
      } else if (idatatype == NC_UINT) {
         for (i = 0; i < num; ++i) {
           if ((long long)*(((unsigned int *)idata)+i) > 2147483647L) return 0;
           *(((int *)odata)+i) = (int) *(((unsigned int *)idata)+i);
         }
      } else if (idatatype == NC_INT64) {
         for (i = 0; i < num; ++i) {
           if (*(((long long *)idata)+i) > 2147483647L ||
               *(((long long *)idata)+i) < (-2147483647L-1)) return 0;
           *(((int *)odata)+i) = (int) *(((long long *)idata)+i);
         }
      } else if (idatatype == NC_FLOAT) {
         for (i = 0; i < num; ++i) {
           if ((long long)*(((float *)idata)+i) > 2147483647 ||
               (long long)*(((float *)idata)+i) < (-2147483647L-1)) return 0;
           *(((int *)odata)+i) = (int) *(((float *)idata)+i);
         }
      } else if (idatatype == NC_DOUBLE) {
         for (i = 0; i < num; ++i) {
           if ((long long)*(((double *)idata)+i) > 2147483647L ||
               (long long)*(((double *)idata)+i) < (-2147483647L-1)) return 0;
           *(((int *)odata)+i) = (int) *(((double *)idata)+i);
         }
      }
  } else if (odatatype == CDF_UINT4) {
      if (idatatype == NC_BYTE) {
         for (i = 0; i < num; ++i) {
           *(((unsigned int *)odata)+i) = 
                                 (unsigned int) *(((signed char *)idata)+i);
         }
      } else if (idatatype == NC_UBYTE) {
         for (i = 0; i < num; ++i) {
           *(((unsigned int *)odata)+i) = 
                                 (unsigned int) *(((unsigned char *)idata)+i);
         }
      } else if (idatatype == NC_SHORT) {
         for (i = 0; i < num; ++i) {
           if (*(((short *)idata)+i) < 0) return 0;
           *(((unsigned int *)odata)+i) = (unsigned int) *(((short *)idata)+i);
         }
      } else if (idatatype == NC_USHORT) {
         for (i = 0; i < num; ++i) { 
           *(((unsigned int *)odata)+i) = 
                                 (unsigned int) *(((unsigned short *)idata)+i);
         }
      } else if (idatatype == NC_INT || idatatype == NC_LONG) {
         for (i = 0; i < num; ++i) {
           if (*(((int *)idata)+i) < 0) return 0;
           *(((unsigned int *)odata)+i) = (unsigned int) *(((int *)idata)+i);
         }
      } else if (idatatype == NC_UINT) {
         memcpy(odata, idata, 4*num);
      } else if (idatatype == NC_INT64) {
         for (i = 0; i < num; ++i) {
           if (*(((long long *)idata)+i) > 2147483647L ||
               *(((long long *)idata)+i) < (-2147483647L-1)) return 0;
           *(((unsigned int *)odata)+i) = 
                                 (unsigned int) *(((long long *)idata)+i);
         }
      } else if (idatatype == NC_FLOAT) {
         for (i = 0; i < num; ++i) {
           if ((long long)*(((float *)idata)+i) > 2147483647L ||
               (long long)*(((float *)idata)+i) < (-2147483647L-1)) return 0;
           *(((unsigned int *)odata)+i) = (unsigned int) *(((float *)idata)+i);
         }
      } else if (idatatype == NC_DOUBLE) {
         for (i = 0; i < num; ++i) {
           if ((long long)*(((double *)idata)+i) > 2147483647L ||
               (long long)*(((double *)idata)+i) < (-2147483647L-1)) return 0;
           *(((unsigned int *)odata)+i) = (unsigned int) *(((double *)idata)+i);
         }
      }
  } else if (odatatype == CDF_INT8) {
      if (idatatype == NC_BYTE) {
         for (i = 0; i < num; ++i) {
           *(((long long *)odata)+i) = (long long) *(((signed char *)idata)+i);
         }
      } else if (idatatype == NC_UBYTE) {
         for (i = 0; i < num; ++i) {
           *(((long long *)odata)+i) = 
                                 (long long) *(((unsigned char *)idata)+i);
         }
      } else if (idatatype == NC_SHORT) {
         for (i = 0; i < num; ++i) {
           *(((long long *)odata)+i) = (long long) *(((short *)idata)+i);
         }
      } else if (idatatype == NC_USHORT) {
         for (i = 0; i < num; ++i) {
           *(((long long *)odata)+i) = 
                                 (long long) *(((unsigned short *)idata)+i);
         }
      } else if (idatatype == NC_INT || idatatype == NC_LONG) {
         for (i = 0; i < num; ++i) {
           *(((long long *)odata)+i) = (long long) *(((int *)idata)+i);
         }
      } else if (idatatype == NC_UINT) {
         for (i = 0; i < num; ++i) {
           *(((long long *)odata)+i) = (long long) *(((unsigned int *)idata)+i);
         }
      } else if (idatatype == NC_INT64) {
         memcpy(odata, idata, 8*num);
      } else if (idatatype == NC_FLOAT) {
         for (i = 0; i < num; ++i) {
           if ((long long)*(((float *)idata)+i) > 2147483647 ||
               (long long)*(((float *)idata)+i) < (-2147483647L-1)) return 0;
           *(((long long *)odata)+i) = (long long) *(((float *)idata)+i);
         }
      } else if (idatatype == NC_DOUBLE) {
         for (i = 0; i < num; ++i) {
           if ((long long)*(((double *)idata)+i) > 2147483647L ||
               (long long)*(((double *)idata)+i) < (-2147483647L-1)) return 0;
           *(((long long *)odata)+i) = (long long) *(((double *)idata)+i);
         }
      }
  } else if (odatatype == CDF_FLOAT) {
      if (idatatype == NC_BYTE) {
         for (i = 0; i < num; ++i) {
           *(((float *)odata)+i) = (float) *(((signed char *)idata)+i);
         }
      } else if (idatatype == NC_UBYTE) {
         for (i = 0; i < num; ++i) {
           *(((float *)odata)+i) = (float) *(((unsigned char *)idata)+i);
         }
      } else if (idatatype == NC_SHORT) {
         for (i = 0; i < num; ++i) {
           *(((float *)odata)+i) = (float) *(((short *)idata)+i);
         }
      } else if (idatatype == NC_USHORT) {
         for (i = 0; i < num; ++i) {
           *(((float *)odata)+i) = (float) *(((unsigned short *)idata)+i);
         }
      } else if (idatatype == NC_INT || idatatype == NC_LONG) {
         for (i = 0; i < num; ++i) {
           *(((float *)odata)+i) = (float) *(((int *)idata)+i);
         }
      } else if (idatatype == NC_UINT) {
         for (i = 0; i < num; ++i) { 
           *(((float *)odata)+i) = (float) *(((unsigned int *)idata)+i);
         }
      } else if (idatatype == NC_INT64) {
         for (i = 0; i < num; ++i) {
           *(((float *)odata)+i) = (float) *(((long long *)idata)+i);
         }
      } else if (idatatype == NC_FLOAT) {
         memcpy(odata, idata, 4*num);
      } else if (idatatype == NC_DOUBLE) {
         for (i = 0; i < num; ++i) {
           *(((float *)odata)+i) = (float) *(((double *)idata)+i);
         }
      }
  } else if (odatatype == CDF_DOUBLE) {
      if (idatatype == NC_BYTE) {
         for (i = 0; i < num; ++i) {            
           *(((double *)odata)+i) = (double) *(((signed char *)idata)+i);
         }
      } else if (idatatype == NC_UBYTE) {
         for (i = 0; i < num; ++i) {
           *(((double *)odata)+i) = (double) *(((unsigned char *)idata)+i);
         }
      } else if (idatatype == NC_SHORT) {
         for (i = 0; i < num; ++i) {
           *(((double *)odata)+i) = (double) *(((short *)idata)+i);
         }
      } else if (idatatype == NC_USHORT) {
         for (i = 0; i < num; ++i) {
           *(((double *)odata)+i) = (double) *(((unsigned short *)idata)+i);
         }
      } else if (idatatype == NC_INT || idatatype == NC_LONG) {
         for (i = 0; i < num; ++i) {
           *(((double *)odata)+i) = (double) *(((int *)idata)+i);
         }
      } else if (idatatype == NC_UINT) {
         for (i = 0; i < num; ++i) {
           *(((double *)odata)+i) = (double) *(((unsigned int *)idata)+i);
         }
      } else if (idatatype == NC_INT64) {
         for (i = 0; i < num; ++i) {
           *(((double *)odata)+i) = (double) *(((long long *)idata)+i);
         }
      } else if (idatatype == NC_FLOAT) {
         for (i = 0; i < num; ++i) {
           *(((double *)odata)+i) = (double) *(((float *)idata)+i);
         }
      } else if (idatatype == NC_DOUBLE) {
         memcpy(odata, idata, 8*num);
      }
  }
  return 1;
}

int getFormatAttributeLength (int ncid, int varId) {
   int found, ncStatus;
   nc_type type;
   char **strings;
   char *string;
   size_t len, lenx;

   found = 0;
   len = 0;
   strings = (char **)malloc (sizeof(char *) * 1 /* len */);
   ncStatus = nc_inq_atttype(ncid, varId, "Format", &type);
   if (ncStatus == NC_NOERR) {
     if (type == NC_CHAR) {
       ncStatus = nc_inq_attlen(ncid, varId, "Format", &len);
       string = malloc(sizeof(char)*(len+1)); 
       ncStatus = nc_get_att_text(ncid, varId, "Format", string);
       *(string+len) = '\0';
     } else {
       ncStatus = nc_get_att_string(ncid, varId, "Format", strings);
       len = strlen(strings[0]);
       string = (char *) malloc (sizeof(char) * (len+1));
       strcpy (string, strings[0]);
       nc_free_string (1, strings);
     }
     found = 1;
   } else {
     ncStatus = nc_inq_atttype(ncid, varId, "FORMAT", &type);
     if (ncStatus == NC_NOERR) {
       if (type == NC_CHAR) {
         ncStatus = nc_inq_attlen(ncid, varId, "FORMAT", &len);
         string = malloc(sizeof(char)*(len+1));
         ncStatus = nc_get_att_text(ncid, varId, "FORMAT", string);
         *(string+len) = '\0';
       } else {
         ncStatus = nc_get_att_string(ncid, varId, "FORMAT", strings);
         len = strlen(strings[0]);
         string = (char *) malloc (sizeof(char) * (len+1));
         strcpy (string, strings[0]);
         nc_free_string (1, strings);
       }
       found = 1;
     } else {
       ncStatus = nc_inq_atttype(ncid, varId, "format", &type);
       if (ncStatus == NC_NOERR) {
         if (type == NC_CHAR) {
           ncStatus = nc_inq_attlen(ncid, varId, "format", &len);
           string = malloc(sizeof(char)*(len+1));
           ncStatus = nc_get_att_text(ncid, varId, "format", string);
           *(string+len) = '\0';
         } else {
           ncStatus = nc_get_att_string(ncid, varId, "format", strings);
           len = strlen(strings[0]);
           string = (char *) malloc (sizeof(char) * (len+1));
           strcpy (string, strings[0]);
           nc_free_string (1, strings);
         }
         found = 1;
       } 
     }
   }
   if (found == 1) {
     char *cPtr1;
     int ii;
     cPtr1 = string;
     for (ii =0; ii < (int)len; ++ii) {
       if (Decimal((int)*(cPtr1+ii))) break;
       ++cPtr1;
     }
     if (ii < (int) len) {
       int leny;
       if (sscanf (cPtr1, "%d", &leny) != 1)
         lenx = -1;
       lenx = (size_t) leny;
     } else
         lenx = -1;
     free (string);
   } else
     lenx = 100;
   return lenx;
}

/******************************************************************************
 * add_cdf_dimension_extra.
 * This function adds a few extra attributes/entries to the added variables
 * that represent the dimensions from the netCDF.
 ******************************************************************************/

void add_cdf_dimension_extra (long varNum, char *dimName) {
   long attrNum, dataType, numElems;
   CDFstatus status;
   void *value;

   status = CDFlib (SELECT_, zVAR_, varNum,
                    GET_, zVAR_DATATYPE_, &dataType,
                          zVAR_NUMELEMS_, &numElems,
                    NULL_);
   if (status != CDF_OK) {
     cdf_status_handler (status, "010...GET_, DATATYPE");
     return;
   }
   status = CDFlib (CREATE_, ATTR_, "FILLVAL", VARIABLE_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS) {
     cdf_status_handler (status, "010...CREATE_, ATTR_");
     return;
   }
   status = CDFlib (SELECT_, ATTR_NAME_, "FILLVAL",
                    CONFIRM_, zENTRY_EXISTENCE_, varNum,
                    NULL_);
   if (status != CDF_OK) {
     value = (void *) malloc ((size_t) CDFelemSize (dataType));
     DefaultPadValue (dataType, numElems, value);
     status = CDFlib (SELECT_, ATTR_NAME_, "FILLVAL",
                               zENTRY_, varNum,
                      PUT_, zENTRY_DATA_, dataType, numElems, value,
                      NULL_);
     if (status != CDF_OK) {
       free (value);
       cdf_status_handler (status, "010...PUT_, ENTRY_");
       return;
     }
     free (value);
   }
   strcat (addedAttrs, "FILLVAL ");
   status = CDFlib (CREATE_, ATTR_, "CATDESC", VARIABLE_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS) {
     cdf_status_handler (status, "011...CREATE_, ATTR_");
     return;
   }
   status = CDFlib (SELECT_, ATTR_NAME_, "CATDESC",
                             zENTRY_, varNum,
                    PUT_, zENTRY_DATA_, (long) CDF_CHAR,
                                        (long) strlen(dimName), dimName,
                    NULL_);
   if (status != CDF_OK) {
     cdf_status_handler (status, "011...PUT_, ENTRY_");
     return;
   }
   strcat (addedAttrs, "CATDESC ");
   status = CDFlib (CREATE_, ATTR_, "FORMAT", VARIABLE_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS) {
     cdf_status_handler (status, "012...CREATE_, ATTR_");
     return;
   }
   status = CDFlib (SELECT_, ATTR_NAME_, "FORMAT",
                             zENTRY_, varNum,
                    PUT_, zENTRY_DATA_, (long) CDF_CHAR, (long) 2, "I3",
                    NULL_);
   if (status != CDF_OK) {
     cdf_status_handler (status, "012...PUT_, ENTRY_");
     return;
   }
   strcat (addedAttrs, "FORMAT ");
   status = CDFlib (CREATE_, ATTR_, "FIELDNAM", VARIABLE_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS) {
     cdf_status_handler (status, "013...CREATE_, ATTR_");
     return;
   }
   status = CDFlib (SELECT_, ATTR_NAME_, "FIELDNAM",
                             zENTRY_, varNum,
                    PUT_, zENTRY_DATA_, (long) CDF_CHAR,
                                        (long) strlen(dimName), dimName,
                    NULL_);    
   if (status != CDF_OK) {     
     cdf_status_handler (status, "013...PUT_, ENTRY_");
     return;
   }
   strcat (addedAttrs, "FIELDNAM ");
   status = CDFlib (CREATE_, ATTR_, "LABLAXIS", VARIABLE_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS) {
     cdf_status_handler (status, "014...CREATE_, ATTR_");
     return;
   }
   status = CDFlib (SELECT_, ATTR_NAME_, "LABLAXIS",
                             zENTRY_, varNum,
                    PUT_, zENTRY_DATA_, (long) CDF_CHAR, 
                                        (long) strlen(dimName), dimName,
                    NULL_);
   if (status != CDF_OK) {
     cdf_status_handler (status, "014...PUT_, ENTRY_");
     return;
   }
   strcat (addedAttrs, "LABLAXIS ");
   status = CDFlib (CREATE_, ATTR_, "VAR_TYPE", VARIABLE_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS) {
     cdf_status_handler (status, "015...CREATE_, ATTR_");
     return;
   }
   status = CDFlib (SELECT_, ATTR_NAME_, "VAR_TYPE",
                             zENTRY_, varNum,
                    PUT_, zENTRY_DATA_, (long) CDF_CHAR,
                                        (long) strlen("support_data"),
                                        "support_data",
                    NULL_);    
   if (status != CDF_OK) {     
     cdf_status_handler (status, "015...PUT_, ENTRY_");
     return;
   }
   strcat (addedAttrs, "VAR_TYPE ");
   status = CDFlib (CREATE_, ATTR_, "VALIDMIN", VARIABLE_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS) {
     cdf_status_handler (status, "016...CREATE_, ATTR_");
     return;
   }
   value = (void *) malloc ((size_t) CDFelemSize (dataType));
   DefaultPadValue (dataType, numElems, value);
/*   value = (char) -127; */
   status = CDFlib (SELECT_, ATTR_NAME_, "VALIDMIN",
                             zENTRY_, varNum,
                    PUT_, zENTRY_DATA_, dataType, numElems, value,
                    NULL_);
   if (status != CDF_OK) {
     free (value);
     cdf_status_handler (status, "016...PUT_, ENTRY_");
     return;
   }
   strcat (addedAttrs, "VALIDMIN ");
   status = CDFlib (CREATE_, ATTR_, "VALIDMAX", VARIABLE_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS) {
     cdf_status_handler (status, "017...CREATE_, ATTR_");
     return;
   }
/*   value = (char) 127; */
   DefaultMaxValue (dataType, numElems, value);
   status = CDFlib (SELECT_, ATTR_NAME_, "VALIDMAX",
                             zENTRY_, varNum,
                    PUT_, zENTRY_DATA_, dataType, numElems, value,
                    NULL_);    
   if (status != CDF_OK) {     
     free (value);
     cdf_status_handler (status, "017...PUT_, ENTRY_");
     return;
   }
   free (value);
   strcat (addedAttrs, "VALIDMAX ");
   status = CDFlib (CREATE_, ATTR_, "SCALETYP", VARIABLE_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS) {
     cdf_status_handler (status, "018...CREATE_, ATTR_");
     return;
   }
   status = CDFlib (SELECT_, ATTR_NAME_, "SCALETYP",
                             zENTRY_, varNum,
                    PUT_, zENTRY_DATA_, (long) CDF_CHAR,
                                        (long) strlen("linear"), 
                                        "linear",
                    NULL_);
   if (status != CDF_OK) {
     cdf_status_handler (status, "018...PUT_, ENTRY_");
     return;
   }
   strcat (addedAttrs, "SCALETYP");
   status = CDFlib (CREATE_, ATTR_, "NC_ADDED", VARIABLE_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS) {
     cdf_status_handler (status, "018...CREATE_, ATTR_");
     return;
   }
   status = CDFlib (SELECT_, ATTR_NAME_, "NC_ADDED",
                             zENTRY_, varNum,
                    PUT_, zENTRY_DATA_, CDF_CHAR, (long) strlen(addedAttrs),
                                                  addedAttrs,
                    NULL_);
   if (status != CDF_OK) cdf_status_handler (status, "020...PUT_, ENTRY_");
}

/******************************************************************************
 * add_cdf_attribute_vartype.
 * This function adds VAR_TYPE variable attribute and its entries.
 ******************************************************************************/

void add_cdf_attribute_vartype (int ncid, int nVars) {
   CDFstatus status;
   int nStatus;
   int  id, dimid;
   char dimName[NC_MAX_NAME+1], varName[NC_MAX_NAME+1];
   long numElems, attrNum;

   status = CDFlib (CREATE_, ATTR_, "VAR_TYPE", VARIABLE_SCOPE, &attrNum,
                    NULL_);
   if (status != CDF_OK && status != ATTR_EXISTS) {
     cdf_status_handler (status, "015...CREATE_, ATTR_");
     return;
   }
   for (id = 0; id < nVars; id++) {
       nStatus = nc_inq_varname (ncid, id, varName);
       if (nStatus != NC_NOERR) handle_error(690, nStatus);
       status = CDFlib (SELECT_, ATTR_NAME_, "VAR_TYPE",
                                 zENTRY_NAME_, varName,
                        GET_, zENTRY_NUMELEMS_, &numElems,
                        NULL_);    
       if (status == CDF_OK) continue;
       nStatus = nc_inq_dimid (ncid, varName, &dimid);
       if (nStatus == NC_NOERR)
         status = CDFlib (PUT_, zENTRY_DATA_, (long) CDF_CHAR,
                                              (long) strlen("support_data"),
                                              "support_data",
                          NULL_);    
       else 
         status = CDFlib (PUT_, zENTRY_DATA_, (long) CDF_CHAR,
                                              (long) strlen("data"), "data",
                          NULL_);    
       if (status != CDF_OK) {     
         cdf_status_handler (status, "015...PUT_, ENTRY_");
         return;
       }
       status = CDFlib (SELECT_, ATTR_NAME_, "NC_ADDED",
                                 zENTRY_NAME_, varName,
                        GET_, zENTRY_NUMELEMS_, &numElems,
                        NULL_);
       if (status != CDF_OK) {
         status = CDFlib (PUT_, zENTRY_DATA_, CDF_CHAR, 8L, "VAR_TYPE",
                          NULL_);
       } else {
         char *new1 = (char *) malloc (sizeof(char) * ((int)numElems + 1));
         char *new2 = (char *) malloc (sizeof(char) * ((int)numElems + 2 + 8));
         status = CDFlib (GET_, zENTRY_DATA_, new1,
                          NULL_); 
         if (status != CDF_OK)
           cdf_status_handler (status, "021...GET_, ENTRY_");
         memcpy (new2, new1, (size_t)numElems);
         memcpy (new2+numElems, " ", (size_t)1);
         memcpy (new2+numElems+1, "VAR_TYPE", (size_t)8);
         status = CDFlib (PUT_, zENTRY_DATA_, (long) CDF_CHAR, (numElems + 9),
                                                               new2,
                          NULL_);
         free (new1);
         free (new2);
      }
      if (status != CDF_OK) cdf_status_handler (status, "022...PUT_, ENTRY_");
    }
}

void DefaultMaxValue (Int32 dataType, Int32 numElems, void *maxValue)
{
  size_t nBytes = (size_t) CDFelemSize (dataType);
  Byte1 *ptr = maxValue;
  char aNull = (char) 0;
  int i;
  switch (dataType) {
    case CDF_BYTE:
    case CDF_INT1: {
      sChar aINT1 = (sChar) DEFAULT_BYTE_MAXVALUE;
      for (i = 0; i < numElems; i++, ptr += nBytes)
        memmove (ptr, &aINT1, nBytes);
      break;
    }
    case CDF_UINT1: {
      uChar aUINT1 = (uChar) DEFAULT_UINT1_MAXVALUE;
      for (i = 0; i < numElems; i++, ptr += nBytes)
        memmove (ptr, &aUINT1, nBytes);
      break;
    }
    case CDF_INT2: {
      Int16 aINT2 = (Int16) DEFAULT_INT2_MAXVALUE;
      for (i = 0; i < numElems; i++, ptr += nBytes)
        memmove (ptr, &aINT2, nBytes);
      break;
    }
    case CDF_UINT2: {
      uInt16 aUINT2 = (uInt16) DEFAULT_UINT2_MAXVALUE;
      for (i = 0; i < numElems; i++, ptr += nBytes)
        memmove (ptr, &aUINT2, nBytes);
      break;
    }
    case CDF_INT4: {
      Int32 aINT4 = (Int32) DEFAULT_INT4_MAXVALUE;
      for (i = 0; i < numElems; i++, ptr += nBytes)
        memmove (ptr, &aINT4, nBytes);
      break;
    }
    case CDF_UINT4: {
      uInt32 aUINT4 = (uInt32) DEFAULT_UINT4_MAXVALUE;
      for (i = 0; i < numElems; i++, ptr += nBytes)
        memmove (ptr, &aUINT4, nBytes);
      break;
    }
    case CDF_INT8: {
      Int64 aINT8 = (Int64) DEFAULT_INT8_MAXVALUE;
      for (i = 0; i < numElems; i++, ptr += nBytes)
        memmove (ptr, &aINT8, nBytes);
      break;
    }
    case CDF_REAL4:
    case CDF_FLOAT: {
      float aREAL4 = (float) DEFAULT_REAL4_MAXVALUE;
      for (i = 0; i < numElems; i++, ptr += nBytes)
        memmove (ptr, &aREAL4, nBytes);
      break;
    }
    case CDF_REAL8:
    case CDF_DOUBLE: {
      double aREAL8 = (double) DEFAULT_REAL8_MAXVALUE;
      for (i = 0; i < numElems; i++, ptr += nBytes)
        memmove (ptr, &aREAL8, nBytes);
      break;
    }
    case CDF_EPOCH: {
      double aREAL8 = (double) DEFAULT_EPOCH_MAXVALUE;
      for (i = 0; i < numElems; i++, ptr += nBytes)
        memmove (ptr, &aREAL8, nBytes);
      break;
    }
    case CDF_EPOCH16: {
      double aREAL16[2];
      aREAL16[0] = (double) DEFAULT_EPOCH16_MAXVALUE;
      aREAL16[1] = (double) DEFAULT_EPOCH16_MAXVALUE;
      for (i = 0; i < numElems; i++, ptr += nBytes)
        memmove (ptr, &aREAL16, nBytes);
      break;
    }
    case CDF_TIME_TT2000: {
      long long aINT8 = (long long) DEFAULT_TT2000_MAXVALUE;
      for (i = 0; i < numElems; i++, ptr += nBytes)
        memmove (ptr, &aINT8, nBytes);
      break;
    }
    case CDF_CHAR:
    case CDF_UCHAR: {
      char aCHAR = DEFAULT_CHAR_MAXVALUE;
      memmove (ptr, &aCHAR, nBytes);
      ptr += nBytes;
      for (i = 1; i < numElems; i++, ptr += nBytes)
        memmove (ptr, &aNull, nBytes);
      break;
    }
  }
  return;
}

